/***************************************************************************
                             download_record.c
                             -----------------

    Code to conveniently get info from a downloadRecord structure in form
     of the strings that are going to be displayed in the list widget.


    begin                : Wed Jan 22 2003
    copyright            : (C) 2003 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "download_record.h"
#include "downloads_data.h"
#include "downloads_page.h"
#include "gaps.h"
#include "global.h"
#include "gui_options.h"
#include "misc.h"
#include "notebook.h"

/******************************************************************************
 *
 *   download_calculate_seconds_left_to_complete
 *
 *   calculates the estimated number of seconds
 *     left to complete the download
 */

void
download_calculate_seconds_left_to_complete (downloadRecord *dl)
{
	time_t 	secs_downloading;
	guint32	bytes_left;
	gfloat 	speed_in_bytes_per_sec;
	gfloat 	longrun_avg_speed_in_bytes_per_sec;
	gfloat 	secs_left_based_on_current_speed;
	gfloat 	secs_left_based_on_longrun_average;
	gfloat 	weight_long_avg;
	gfloat 	weight_current;

	g_return_if_fail (dl!=NULL);

	if (dl->speed <= 0.05)
	{
		dl->secs_left = 0;				/* not downloading => no estimate */
		return;
	}

	dl->lasttime = time(NULL);			/* last seen downloading: now */

	if (dl->transferred > dl->size)
	{
		dl->secs_left = -1;				/* -1 = 'not long' */
		return;
	}

	bytes_left = dl->size - dl->transferred;
	speed_in_bytes_per_sec = dl->speed * 1024.0;
	secs_left_based_on_current_speed = (bytes_left * 1.0) / (speed_in_bytes_per_sec + 0.001);

	/* do we have a useful 'time added' value? */
	secs_downloading = time(NULL) - dl->time_added;
	if (dl->time_added <= 0 || secs_downloading == 0 || secs_downloading == time(NULL))
	{
		dl->secs_left = (gint) secs_left_based_on_current_speed;
		return;
	}

	/* for now stick with the old system */
	dl->secs_left = (gint) secs_left_based_on_current_speed;
	return;

#ifdef HUMBUG_DO_NOT_COMPILE

	longrun_avg_speed_in_bytes_per_sec = (dl->transferred * 1.0) / (secs_downloading * 1.0);
	secs_left_based_on_longrun_average = (bytes_left * 1.0) / (longrun_avg_speed_in_bytes_per_sec + 0.001);

	g_assert (dl->size > 0);

	/* The closer we get to the end (=the smaller bytes_left is), the
	 *  less weight we assign to the long run average
	 * Note: x0.5, so total weight is still 1.0
	 *
	 * XXX - Problem: we are treating times that the download has been
	 *       paused just like as if it had been actively looking, which
	 *       is not really right.
	 *
	 *       Should weight_long_avg(bytes_left) be non-linear?
	 */

	weight_long_avg = ((bytes_left * 1.0) / (dl->size * 1.0));

	/* make sure the long run average only accounts for max. 30% */
	if (weight_long_avg > 0.3)
		weight_long_avg = 0.3;

	weight_current  = 1.0 - weight_long_avg;

	dl->secs_left = (gint) ( (secs_left_based_on_current_speed * weight_current)
							+ (secs_left_based_on_longrun_average * weight_long_avg));

/*
	g_print (N_("[%s] - %02.1fkB/s\nCUR %04.1fh - LRAVG %04.1fh - TOT: %04.1fh - WL: %02.2f + WC: %02.2f = WT: %02.2f\n"),
		dl->nojunkname,
		dl->speed,
		(secs_left_based_on_current_speed/3600.0),
		(secs_left_based_on_longrun_average/3600.0),
		((dl->secs_left*1.0)/3600.0),
		weight_long_avg, weight_current, weight_long_avg+weight_current
	);
*/
#endif

}


/******************************************************************************
 *
 *   download_get_time_left_string
 *
 */

const gchar *
download_get_time_left_string (downloadRecord *dl)
{
	static gchar timeleft[64];

	timeleft[0] = 0x00;
	g_return_val_if_fail (dl!=NULL, timeleft);

	if (dl->secs_left == 0)
		return timeleft;

	if (dl->secs_left > 3600)
	{
		g_snprintf (timeleft, sizeof(timeleft), N_("%2.1fh"), (dl->secs_left*1.0)/3600.0);
	}
	else if (dl->secs_left >= 60)
	{
		g_snprintf (timeleft, sizeof(timeleft), N_("%2.1fm"), dl->secs_left*1.0/60.0);
	}
	else if (dl->secs_left > 0)
	{
		g_snprintf (timeleft, sizeof(timeleft), N_("%us"), dl->secs_left);
	}
	else if (dl->secs_left < 0)
	{
		g_snprintf (timeleft, sizeof(timeleft), N_("%s"), _("not long"));
	}

	return timeleft;
}


/******************************************************************************
 *
 *   download_get_status_string
 *
 */

const gchar *
download_get_status_string (downloadRecord *dl)
{
	guint8 status;

	g_return_val_if_fail (dl != NULL, FIXME_STRING);

	status = dl->status;

	if (status >= STATUS_LAST)
		return FIXME_STRING;

	/* if it is downloading or was in the last 6 seconds, make it 'downloading...' */
	if (status == STATUS_LOOKING && (dl->speed > 0.0 || (time(NULL)-dl->lasttime) <= 6))
	{
		status = STATUS_DOWNLOADING;
	}

	switch (status)
	{
		case STATUS_HASHING     :	return _("(Hashing...)");
		case STATUS_QUEUED      :	return _("Queued.");
		case STATUS_LOOKING     :	return _("Looking...");
		case STATUS_DOWNLOADING :	return _("Downloading...");
		case STATUS_PAUSED      :	return _("Paused.");
		case STATUS_IDS         :	return _("Insufficient Disk Space.");
		case STATUS_NOSRCS      :	return _("No Sources.");
		case STATUS_DONE        :	return _("Done.");
		case STATUS_HASHING2    :	return _("Hashing...");
		case STATUS_ERRLOADING  :	return _("Error Loading.");
		case STATUS_COMPLETING  :	return _("Completing...");
		case STATUS_COMPLETE    :	return _("Complete.");
		case STATUS_CORRUPTED   :	return _("Complete, but corrupted.");
		case STATUS_ERRHASHING  :	return _("Error Hashing");
		case STATUS_TRANSFERRING:	return _("Transferring...");
		case STATUS_RESUMING    :	return _("Resuming...");
		case STATUS_PAUSING     :	return _("Pausing...");

		default:
		break;
	}

	g_assert_not_reached();
}


/******************************************************************************
 *
 *   download_get_transfered_string
 *
 */

const gchar *
download_get_transfered_string (downloadRecord *dl)
{
	static gchar transstr[64];

	transstr[0] = 0x00;
	g_return_val_if_fail (dl != NULL, FIXME_STRING);

	/* if meaningless transfered value, see if we can
	 *   calculate the real one from gaplist
	 */
	if (dl->transferred >= dl->size && (dl->gaprecord))
	{
		if ((dl->gaprecord)->gaplist)
		{
			dl->transferred = get_transfered_from_gaps(dl);
			/* Now the secs_left field is still set using the old value for
			 * 'transfered' though, so we might need to recalculate secs_left here based on the
			 * new 'transfered' value.
			 */
			download_calculate_seconds_left_to_complete(dl);
		}
	}

	if (option_dltransinpercent)
	{
		g_snprintf (transstr, sizeof(transstr), N_("%2.1f%%"), (100.0*dl->transferred)/(1.0*dl->size));
		return transstr;
	}

	g_snprintf (transstr, sizeof(transstr), N_("%s"), human_size(dl->transferred));
	return transstr;
}


/******************************************************************************
 *
 *   download_get_amount_left_string
 *
 */

const gchar *
download_get_amount_left_string (downloadRecord *dl)
{
	static gchar leftstr[64];

	g_return_val_if_fail (dl != NULL, FIXME_STRING);

	if (dl->transferred >= dl->size)
		return N_("???");

	g_snprintf (leftstr, sizeof(leftstr), N_("%s"), human_size(dl->size - dl->transferred));

	return leftstr;
}


/******************************************************************************
 *
 *   download_get_speed_string
 *
 */

const gchar *
download_get_speed_string (downloadRecord *dl, gboolean showunit)
{
	static gchar speedstr[64];

	g_return_val_if_fail (dl != NULL, FIXME_STRING);

	switch (dl->status)
	{
		case STATUS_PAUSED:
		case STATUS_PAUSING:
		case STATUS_RESUMING:
		case STATUS_HASHING:
		case STATUS_HASHING2:
		case STATUS_QUEUED:
		case STATUS_ERRLOADING:
		case STATUS_ERRHASHING:
		{
			dl->speed = 0.0;	// just making sure...
			return N_("");
		}
		default:
		break;
	}

	g_snprintf (speedstr, sizeof(speedstr), N_("%3.1f%s"), dl->speed, (showunit == TRUE) ? _(" kB/s") : N_(""));
	return speedstr;
}


/******************************************************************************
 *
 *   download_get_priority_string
 *
 */

const gchar *
download_get_priority_string (downloadRecord *dl)
{
	g_return_val_if_fail (dl != NULL, FIXME_STRING);

	switch (dl->priority)
	{
		case 0:	return _("low");
		case 1:	return _("normal");
		case 2:	return _("high");
		default:
		break;
	}

	return FIXME_STRING;
}


/******************************************************************************
 *
 *   download_get_availability_string
 *
 */

const gchar *
download_get_availability_string (downloadRecord *dl)
{
	static gchar availstr[32];

	g_return_val_if_fail (dl != NULL, FIXME_STRING);

	switch (dl->status)
	{
		case STATUS_HASHING:
		case STATUS_HASHING2:
		case STATUS_ERRLOADING:
		case STATUS_ERRHASHING:
		{
			return N_(" ");
		}
		default:
		break;
	}

	g_snprintf (availstr, sizeof(availstr), N_("%u%%"), dl->availability);

	return availstr;
}


/******************************************************************************
 *
 *   download_get_last_seen_downloading_string
 *
 */

const gchar *
download_get_last_seen_downloading_string (downloadRecord *dl, gboolean showago)
{
	time_t now = time(NULL);
	time_t last;

	g_return_val_if_fail ( dl != NULL, FIXME_STRING);

	last = dl->lasttime;

	if (last == now)
		return _("now");

	/* more than 2 years ago? => never */
	if ((now-last) > 60*60*24*365*2)
		return _("never");

	if (showago)
	{
		static gchar lastseenstr[64];
		g_snprintf (lastseenstr, sizeof(lastseenstr), _("%s ago"), seconds_into_human_time (now - last));
		return lastseenstr;
	}

	return seconds_into_human_time (now - last);
}


/******************************************************************************
 *
 *   download_get_last_seen_complete_string
 *
 */

const gchar *
download_get_last_seen_complete_string (downloadRecord *dl, gboolean showago)
{
	time_t now = time(NULL);
	time_t last;

	g_return_val_if_fail ( dl != NULL, FIXME_STRING);

	last = dl->lasttime_complete;

	if (last == now)
		return _("now");

	/* more than 2 years ago? => never */
	if ((now-last) > 60*60*24*365*2)
		return _("never");

	if (showago)
	{
		static gchar lastseenstr[64];
		g_snprintf (lastseenstr, sizeof(lastseenstr), _("%s ago"), seconds_into_human_time (now - last));
		return lastseenstr;
	}

	return seconds_into_human_time (now - last);
}


/******************************************************************************
 *
 *   download_set_stats
 *
 *
 */

void
download_set_stats ( guint16 index,
                     guint8 status,
                     gfloat speed,
                     guint32 trans,
                     guint8 avail,
                     guint8 sources)
{
	downloadRecord	*dl = downloads_get_record_by_index(index);

	if (!dl)
	{
		g_printerr (_("received DL stats for item not in list?! (index=%u)\n"), index);
		return;
	}

	dl->stats_okay = TRUE;
	dl->transferred = trans;
	dl->availability = avail;
	dl->sources = sources;

//	g_print ("got stats for %s: st=%u avail=%u, trans=%u, speed=%2.1f\n", dl->nojunkname, status, avail, trans,speed);

	/* don't change status from 'pausing...' into looking
	 * or 'resuming...' into paused ('cause of delay of core response)
	 *
     */

	if ((dl->status!=STATUS_PAUSING || status!=STATUS_LOOKING) &&
   		(dl->status!=STATUS_RESUMING || status!=STATUS_PAUSED))
	{
		dl->status = status;
	}

	if (dl->status != status)
		dl->lasttime_status_changed = time(NULL);

	if (dl->status == STATUS_CORRUPTED)
		dl->corrupted = TRUE;

	if (dl->availability == 100 &&
			(status == STATUS_LOOKING || status == STATUS_DOWNLOADING || status == STATUS_COMPLETING))
		dl->lasttime_complete = time(NULL);

	/* Note: Insuff. Disk Space and dlspeed>0 is possible, but probably
	 * wrong most of the time (core bug) (let's not print an error msg though,
	 * as this is a confirmed core bug) */

	if (speed >= 0.0 && speed < 2048.0)
		dl->speed = speed;

	if (   status == STATUS_ERRLOADING
		|| status == STATUS_ERRHASHING
		|| status == STATUS_IDS)
	{
		dl->speed = 0.0;
	}

	download_calculate_seconds_left_to_complete(dl);

	/* MORE THAN 255 SOURCES are a problem, because the GUI-core
	 * protocol only provides 8-bit for the number of sources.
	 * Currently the core effectively sends us num_sources&0xff, so
	 * when there are 257 sources, we'd see a '2'. Which is not really
	 * what we want.
	 *
	 * workaround:
	 * if the core is gap-enabled and we show progress bars, then we
	 * can scan through the progress bar gaplist and look at the gaps
	 * which mark the parts that are available but don't have yet.
	 * We will know the number of sources which have this one part
	 * (luckily it's a 16-bit value there). Although we cannot deduct
	 * from the gaplist how many different sources there are overall,
	 * we know can know the minimum number of different sources from
	 * the highest number of sources that has one and the same part of
	 * the file.
	 * Now, if the number of sources as provided by the core is LOWER
	 * then the minimum number of sources according to the gaplist,
	 * then we can conclude with some probability that we have to add
	 * another 256 to get the real number of sources.
	 * But of course our gaplist could just be horribly out of date or
	 * wrong as well.... */

	if (1==1)
	{
		gint min_sources = find_min_num_of_sources_from_gaplist(dl);
		/* require difference to be at least 20 to avoid wrong
		 * judgements due to less frequent update of the gaplist */
		if (min_sources>0 && dl->sources < min_sources && (min_sources-dl->sources) >= 20)
		{
			/* XXX - do we need to specify a minimum threshold value for the
			 * 'gaplist' value (e.g. 64)? */
			dl->sources += 256;
		}
	}

	/* save cpu power - don't redraw list if it's not visible */
	if (current_page == NOTEBOOK_PAGE_DOWNLOADS)
		downloadsclist_update_status(dl);
}

