/***************************************************************************
            downloads_data.c  -  handles the DL list + DL clist

	* add, remove, and update records to/from/in the downloads data list
	* add records to downloads clist
	* update values in downloads clist
	* clear download list and download clist

                             -------------------
    begin                : Fri Jul 5 2002
    copyright            : (C) 2001-2002 by Tim-Philipp Mller
    email                : t.i.m at orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// TODO:
// - move remove junk routines from misc.c here
// - replace gaps memory alloc+free with g_ routines
// - update only status of visible rows during status update?

#include "colors.h"
#include "downloads_attrib.h"
#include "downloads_data.h"
#include "downloads_page.h"
#include "gaps.h"
#include "global.h"
#include "gui_options.h"
#include "icons.h"
#include "mainwindow.h"
#include "misc.h"		// for remove_junk_from_filename
#include "notebook.h"

GPtrArray		*downloads = NULL;
GPtrArray		*completed = NULL;


/******************************************************************************
 *
 *  downloads_get_num
 *
 *  return number of items in the downloadlist
 *
 */

guint
downloads_get_num (void)
{
	if (!downloads)
		return 0;

	return (downloads->len);
}


/******************************************************************************
 *
 *  downloads_is_record_in_list
 *
 *  return TRUE if the given downloadRecord pointer
 *   is in the downloads list, otherwise FALSE.
 *
 *  (this is used when processing selections)
 *
 */

gboolean
downloads_is_record_in_list (downloadRecord *dl)
{
	guint i;

	g_return_val_if_fail (dl != NULL, FALSE);

	if (!downloads)
		return FALSE;

	for (i=0; i < downloads->len; i++)
	{
		gpointer data = g_ptr_array_index (downloads, i);
		if (data == (gpointer)dl)
			return TRUE;
	}

	return FALSE;
}


/******************************************************************************
 *
 *  downloads_set_all_to_no_stats_received
 *
 *  sets all downloads to 'no stats received', so we know if we
 *   we haven't gotten any stats for a particular download
 *
 *  returns TRUE if there are download records, otherwise FALSE
 */

gboolean
downloads_set_all_to_no_stats_received (void)
{
	if (!downloads)
		return FALSE;

	if (downloads->len == 0)
		return FALSE;

	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if (item)
		{
			item->stats_okay = FALSE;
		}
	);

	return TRUE;
}


/******************************************************************************
 *
 *  downloads_check_for_stats_received
 *
 *  checks whether we received stats for all downloads 
 *
 */

void
downloads_check_for_stats_received (void)
{
	if (!downloads)
		return;

	if (downloads->len == 0)
		return;

	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if ((item) && item->stats_okay == FALSE)
		{
				g_printerr (_("we did not receive DL stats for '%s'\n"), item->name);

				item->sources = 0;
				item->speed = 0.0;
				item->availability = 0;

				/* save cpu power - don't redraw list if it's not visible */
				if (current_page == NOTEBOOK_PAGE_DOWNLOADS)
					downloadsclist_update_status (item);
		}
	);
}


/******************************************************************************
 *
 *  downloads_get_record_by_index
 *
 *  returns the n-th download in the list, otherwise NULL
 *
 */

downloadRecord *
downloads_get_record_by_index (guint n)
{
	if (!downloads)
		return NULL;

	if (downloads->len == 0)
		return NULL;

	if (downloads->len <= n)
		return NULL;

	return (downloadRecord*) g_ptr_array_index(downloads,n);
}





/******************************************************************************
 *
 *  downloads_pause_all
 *
 *  pauses all downloads and prints 'only x MB left' message
 *
 */

void
downloads_pause_all (gfloat space_left_on_device)
{
	static time_t last_called = 0;
	gint          num_paused = 0;

	if (!downloads)
		return;

	/* problem: a send_pause() will also request update on the space stats,
	 * which in turn will again trigger an emergency pause all if we don't
	 * break this cycle here
	 */
	if (last_called!=0)
	{
		if ( (time(NULL)-last_called) < 120)
			return;
	} else last_called = time(NULL);

	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if ((item) && item->status!=STATUS_PAUSED)
		{
			send_pause_download(item->hash);
			num_paused++;
		}
	);

	if (num_paused>0)
	{
		status_message_blue (_("GUI: Only %3.1f MB free in the temp folder! Pausing all downloads.\n"), space_left_on_device);
	}
}


// downloads_find_from_hash
//
// Finds a downloadRecord from the file hash
// Returns downloadRecord or NULL if not found

downloadRecord *
downloads_find_from_hash (const guint8 *hash)
{
	g_return_val_if_fail (hash!=NULL, NULL);
	if (!downloads) return NULL;
	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if ((item) && memcmp(item->hash, hash, 16)==0) return item;
	);
	return NULL;
}


// downloads_find_from_filename
//
// Finds a downloadRecord from the file name
// Returns downloadRecord or NULL if not found

downloadRecord *
downloads_find_from_filename (const gchar *filename)
{
	g_return_val_if_fail (filename!=NULL, NULL);
	if (!downloads) return NULL;
	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if ((item) && (item->name) && strcmp(filename, item->name)==0) return item;
	);
	return NULL;
}



static void
downloads_free_record (downloadRecord *item)
{
	g_return_if_fail (item!=NULL);

	G_FREE(item->name);
	G_FREE(item->nojunkname);
	G_FREE(item->type);
	G_FREE(item->format);

	if (ITEM_GAPRECORD)
	{
		free_pixmaps (item);
		if (ITEM_GAPLIST)
			free (ITEM_GAPLIST);
		free (ITEM_GAPRECORD);
	}

	memset(item,0x00,sizeof(item));
	g_free(item);		// XXX - why was this missing before? (TP,30/08/2002)
}

// downloads_remove_record
//
// Removes and frees a record from the downloads list (and clist).

void
downloads_remove_record (downloadRecord *dl)
{
	g_return_if_fail (dl!=NULL);
	(void) downloadsclist_remove_record (dl);

	if (!downloads) return;
	// Do NOT use g_ptr_array_remove_fast() here - order matters!
	(void) g_ptr_array_remove (downloads, (gpointer)dl);
	downloads_free_record (dl);
}

// downloads_cleanup
//
// frees all downloads items

void
downloads_cleanup (void)
{
	if (!downloads) return;

	if (clist_downloads)
		gtk_clist_freeze(DL_CLIST);

	PROCESS_GPTRARRAY_FROM_END ( downloads, downloadRecord,
		if (item) downloads_remove_record (item);
	);

	if (clist_downloads)
		gtk_clist_thaw(DL_CLIST);
}

// downloads_completed_cleanup
//
// frees all 'completed' items

void
downloads_completed_cleanup (void)
{
	if (!completed) return;

	if (clist_downloads)
		gtk_clist_freeze(DL_CLIST);

	PROCESS_GPTRARRAY_FROM_END ( completed, downloadRecord,
		if (item)
		{
			(void) downloadsclist_remove_record (item);
			// Do NOT use g_ptr_array_remove_fast() here - order matters!
			(void) g_ptr_array_remove (completed, (gpointer)item);
			downloads_free_record (item);
		}
	);

	if (clist_downloads)
		gtk_clist_thaw(DL_CLIST);

	g_ptr_array_free(completed, TRUE);
	completed = NULL;
}


// downloads_add_record
//
// Add a new record to the downloads list. The order of the records in the list is
//	important. It needs to be exactly the same order the core has, otherwise
//	we don't assign the status information to the right records

void
downloads_add_record (const guint8 *hash, const gchar *name, guint32 size, guint partnum, guint8 priority)
{
	downloadRecord	*dl;

	g_return_if_fail (hash!=NULL);
	g_return_if_fail (name!=NULL);
	g_return_if_fail (downloads_find_from_hash(hash)==NULL);

	if (!downloads) downloads = g_ptr_array_new();
	g_return_if_fail (downloads!=NULL);

	dl = g_new0 (downloadRecord, 1);
	g_return_if_fail (dl!=NULL);

	g_memmove (dl->hash, hash, 16);
	dl->name = g_strdup(name);
	dl->nojunkname = g_strdup(name);
	remove_junk_from_filename (dl->nojunkname);
	dl->size = size;
	dl->partnum = partnum;
	dl->availability = 0;
	dl->cancelled = dl->corrupted = FALSE;
	dl->format = g_strdup(misc_get_extension_from_filename (name));
	dl->type = g_strdup(misc_get_filetype_from_extension (dl->format));
	dl->gaprecord = NULL;
	dl->transferred = 0;
	dl->sources = 0;
	dl->status = 0;
	dl->speed = 0.0;
	dl->priority = priority;
	dl->lasttime = 0;
	dl->lasttime_complete = 0;
	dl->time_added = time(NULL);
	dl->lasttime_status_changed = 0;
	dl->lasttime_got_gaps = 0;

	// if we have time attributes in the downloads attributes list
	//	set the download record's values to the ones we have
	//	saved (Note: new download must be last item in array)
	downloadsattriblist_set_download_record_attributes (dl);

	dl->last_scheme = DL_COL_SCHEME_NONE;
	dl->last_txtcol = NULL;
	g_ptr_array_add (downloads, (gpointer) dl);
	downloadsclist_add_record (dl);
}


// downloadsclist_resort
//
// makes sure the clist is sorted after sth has been changed

void
downloadsclist_resort (void)
{
	gtk_clist_set_auto_sort (DL_CLIST, FALSE);
	gtk_clist_set_sort_column (DL_CLIST, downloadlist_sortcolumn);
	gtk_clist_set_sort_type (DL_CLIST, downloadlist_sorttype);
	gtk_clist_set_compare_func (DL_CLIST, downloadclist_compare_func);
	gtk_clist_sort (DL_CLIST);
}



// downloadsclist_add_record
//
// adds a record (row) to the downloads clist

void
downloadsclist_add_record (downloadRecord *dl)
{
	gchar	*cols[NO_OF_DOWNLOAD_COLS];
	gint	 i, row;

	g_return_if_fail (DL_CLIST!=NULL);
	g_return_if_fail (dl!=NULL);

	for (i=0; i<NO_OF_DOWNLOAD_COLS; i++)
		cols[i] = N_("");

	// column 2 - file size
	cols[2] = (gchar*) human_size (dl->size);

	// column 12 - file hash (currently not set or shown)
//	cols[12] = hash_to_hash_str (dl->hash);

	row = gtk_clist_append (DL_CLIST, cols);

	// column 0 - file name (is pixtext, that's why we have to do it separately)
	if (1==1)
	{
		gchar *name = (option_nofnjunk==1) ? dl->nojunkname : dl->name ;
		gchar *tempfn = (option_DLShowPartFileName==1 && dl->partnum>0) ? g_strdup_printf(N_(" (%u.part)"), dl->partnum) : g_strdup(N_(""));
		gchar *col0txt, *pos;
		if (!name) name = (dl->name) ? dl->name : N_("(null)");
		col0txt = g_strdup_printf (N_("%s%s"), name, tempfn);

		/* replace non-printable characters in this locale by spaces */
		pos = col0txt;
		while (*pos)
		{
			if (!isprint(*pos))
				*pos = ' ';
			pos++;
		}

		gtk_clist_set_pixtext (DL_CLIST, row, 0, col0txt, 2, dotwhite_pm, dotwhite_bm);
		g_free (tempfn);
		G_FREE(col0txt);
	}

	gtk_clist_set_row_data (DL_CLIST, row, (gpointer) dl);
	downloadsclist_set_columns_visibility ();
}


// downloadsclist_add_completed_record
//
// adds a completed record (row) to the downloads clist

void
downloadsclist_add_completed_record (downloadRecord *dl)
{
	gchar	*cols[NO_OF_DOWNLOAD_COLS];
	gint	 i, row;

	g_return_if_fail (DL_CLIST!=NULL);
	g_return_if_fail (dl!=NULL);

	for (i=0; i<NO_OF_DOWNLOAD_COLS; i++)
		cols[i] = N_("");

	cols[1] = (gchar*) download_get_status_string(dl);

	// column 2 - file size
	cols[2] = g_strdup(human_size (dl->size));

	// column 4 - left
	cols[4] = g_strdup(human_size(0));

	// column 12 - file hash (currently not set or shown)
	//	cols[12] = hash_to_hash_str (dl->hash);

	row = gtk_clist_append (DL_CLIST, cols);

	/* column 0 - file name (is pixtext, that's why we have to do it separately)
	 * XXX - what kind of temp filename do we show here if the file is completed and fine?
	 */
	if (1==1)
	{
		gchar *name = (option_nofnjunk==1) ? dl->nojunkname : dl->name ;
		gchar *tempfn = (option_DLShowPartFileName==1 && dl->partnum>0) ? g_strdup_printf(N_(" (%u.part)"), dl->partnum) : g_strdup(N_(""));
		gchar *col0txt;
		if (!name) name = (dl->name) ? dl->name : N_("(null)");
		col0txt = g_strdup_printf (N_("%s%s"), name, tempfn);
		gtk_clist_set_pixtext (DL_CLIST, row, 0, col0txt, 2, dotgreen_pm, dotgreen_bm);
		g_free(tempfn);
		G_FREE(col0txt);
	}

	gtk_clist_set_row_data (DL_CLIST, row, (gpointer) dl);
	downloadsclist_set_columns_visibility ();

	G_FREE(cols[2]);
	G_FREE(cols[4]);
	downloadsclist_resort ();
}



// downloadsclist_update_status
//
// update all those status values in the clist

void
downloadsclist_update_status (downloadRecord *dl)
{
	gint		 row;
	gchar		 coltxtfix[1024];
	downloadRecord	*item = dl;	// we use this so we can use the ITEM_GAP* makros ...
	GdkPixmap	*pm;
	GdkBitmap	*bm;
	GdkColor	*txtcol = &black;
	ColorSchemeDL	 colScheme = DL_COL_SCHEME_DEFAULT;

	g_return_if_fail (dl!=NULL);
	
	row = gtk_clist_find_row_from_data (DL_CLIST, (gpointer)dl);
	g_return_if_fail (row>=0);

	/*  Note: columns 0+2 (filename+size) don't need frequent updating */
	gtk_clist_set_text (DL_CLIST, row, 1, download_get_status_string(dl));
	gtk_clist_set_text (DL_CLIST, row, 3, download_get_transfered_string(dl));
	gtk_clist_set_text (DL_CLIST, row, 4, download_get_amount_left_string(dl));
	gtk_clist_set_text (DL_CLIST, row, 5, download_get_speed_string(dl, (DL_CLIST->column[5]).width > 70));
	gtk_clist_set_text (DL_CLIST, row, 6, download_get_time_left_string(dl));
	gtk_clist_set_text (DL_CLIST, row, 7, download_get_priority_string(dl));
	g_snprintf (coltxtfix, sizeof(coltxtfix), N_("%u"), dl->sources);
	gtk_clist_set_text (DL_CLIST, row, 9, coltxtfix);
	gtk_clist_set_text (DL_CLIST, row, 10, download_get_last_seen_downloading_string (dl,(DL_CLIST->column[10]).width > 60));
	gtk_clist_set_text (DL_CLIST, row, 11, download_get_last_seen_complete_string (dl, (DL_CLIST->column[11]).width>60));


	// 8 - % available
	// IMPORTANT: if col # is changed, change pixtext thingy below as well!
	//            and _column_resized() thingy in downloads_page.c
	//		and in free_pixmaps() in gaps.c
	if (option_progressbar==1 && (dl->gaprecord))
	{
		// progress bars...
		if (ITEM_GAPLIST)
		{
			// gaplist but no pixmap? => create the pixmap first
			//  but only if the row is visible
			if ((!ITEM_PIXMAP) &&
				(gtk_clist_row_is_visible (DL_CLIST,row)!=GTK_VISIBILITY_NONE || option_visibleonly==0))
			{
				create_gaplist_pixmap (dl);
			}
			// pixmap there? => show it
			if (ITEM_PIXMAP)
			{
				// but: every set_pixmap increses ref count by 1
				// therefore: make sure, we don't set the same
				// pixmap again if it's already there, otherwise
				// we'll have problems freeing the memory, as we
				// wouldn't know how many unref()s to do
				// - this way ref count stays 1 and we only need
				// to do one unref() to free the pixmap
				GdkPixmap *pm = NULL;
				GdkBitmap *bm = NULL;
				gtk_clist_get_pixmap (DL_CLIST, row, 8, &pm, &bm);
				if (pm!=ITEM_PIXMAP)
				{
					gtk_clist_set_pixmap (DL_CLIST, row, 8, ITEM_PIXMAP, ITEM_BITMAP);
				}
			}
		}
	}

	/* if no progress bars, show availability as 'x%' */
	if (option_progressbar == 0 || (!dl->gaprecord))
	{
		gtk_clist_set_text (DL_CLIST, row, 8, download_get_availability_string(dl));
	}


	// now let's see which colour the status dot should have
	if (dl->status!=STATUS_PAUSED && dl->status!=STATUS_PAUSING && dl->status!=STATUS_RESUMING
		&& dl->status!=STATUS_HASHING && dl->status!=STATUS_HASHING2 && dl->status!=STATUS_QUEUED)
	{
		if (dl->availability<100)
		{
			if (dl->sources!=0)
			{
				// potential dead file gets marked with yello
				// files without full availability, but with sources and where we can still get something
				//  get marked purple
				gfloat	fpercent = (item->transferred*100.0)/(item->size*1.0);
				guint	percent  = (gint) fpercent;
				if (fpercent<100.0 && percent==100) percent=99;
				if (dl->availability <=percent && dl->availability>0)
				{
					colScheme = DL_COL_SCHEME_YELLOW;
				} else colScheme = DL_COL_SCHEME_PURPLE;
			} else colScheme = DL_COL_SCHEME_RED;
		} else colScheme = DL_COL_SCHEME_GREEN;
	}
	if (dl->status==STATUS_ERRLOADING || dl->status==STATUS_ERRHASHING) colScheme = DL_COL_SCHEME_ALLRED;
	colors_set_downloadsclist_scheme (colScheme, &pm, &bm, &txtcol);

	// don't redraw this if it hasn't been changed
	if (1==1)
//	if (dl->last_scheme!=colScheme || dl->last_txtcol!=txtcol)
	{
		gchar		*name = NULL, *newname = NULL;
		GdkPixmap	*dummy2, *dummy3;
		guint8		 dummy1;
		if (gtk_clist_get_pixtext (DL_CLIST, row, 0, &name, &dummy1, &dummy2, (GdkBitmap**)&dummy3)==1)
		{
			newname = g_strdup(name);
			gtk_clist_set_pixtext (DL_CLIST, row, 0, (newname!=NULL) ? newname : N_("???"), 2, pm, bm);	
			g_free(newname);
		}
		gtk_clist_set_foreground (DL_CLIST, row, txtcol);
		dl->last_scheme = colScheme;
		dl->last_txtcol = txtcol;
	}
}


// downloadsclist_update_all_filenames
//
// if for instance some option changes, we need to re-fill the file name fields
//	in the clist according to the new options, or do an initial fill in the first
//	place.

void
downloadsclist_update_all_filenames (void)
{
	if (!downloads) return;
	gtk_clist_freeze (DL_CLIST);
	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if (item)
		{
			gint row = gtk_clist_find_row_from_data (DL_CLIST, (gpointer)item);
			if (row>=0)
			{
				gchar		*coltxt;
				gchar		*oldname;
				gchar		*name;
				gchar		*tempfn = (option_DLShowPartFileName==1 && item->partnum>0) ? g_strdup_printf(N_(" (%u.part)"), item->partnum) : g_strdup(N_(""));
				guint8		 spacing;
				GdkBitmap	*bm;
				GdkPixmap	*pm;

				if (item->nojunkname)
				{
					g_free(item->nojunkname);
					item->nojunkname = (item->name) ? g_strdup(item->name) : NULL;
					if (item->nojunkname) remove_junk_from_filename (item->nojunkname);
				}
				name = (option_nofnjunk==1) ? item->nojunkname : item->name ;
				if (!name) name = (item->name) ? item->name : N_("(null)");
				coltxt = g_strdup_printf (N_("%s%s"), name, tempfn);
				g_return_if_fail (coltxt!=NULL);
				// we don't want 'oldname' - we just want the bitmaps
				(void) gtk_clist_get_pixtext (DL_CLIST, row, 0, &oldname, &spacing, &bm, &pm);
				gtk_clist_set_pixtext (DL_CLIST, row, 0, coltxt, spacing, bm, pm);
				g_free (tempfn);
				g_free (coltxt);
				coltxt = oldname;	// to get rid of compiler warning
			}
		}
	);
	gtk_clist_thaw (DL_CLIST);
}



// downloadsclist_set_columns_visibility
//
// Sets the download clist columns' visibilities how they are supposed to be

void
downloadsclist_set_columns_visibility (void)
{
	guint i;
	// set column visibility as requested by user - hash INVISIBLE for now
	for (i=0; i<NO_OF_DOWNLOAD_COLS-1; i++)
	{
		gtk_clist_set_column_visibility (DL_CLIST,i, (dlcols_visibility[i]==1));
	}
	gtk_clist_set_column_visibility (DL_CLIST,NO_OF_DOWNLOAD_COLS-1, FALSE); // hash col
}


// downloadsclist_update_status_all
//
// Updates the values in the clist to the values in the records. Called after we
//	haven't updated the clist for a while (e.g. because it was not visible)

void
downloadsclist_update_status_all (void)
{
	if (!downloads) return;
	gtk_clist_freeze(DL_CLIST);
	PROCESS_GPTRARRAY( downloads, downloadRecord,
		if (item) downloadsclist_update_status (item);
	);
	gtk_clist_thaw(DL_CLIST);
}


// downloadsclist_remove_record
//
// Returns TRUE if successfully removed, otherwise FALSE

gboolean
downloadsclist_remove_record (downloadRecord *dl)
{
	gint row;
	if ((!window) || (!DL_CLIST)) return FALSE;
	row = gtk_clist_find_row_from_data (DL_CLIST, (gpointer)dl);
	if (row<0) return FALSE;
	gtk_clist_remove (DL_CLIST, row);
	return TRUE;
}


/* downloads_completed_new_from_download
 *
 * adds a copy of the given download record to the 'completed' list/array
 * and to the downloads clist
 *
 */

void
downloads_completed_new_from_download (const downloadRecord *dl)
{
	downloadRecord	*cdl;

	g_return_if_fail (dl!=NULL);

	if (!completed)
	{
		completed = g_ptr_array_new();
		g_return_if_fail (completed!=NULL);
	}

	cdl = g_new0 (downloadRecord,1);
	g_return_if_fail (cdl!=NULL);

	memcpy(cdl->hash, dl->hash, 16);
	cdl->name = g_strdup(dl->name);
	cdl->nojunkname = g_strdup(dl->nojunkname);
	cdl->type = g_strdup(dl->type);
	cdl->format = g_strdup(dl->format);
	cdl->size = dl->size;
	cdl->transferred = dl->size;
	cdl->status = (dl->corrupted) ? STATUS_CORRUPTED : STATUS_COMPLETE ;
	cdl->availability = 100;
	cdl->sources = 999;		/* so people who sort by no. of source have it at the top */
	downloadsclist_add_completed_record (cdl);

	g_ptr_array_add (completed, (gpointer)cdl);
	
	return;
}


/* downloads_clear_all_completed
 *
 *
 */

void
downloads_clear_all_completed (void)
{
	if (!completed)
		return;

	gtk_clist_freeze(DL_CLIST);
	PROCESS_GPTRARRAY (completed, downloadRecord,
		if (item)
		{
			gint row = gtk_clist_find_row_from_data (DL_CLIST, (gpointer)item);
			if (row>=0)
			{
				gtk_clist_remove (DL_CLIST,row);
				downloads_free_record (item);
			}
		}
	);
	g_ptr_array_free(completed,TRUE);
	completed = NULL;
	gtk_clist_thaw(DL_CLIST);
}



