/*
*  http_get.c for ed2k_gui
*
*  GET HTML PAGE (SERVERLIST) VIA HTTP PROTOCOL
*
* (c) 2001-2002 TP Muller <t.i.m@orange.net>
*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "config.h"
#include "http_get.h"
#include "send_command.h"
#include "parse_message.h"
#include "search_data.h"
#include "search_page.h"
#include "servers_page.h"
#include "status_page.h"
#include "gui_options.h"
#include "misc.h"
#include "misc_strings.h"
#include "global.h"
#include "tag.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <gtk/gtk.h>
#define GNET_EXPERIMENTAL
#include <gnet/gnet.h>

// global variables

// local variables and defines

#warning ***
#warning *** find appropriate timeout
#warning *** is it in secs or msecs?
#warning ***
#define HTTP_GET_TIMEOUT	45*1000


// local functions

static gboolean		 http_get_connection_callback (GConn *conn, GConnStatus status, gchar *buffer,
							gint length, gpointer getstring);
static void			 http_get_process_serverlist_page (gchar *page, guint bufsize, gboolean clearlist, gint redirections);
static void			 http_get_connection_cleanup (GConn *conn, struct http_get_callback_data *data);
static gchar		*http_get_beginning_of_data (const gchar *page);
static gint			 extract_server_links_from_metfile (const gchar *data, guint datasize, gboolean clearlist);
static gboolean		 http_get_ip_is_evil (guint8 n1, guint8 n2, guint8 n3, guint8 n4, const gchar **offender);


// extract_server_links
//
// Takes a page with ed2k://|server|-links and extracts them one by one and
//	adds them to the server list. It also clears the list before the first
//	link is added on request.
//
// returns number of added servers

gint
extract_server_links (const gchar *pagebuf, gboolean clear_serverlist)
{
	guint32	n1, n2, n3, n4, port, num=0, invalid=0, new=0;
	gchar	**linkarray, **link, extrapriority;
	g_return_val_if_fail (pagebuf!=NULL, 0);

	linkarray = g_strsplit (pagebuf, N_("|server|"), 750);

	if (!linkarray) return 0;
	link = linkarray;
	while (*link)
	{
		gboolean	 foundlink;
		const gchar	*offender;

		foundlink = TRUE;

		/* We accept two formats - the standard format (5 parameter) and EXP's format, with 
		 * an additional figure '0'-'2' at the end specifying the priority of the server
		 */

		if (sscanf (*link, N_("%u.%u.%u.%u|%u|%c"), &n1, &n2, &n3, &n4, &port, &extrapriority) != 6)
		{
			extrapriority = '0';	/* normal priority if no priority is specified */
			if (sscanf (*link, N_("%u.%u.%u.%u|%u|"), &n1, &n2, &n3, &n4, &port) != 5)
			{
				foundlink = FALSE;
			}
		} 

		if (foundlink)
		{
			if (n1==10 || (n1==172 && n2>=16 && n2<=31) || (n1==192 && n2==168) || (n1==127 && n2==0 && n3==0))
			{
				status_message_blue (_("GUI: server list: skipped useless server with private IP %u.%u.%u.%u.\n"), n1,n2,n3,n4);
				foundlink = FALSE;
			}
		}

		if (foundlink)
		{
			if (http_get_ip_is_evil (n4, n3, n2, n1, &offender))
			{
				status_message_blue (_("GUI: server list: skipped suspicious IP from '%s'.\n"), offender);
				foundlink = FALSE;
			}
		}

		if (foundlink)
		{
			guint32	ip, priority=0;
			ip = GUINT32_TO_LE (n1 | (n2 << 8) | (n3 << 16) | (n4 << 24));

			if (extrapriority == '1') priority = 1;
			if (extrapriority == '2') priority = 2;

//			g_print ("====> server %u.%u.%u.%u : %u\n", n1,n2,n3,n4,port);
			if (!servers_find_from_address (ip,port))
			{
				// get rid of all servers if wanted
				if (clear_serverlist==1 && num==0)
				{
					serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_ALL_FLAG));
				}

				send_add_server (ip, port);
				if (priority>0) send_set_server_priority (ip, port, priority);
				new++;
			}
			num++;
		} else if (*(link+1)!=NULL && link!=linkarray) invalid++;	// last+first line can be invalid
		link++;
	}

	status_msg (_("Fetched %u servers. %u were new. %u links were invalid. Total now: %u servers\n"),
				 num, new, invalid, (servers!=NULL) ? servers->len+new : new);

	g_strfreev(linkarray);
	return num;
}



/* extract_ed2k_links
 *
 * (call misc_string_unescape_url() before if necessary)
 *
 */

gint
extract_ed2k_links (const gchar *page)
{
	gchar		*n;
	gint	 	 num = 0;
	searchList	*sl = NULL;

	g_return_val_if_fail (page!=NULL,-1);

	if (option_download_ed2k_link_immediately == 0)
	{
		if (searchlists)
		{
			PROCESS_GPTRARRAY ( searchlists, searchList,
				if ((item) && strcmp(item->label, LINK_SEARCHLIST_LABEL)==0)
				{
					sl = item;
					break;
				}
			);
		}
		if (!sl) sl = searchlist_make_new_searchlist (LINK_SEARCHLIST_LABEL);
		g_return_val_if_fail (sl!=NULL,-1);
	} else sl=NULL;

	n = strstr (page, N_("ed2k:"));
	while (n)
	{
		gchar	*fn, *hashstr, hash[16];
		guint	 size;

		fn = NULL;
		size = 0;
		hashstr = NULL;

		if (misc_check_if_ed2klink_is_valid(n,&fn,&size,&hashstr))
		{
			if ((fn) && (hashstr) && size>0)
			{
				memcpy (hash, hash_str_to_hash(hashstr), 16);

				/* remove '/' characters that some tossers put into filenames */
				(void) g_strdelimit (fn, N_("/"), '_');

				if (option_download_ed2k_link_immediately == 0)
				{
					if (sl)
					{
						guint idxpos=0, pagenum;
						if ((sl) && (sl->results))
							idxpos = (sl->results)->len;
						searchlist_add_or_update_record (sl, idxpos, hash, size, 99999, fn, NULL, NULL, NULL, NULL, NULL, NULL, 0, NULL);
						/* and switch to ed2k-links page */
						pagenum = gtk_notebook_page_num (GTK_NOTEBOOK(search_notebook), sl->frame);
						if (pagenum>=0) gtk_notebook_set_page (GTK_NOTEBOOK(search_notebook), pagenum);
					} else g_printerr (_("couldn't find or make 'ed2k-link' searchlist!"));
				} else send_new_download (hash, size, fn); /* download immediately! */
				
				num++;
				G_FREE(fn);
				G_FREE(hashstr);
			}
		}
		n = strstr (n+15, N_("ed2k:"));
	}

	// we only want to do this after all the links have been added, not after
	//	every single item
	if ((sl) && (sl->clist))
		searchclist_resort(sl->clist);

	return num;
}


// http_get_retrieve_serverlist
//
// Asynchroneously (ie. without blocking) retrieves a page from an http-server and
//	passes the result on to a routine to extract the ed2k-server links on that
//	page.

void
http_get_retrieve_serverlist (const gchar *url, guint redirections)
{
	struct http_get_callback_data	*data;
	GConn				*conn;
	GURL				*fetchurl, *proxyurl = NULL;
	gchar				*proxyenv, *firstgetline, *conn_host;
	guint				 conn_port;

	g_return_if_fail (url!=NULL);

	if (redirections >= MAX_NUMBER_OF_REDIRECTIONS)
	{
		status_warning (_("GUI: Too many redirections while trying to fetch server list! Giving up.\n"));
		return;
	}

	fetchurl = gnet_url_new (url);
	if (!fetchurl)
	{
		g_printerr (_("GUI: Could not transform URL to GURL structure!\n"));
		return;
	}

	// check server list URL
	if (!misc_gnet_http_url_okay(fetchurl))
	{
		status_msg (_("GUI: malformed http URL\n"));
		return;
	} else {
		// use standard port 80 if no port is specified
		if (fetchurl->port==0) gnet_url_set_port (fetchurl, 80);
		status_msg (_("GUI: Parsed URL alright (%s)\n"), gnet_url_get_nice_string(fetchurl));
	}

	// check if the http proxy environment variable is set (and not empty)
	if (!(proxyenv=getenv (N_("http_proxy"))))
	{
		proxyenv = getenv (N_("HTTP_PROXY"));
	}
	if ((proxyenv) && (*proxyenv!=0x00))
	{
		proxyurl = gnet_url_new (proxyenv);
	}

	if (proxyurl)	// using http proxy ...
	{
		status_msg (_("GUI: proxy environmental variable (http_proxy) is set ...\n"));

		// check proxy server URL
		if (!misc_gnet_http_url_okay(proxyurl))
		{
			/* maybe the '/' at the end has been forgotten? */
			if ((proxyurl)&&(!proxyurl->resource))
				gnet_url_set_resource(proxyurl, N_("/"));

			if (!misc_gnet_http_url_okay(proxyurl))
			{
				status_msg (_("GUI: malformed proxy URL ('%s')\n"), (getenv (N_("http_proxy"))!=NULL) ? getenv (N_("http_proxy")) : getenv (N_("HTTP_PROXY")));
				status_msg (_("GUI: proxy URL must be in this format: 'http://<host>:<port>/'\n"));
				return;
			}
		}

		status_msg (_("GUI: Parsed proxy URL alright (%s)\n"), gnet_url_get_nice_string(proxyurl));

		if (proxyurl->port==0)
		{
			gnet_url_set_port (proxyurl, 8080);
			status_msg (_("GUI: assuming standard port 8080 for http proxy.\n"));
		}

		firstgetline = g_strdup_printf (N_("GET http://%s:%u%s HTTP/1.1\r\n"),
							fetchurl->hostname, fetchurl->port, fetchurl->resource);
		conn_host = proxyurl->hostname;
		conn_port = proxyurl->port;

	} else {	// NOT using http proxy ...
		gchar *query = (fetchurl->query) ? g_strdup_printf (N_("?%s"), fetchurl->query) : strdup("");
		firstgetline = g_strdup_printf ("GET %s%s HTTP/1.1\r\n", fetchurl->resource, query);
//		g_print ("query = '%s'\nfragment = '%s'\n", fetchurl->query, fetchurl->fragment);
		conn_host = fetchurl->hostname;
		conn_port = fetchurl->port;
		g_free(query);
	}

	g_return_if_fail (firstgetline!=NULL);

	// make data structure for callback routine (we need that because we can only hand over one user data value
	data = g_new0 (struct http_get_callback_data, 1);
	g_return_if_fail (data!=NULL);
	data->getstring = g_strdup_printf (N_("%s"
				"Accept: text/html\r\n"
//				"Accept-Encoding: x-compress; x-zip; x-gzip\r\n"
				"Accept-Language: en-us\r\n"
				"User-Agent: Mozilla/4.0 (compatible; ed2k_gui; %s)\r\n"
				"Host: %s\r\n"
				"Connection: close\r\n\r\n"), firstgetline, OP_SYSTEM_STRING, fetchurl->hostname);
	g_return_if_fail (data->getstring!=NULL);
	data->buf = NULL;
	data->bufsize = 0;
	data->contentlength = -1;
	data->headerlength = -1;
	data->clearserverlist = (option_clear_slist_when_fetching_new_ones==1 && strcmp(url, option_serverlist_URL)==0);
	data->using_httpproxy = (proxyurl!=NULL);
	data->redirections = redirections;

	// Now make connection, and connect if successful (the callback will handle the rest then)
	// XXX - this may block?! (for the DNS resolution)
	conn = gnet_conn_new (conn_host, conn_port, http_get_connection_callback, (gpointer) data);
	if (!conn)
	{
		status_msg (_("GUI: couldn't resolve hostname '%s'\n"), (proxyurl!=NULL) ? proxyurl->hostname : fetchurl->hostname);
	} else gnet_conn_connect (conn, HTTP_GET_TIMEOUT);

	if (fetchurl) gnet_url_delete (fetchurl);
	if (proxyurl) gnet_url_delete (proxyurl);
	g_free (firstgetline);
	// string in 'get' will be freed by connection callback function
}



// http_get_connection_callback
//
// This is the connection callback to the http/proxy server that does all the
//	reading/writing work

static gboolean
http_get_connection_callback (GConn *conn, GConnStatus status, gchar *buffer, gint length, gpointer datastruct)
{
	struct http_get_callback_data	*data = (struct http_get_callback_data*) datastruct;

	g_return_val_if_fail (data!=NULL, FALSE);

	switch (status)
	{
		case GNET_CONN_STATUS_CONNECT:
		{
			status_msg (_("GUI: connected to http%s server\n"), (data->using_httpproxy) ? _(" proxy"): N_(""));
			// we are connected, let's send our GET request and read the answer
			g_return_val_if_fail (data->getstring!=NULL, FALSE);
//			g_print ("GETSTRING:\n%s\n", data->getstring);
			gnet_conn_write (conn, data->getstring, strlen(data->getstring), HTTP_GET_TIMEOUT);
			gnet_conn_readany (conn, data->readbuf, sizeof(data->readbuf), HTTP_GET_TIMEOUT);
		}
		break;

		case GNET_CONN_STATUS_CLOSE:
		{
			status_msg (_("GUI: read serverlist. Connection closed.\n"));
			http_get_process_serverlist_page (data->buf, data->bufsize, data->clearserverlist, data->redirections);
			http_get_connection_cleanup (conn, data);
			return FALSE;	// we're done.
		}
		break;

		case GNET_CONN_STATUS_READ:
		{
			if (length>0)
			{
				data->buf = g_realloc (data->buf, data->bufsize+length+1);
				g_return_val_if_fail (data->buf!=NULL, FALSE);
				g_memmove (data->buf+data->bufsize, buffer, length);
				data->buf[data->bufsize+length]=0x00;		// remember: we allocated +1 for terminating 0
				data->bufsize += length;
//				g_print ("read %u bytes (total = %8u)\n", length, data->bufsize);

				// check for Content-Length: header line
				if (data->contentlength < 0)
				{
					static char clstr[] = N_("Content-Length:");
					// we need to search the whole buffer, because the string could be
					//	split across the border of two chunks read
					gchar	*conlen = strstr(data->buf, clstr);
					if (conlen)
					{
						gchar	*termchar = NULL;
						data->contentlength = strtol (conlen+sizeof(clstr), &termchar, 10);
						// in this unlikely case the number has been split across two chunks
						if ((termchar) && *termchar!='\n' && *termchar!='\r')
						{
							data->contentlength = -1;
						} //else g_print ("content-length = %u *****\n", data->contentlength);
					}
					
				}

				if (data->headerlength < 0)
				{
					guint	 endlen = 4;
					gchar	*headerend = strstr(data->buf, N_("\r\n\r\n"));
					if (!headerend)
					{
						endlen = 2;
						headerend = strstr(data->buf, N_("\n\n"));
						if (!headerend) headerend = strstr(data->buf, N_("\r\r"));
					}
					if (headerend)
					{
						data->headerlength = headerend+endlen-data->buf;
					}
				}

				if (data->contentlength > 0  &&  data->headerlength > 0
					&&  data->bufsize >= (data->contentlength + data->headerlength))
				{
					status_msg (_("GUI: read serverlist. Closing connection.\n"));
//					g_print ("---------->\n%s\n<------------\n", data->buf);
					http_get_process_serverlist_page (data->buf, data->bufsize, data->clearserverlist, data->redirections);
					http_get_connection_cleanup (conn, data);
					return FALSE;	// we're done.
				}
			}
		}
		return TRUE;

		case GNET_CONN_STATUS_WRITE:
		{
			g_return_val_if_fail (data->getstring!=NULL, FALSE);
			g_free(buffer);		// frees the getstring (buffershould be = data->getstring) after it's been sent
			data->getstring = NULL;
		}
		break;

		case GNET_CONN_STATUS_TIMEOUT:
		{
			status_msg (_("GUI: Timeout on http connection.\n"));
			http_get_connection_cleanup (conn, data);
		}
		break;

		case GNET_CONN_STATUS_ERROR:
		{
			const gchar *errconnstr = _("GUI: Error on http connection.");
			if (errno>0)
			{
				status_msg (N_("%s\n"), errconnstr);
			} else gui_print_error_function (errconnstr);
			http_get_connection_cleanup (conn, data);
		}
		break;
	}
	return FALSE;
}

// http_get_connection_cleanup
//
// deletes the connection and buffers etc.

static void
http_get_connection_cleanup (GConn *conn, struct http_get_callback_data *data)
{
	if (conn)
	{
		gnet_conn_disconnect (conn, TRUE);
		gnet_conn_delete (conn, TRUE);
	}
	if (data)
	{
		if (data->getstring) g_free (data->getstring);
		if (data->buf) g_free (data->buf);
		g_free (data);
	}
}


// http_get_process_serverlist_page
//
// Extracts the ed2k-servers from the page, and displays success/failure messages

static void
http_get_process_serverlist_page (gchar *page, guint bufsize, gboolean clearlist, gint redirections)
{
	guint	 num;
	gchar	*data; 

	g_return_if_fail (page!=NULL);	// we should at least have headers!

	/* first check whether it's a redirection (server response 302) */

	if (1==1)
	{
		guint	 protomajor, protominor, code = 0;
		gchar	 explanation[512];
		gchar	*cr;

		if (sscanf(page,N_("HTTP/%u.%u %u %512s"), &protomajor, &protominor, &code, explanation)>=3)
		{
			/* redirection? */
			if (code == 302)
			{
				gchar *location = strstr(page,N_("Location: "));
				if (location)
				{
					gchar *newurl = location + strlen(N_("Location: "));
					cr = strchr(newurl,'\r');
					if (!cr) cr = strchr(newurl,'\n');
					if (cr) *cr = 0x00;
					status_message_blue (_("GUI: Redirected to new location '%s'.\n"), newurl);
					http_get_retrieve_serverlist (newurl, redirections+1);
					return;
				}
				status_warning (_("GUI: Redirected, but no new location specified?!\n"));
				return;
			}
		}
	}

	data = http_get_beginning_of_data (page);
	if ((data) && *data == 0x0e)
	{
		status_message_blue (N_("%s\n"), _("GUI: This appears to be a server.met file."));
		num = extract_server_links_from_metfile (data, bufsize-(data-page), clearlist);
	} else num = extract_server_links(page, clearlist); 

	if (num == 0)
	{
		gchar	*cr = strchr(page,'\r');
		if (!cr) cr = strchr(page,'\n');
		if (cr)
		{
			cr[1]=0x00;	/* we only want the first line of the server response + new line character */
			cr[0]='\n';	/* we don't want a '\r' */
		}
		status_msg (_("GUI: No ed2k://|server| links found on page :( Server returned this: %s\n"), page);
	} else send_get_serverlist();
}



/* http_get_beginning_of_data
 *
 * Takes a page received via http including headers and skips the headers,
 * pointing to the first byte of data
 *
 */

static gchar *
http_get_beginning_of_data (const gchar *page)
{
	gchar *data;

	g_return_val_if_fail (page!=NULL, NULL);

	data = strstr (page, "\r\n\r\n");
	if (data)
		data += 4;

	return data;
}


/* extract_server_links_from_metfile
 *
 * parses a server.met file
 *
 * returns number of added servers
 *
 */

static gint
extract_server_links_from_metfile (const gchar *data, guint datasize, gboolean clearlist)
{
	gint		 num_servers, num;
	gint		 num_added=0, alreadyknown=0, invalid=0;
	guint8		 header;
	ramfile		*rf;

	g_return_val_if_fail (data!=NULL, 0);
	g_return_val_if_fail (datasize<(1024*1024), 0);	/* a server.met larger than 1MB? */

	rf = ramfile_new (datasize);
	g_return_val_if_fail (rf!=NULL, 0);

	ramfile_write (rf, (guint8*) data, datasize*sizeof(gchar));
	ramfile_set_pos (rf,0);
	header = ramfile_read8(rf);

	if (header != 0x0e)
	{
		status_warning (_("GUI: server.met file: wrong header?!\n"));
		ramfile_free(rf);
		return 0;
	}

	num_servers = ramfile_read32(rf);
	num = num_servers;

	if (num_servers == 0xffffffff)
	{
		status_warning (_("GUI: server.met file: premature end of file?!\n"));
		ramfile_free(rf);
		return 0;
	}

	if (num_servers == 0)
	{
		status_warning (_("GUI: server.met file: claims to have 0 servers?!\n"));
		ramfile_free(rf);
		return 0;
	}

	if (num_servers > 2000)
	{
		status_warning (_("GUI: server.met file: claims to have more than 2000 servers?! (%u)\n"), num_servers);
		ramfile_free(rf);
		return 0;
	}

	while (num_servers > 0)
	{
		guint32		 ip = ramfile_readIP(rf);
		guint16 	 port = ramfile_read16(rf);
		guint8		 n1,n2,n3,n4;
		GPtrArray	*taglist;
		gboolean	 add;
		const gchar	*offender;

		add = TRUE;
		if (ip == 0xffffffff || port == 0xffff)
		{
			status_warning (_("GUI: server.met file: premature end of file?!\n"));
			ramfile_free(rf);
			return num_added;
		}

		n4 = *((guint8*)&ip);
		n3 = *(((guint8*)&ip)+1);
		n2 = *(((guint8*)&ip)+2);
		n1 = *(((guint8*)&ip)+3);

		if (misc_ip_is_local(ip))
		{
			status_message_blue (_("GUI: server.met file: skipped useless server with private IP %u.%u.%u.%u.\n"), n4,n3,n2,n1);
			add = FALSE;
		}

		if (http_get_ip_is_evil (n1, n2, n3, n4, &offender))
		{
			status_message_blue (_("GUI: server.met file: skipped suspicious IP from '%s'.\n"), offender);
			add = FALSE;
		}

/*		g_print ("got IP %8x:%u\n", ip, port); */

		taglist = taglist_read_from_ramfile (rf);
		if (!taglist)
		{
			status_warning (_("GUI: server.met file: premature end of file?!\n"));
			ramfile_free(rf);
			return num_added;
		}

		if (add)
		{
			if (!servers_find_from_address (ip,port))
			{
				NGS_tag	*priority = taglist_find_tag_from_tagname (taglist, TAGNAME_PREFERENCE_SHORT);

				/* get rid of all servers if wanted */
				if ((clearlist)&& num==0)
				{
					serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_ALL_FLAG));
				}

				send_add_server (ip, port);
				num_added++;

				if ((priority) && GPOINTER_TO_UINT(priority->data)!=0)
				{
					send_set_server_priority (ip, port, GPOINTER_TO_UINT(priority->data));
				}
			} else alreadyknown++;
		} else invalid++;

		taglist_free(taglist);
		taglist = NULL;

		num_servers--;
	}

	status_msg (_("Fetched %u servers. %u were new. %u links were invalid. Total now: %u servers\n"),
				 num, num_added, invalid, (servers!=NULL) ? (servers->len)+num_added : num_added);

	ramfile_free(rf);

	return num;
}




/* http_get_ip_is_evil
 *
 * returns TRUE if we do not want to contact the given IP (i.e. p2p sniffer companies etc.)
 *
 * IP address is n4.n3.n2.n1
 *
 * if return value is true, 'offender' will be set to name of offender
 */

static gboolean
http_get_ip_is_evil (guint8 n1, guint8 n2, guint8 n3, guint8 n4, const gchar **offender)
{

	// BayTSP (p2p monitoring)
	// 209.204.128.*				0x**80ccd1
	// 209.204.191.*				0x**bfccd1
	// 216.49.0.0 - 216.49.63.255
	if ((n4==209 && n3==204 && (n2==128 || n2==191))
		|| (n4==216 && n3==49 && n2<64))
	{
		if (offender)
			*offender = N_("BayTSP (p2p monitoring)");
		return TRUE;
	}

	// Cyveillance (P2P monitoring)
	// 63.64.0.*					0x**00403f
	// 63.127.255.*					0x**ff7f3f
	// 63.148.99.*					0x**63943f
	// 65.118.41.*					0x**297641

	if ((n4==63 && ((n3==64 && n2==0) || (n3==127&&n2==255) || (n3==148&&n2==99)))
		|| (n4==65 && n3==118 && n2==41))
	{
		if (offender)
			*offender = N_("Cyveillance (p2p monitoring)");
		return TRUE;
	}

	// MediaDefender (P2P fakes/spam/monitoring)
	// 64.225.202.*					0x**cae140
	// 66.79.0.*					0x**004f42
	// 66.79.10.*					0x**0a4f42
	// 66.79.95.*					0x**5f4f42

	if ((n4==64 && n3==225 && n2==202) || (n4==66 && n3==79 && (n2==0 || n2==10 || n2==95)))
	{
		if (offender)
			*offender = N_("MediaDefender (p2p fakes/spam/monitoring)");
		return TRUE;
	}

	// Media Force (P2P monitoring)
	// 4.43.96.*					0x**602b04
	// 65.192.0.*					0x**00c041
	// 65.217.219.*					0x**dbd941
	// 65.223.255.*					0x**ffdf41

	if ((n4==4 && n3==43 && n2==96) || (n4==65 && n3==192 && n2==0)
		|| (n4==65 && n3==217 && n2==219) || (n4==65 && n3==223 && n2==255))
	{
		if (offender)
			*offender = N_("Media Force  (p2p monitoring)");
		return TRUE;
	}

	return FALSE;
}
