/*
 * main.c
 *
 * ed2k_gui - a nice gui for the donkey
 *
 * (c) 2001-2002 Tim-Philipp Muller <t.i.m@orange.net>
 *
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "blacklist.h"
#include "colors.h"
#include "connect_gui.h"
#include "downloads_attrib.h"
#include "dns_lookup.h"
#include "filter.h"
#include "global.h"
#include "gui_flags.h"
#include "gui_options.h"
#include "guicoreconn.h"
#include "http_get.h"	/* for no_of_servers_fetched */
#include "icons.h"
#include "linkhandler.h"
#include "mainwindow.h"
#include "misc.h"
#include "notebook.h"
#include "packet.h"
#include "parse_message.h"
#include "pipe.h"
#include "search_page.h"
#include "send_command.h"
#include "servers_page.h"
#include "shared_page.h"
#include "signal_handler.h"
#include "spawn_core.h"
#include "status_page.h"
#include "statusbar.h"
#include "uploads_page.h"

/* standard stuff */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>

/* global variables */

#if defined(__FreeBSD__) && defined(WITH_MALLOC_DEBUG)
  const gchar *_malloc_options = N_("AJVX");
#endif


/* local variables */

static time_t	 last_emergency_server_fetch = 0;
static gint		 general_update_timer_id = -1;

static gboolean	 use_options_data_for_shutdown = FALSE;
static gchar	*advcmd_to_execute = N_("");

static gchar	*main_ed2k_link_passed_on_commandline = NULL;

/* local functions */
static void		 main_parse_commandline_args (gint argc, gchar **argv);



/* functions */

static void
print_helpscreen (const gchar *extramsg, gint exitcode)
{
	fprintf (stderr, N_("%s\n\n"), extramsg);
	printf (_("linux gui for eDonkey2000 (www.eDonkey2000.com) - build %s %s\n"
		"(c) 2001 Tim-Philipp Muller <tim@edonkey2000.com>\n\n"
		"There are NO command line arguments you HAVE to specifiy.\n\n"
		"Optional command line options:\n"
		"-c\tconnect right away to the core without bringing up the 'connect GUI'\n"
		"\tdialog, using the default host,port,username and password specified on\n"
		"\tthe GUI's options page (and saved in gui_options on program exit).\n"
		"\tThis will only work if there is already a gui_options file.\n"
		"-i1,-i0\t\tThis will determine whether an application icon is set (-i1) or not (-i0)\n"
		"-p path\t\tspecifies the path which shall be used for the gui config files\n"
		"-n num\t\tspecifies 'instance' of GUI if you want to run multiple instances\n"
		"\t\t\tof the GUI at the same time and keep separate config and stats files\n"
		"\t\t\tfor them. e.g.: if num=2, then 'gui_options.2' is used\n"
		"\t\t\t(filter+blacklist files will be shared unless you specify different dirs)\n"
		"-q\t\tbe quiet when doing '-s' or '-a' commands (see below)\n"
		"-s host port username password\t\t just connect and shutdown core\n"
		"-a host port username password cmd\t just connect and send advanced command\n\n"),
		__DATE__, __TIME__);
	exit(exitcode);
}


/******************************************************************************
 *
 *   fill_in_special_commandline_args
 *
 *
 */

static void
fill_in_special_commandline_args (gchar *a0, gint a1, gchar *a2, gchar *a3, gchar *a4)
{
	strrealloccpy (&shutdown_host,  a0);
	shutdown_port = a1;
	strrealloccpy (&shutdown_user,  a2);
	strrealloccpy (&shutdown_pass,  a3);
	strrealloccpy (&shutdown_advcmd,a4);
	if ((!shutdown_host) || (!shutdown_user) || (!shutdown_pass) || (!shutdown_port) || (!shutdown_advcmd))
	{
		fprintf (stderr, _(">>>>> ERROR allocating memory!\n"));
		exit(1);
	}
	return;
}


/******************************************************************************
 *
 *   main_parse_commandline_args
 *
 *
 */

static void
main_parse_commandline_args (gint argc, gchar **argv)
{
	gchar *unrecognised = _("UNRECOGNISED COMMAND-LINE OPTION");
	guint i;

	if (argc<=1) return;	/* any args at all? */

	for (i=1; i<argc; i++)
	{
		char *token=argv[i];
		if (!token) return;
		if (token[0]==0x00) return;
		if (token[0]=='-')
		{
			token++;
			if (token[0]=='-') token++;
			switch (token[0])
			{
				case 'h':	print_helpscreen(N_(""), 0);
						break;

				case 'c':	connect_right_away = 1;
						break;

				case 'q':	only_be_quiet	   = 1;
						break;

				case 's':	/* just shutdown core */
				{
					if (argc>=(1+i+4))
					{
						only_send_shutdown = 1;
						fill_in_special_commandline_args (argv[i+1], atoi(argv[i+2]), argv[i+3], argv[i+4], N_(""));
						if (!only_be_quiet)
						{
							printf (_(">>>>> shutdown core @ %s:%u (%s:%s)\n"),
									shutdown_host, shutdown_port, shutdown_user, shutdown_pass);
						}
						i=i+4;
					} else {
						only_send_shutdown = 1;
						use_options_data_for_shutdown = TRUE;
						if (!only_be_quiet)
							printf (_(">>>>> doing core shutdown (with saved parameters)\n"));
					}
				}
				break;

				case 'a':	/* just send advanced command */
				{
					if (argc>=(1+i+5))
					{
						only_send_advcmd = 1;
						fill_in_special_commandline_args (argv[i+1], atoi(argv[i+2]), argv[i+3], argv[i+4], argv[i+5]);
						if (!only_be_quiet) printf (_(">>>>> send advanced command to core @ %s:%u (%s:%s) - '%s'\n"),
									shutdown_host, shutdown_port, shutdown_user, shutdown_pass, shutdown_advcmd);
						i=i+5;
					} else if (argc>=(1+i+1))
					{
						only_send_advcmd = 1;
						use_options_data_for_shutdown = TRUE;
						advcmd_to_execute = argv[i+1];
						i++;
					} else print_helpscreen(_("The '-a' option requires ONE or FIVE extra parameters"),1);
				}
				break;

				case 'i':	// don't set application icon (for Jaroslav)
				{
					if (token[1]=='0')
					{	// we need to set this != 1/0 so read_gui_options
						// knows that these have been set via command line
						option_no_app_icon = 1;		// no app icon
					} else option_no_app_icon = 0;		// yeah, show icon
				}
				break;
				
				case 'p':	/* specify path where config files are stored */
				{
					if (token[1]==0x00)	/* -p /home/joe/ed2k/ variant */
					{
						if (argc>=(1+i+1))
						{
							gchar real[PATH_MAX+NAME_MAX+2];
							if (realpath (argv[i+1], real))
							{
								if (prefpath)
									g_free(prefpath);
								prefpath = g_strdup(real);
							}
							i++;
						} else print_helpscreen (_("The '-p' option requires an extra parameter"),1);
					} else {
						if (prefpath)
							g_free(prefpath);
						prefpath = g_strdup(token+1);
					}
				}
				if (prefpath)
				{
					printf (_("Path for gui configuration files: %s\n"), prefpath);
					if (access(prefpath,F_OK|W_OK)!=0)
					{
						print_helpscreen (_("The path supplied doesn't exist or we can't write to it."),1);
						g_free(prefpath);
						prefpath=NULL;
					}
				}
				break;

				case 'n':	/* specify 'instance' of GUI */
				{
					if (token[1]==0x00)	// case '-n 2'
					{
						if (argc>=(1+i+1))
						{
							option_gui_instance = atoi(argv[i+1]);
							i++;
						} else print_helpscreen (_("The '-n' option requires an extra parameter"),1);
					} else option_gui_instance = atoi(token+1);
				}
				break;
				default:	fprintf (stderr, N_("---> '%s' <--- "), token);
						print_helpscreen(unrecognised,1);
			}
		} else {
			if (g_strncasecmp(token,N_("ed2k:"),5)!=0)
			{
				fprintf (stderr, N_("---> '%s' <--- "), token);
				print_helpscreen(unrecognised,1);
			} else main_ed2k_link_passed_on_commandline = g_strdup(token);
		}
	}
}


/******************************************************************************
 *
 *   find and init pref path
 *
 *   checks whether ~/.ed2k_gui/ exists, if not creates it
 *   should only be used if path has not been specified via command line
 * 
 *   returns -1 on error, 1 if successful
 */

static gint
find_and_init_pref_path (void)
{
	gchar *homedir;

	if (prefpath) /* already initialised or specified via command line */
		return 1;

	homedir = g_get_home_dir();
	g_return_val_if_fail (homedir!=NULL, -1);

	prefpath = g_strdup_printf (N_("%s%s.ed2k_gui%c"),
					homedir,
					(homedir[strlen(homedir)-1]==SLASH_CHAR) ? N_("") : SLASH_STRING,
					SLASH_CHAR);

	g_return_val_if_fail (prefpath!=NULL, -1);

	/* does path   $HOME/.ed2k_gui/  not exist? */
	if (access(prefpath, F_OK)!=0)
	{
		/* ~/.ed2k_gui/ does NOT exists => create it */
		if (mkdir(prefpath,S_IRWXU)!=0)
		{
			g_printerr(_("--- couldn't create directory '%s' for GUI configuration files! (using home dir)\n"), prefpath);
			g_free(prefpath);
			prefpath = homedir;
			return -1;
		} /* else: mkdir() successful -   ~/.ed2k_gui/ exists now :) */
	} else {
		/*   ~/.ed2k_gui/ exists, but do we have write permission as well? */
		if (access(prefpath,W_OK)!=0)
		{
			/* no => try to change permissions */
			if (chmod(prefpath, S_IRWXU)!=0)
			{
				g_printerr (_("--- couldn't set write permissions for GUI config file directory '%s' (using home dir)\n"), prefpath);
				g_free(prefpath);
				prefpath = homedir;
				return -1;
			} /* else: changed permissions successfully */
		} /* else: yup, we have write permissions */
	}
	return 1;
}



/******************************************************************************
 *
 *   general_update_timer
 *
 *   do stuff that needs to be done every x minutes or so
 */

gint
general_update_timer (gpointer data)
{
	static int 	minutes = 0;

	/* every minute */
	if (current_page == NOTEBOOK_PAGE_DOWNLOADS)
		print_download_stats_line ();

	if (current_page == NOTEBOOK_PAGE_UPLOADS)
		print_upload_stats_line ();

	if (CORE_IS_ALIVE)
		send_get_peer_stats ();

	downloadsattriblist_update_from_downloadlist ();

	if (downloads_get_num()<50)
		update_space_stat = 1;	// update after next status packet

	/* every 5 minutes */
	if (minutes%5 == 0)
	{
		update_space_stat = 1;	/* update after next status packet */

		/* we don't want to totally run out of servers, do we? */
		if ( option_fetch_when_too_few_servers==1
			&& (!IS_OVERNET)
			&& ((!servers) || ((servers) && servers->len<10))
			&& strcmp(option_nickname, DUMMY_NICKNAME)!=0)
		{
			if (CORE_IS_ALIVE)
			{
				static guint	whichlist = 0;
				whichlist++;
				if (whichlist==5) whichlist=1;
				/* if fetch unsuccessful, we don't want to re-fetch
				 * unsuccessfully every 5 minutes, do we?
				 *
				 *   max. is every 30 minutes
				 */
				if (last_emergency_server_fetch==0 || (time(NULL)-last_emergency_server_fetch)>1800)
				{
					last_emergency_server_fetch = time(NULL);
					status_message_blue (_("GUI: less than 10 servers in serverlist - trying to fetch more via http.\n"));
					serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(SERVERS_MENUITEM_FETCH_SERVERLIST|(whichlist<<8)));
				}
			}
		}

		if ((CORE_IS_ALIVE) && (!IS_OVERNET))
		{
			if (option_autoremove==1)	/* in case core is too lazy */
			{
				serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_DEAD_FLAG));
			}

			if (option_autoremoveinvalid==1)
			{
				serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_LOCAL_FLAG));
			}

			send_get_serverlist();
			/* make sure we're connected to a server
			 * when 'always stay connected' option is set
			 */
/*			The core will handle that

			if (option_always_connected == 1 && current_server_users == 0)
			{
				send_connect_to_server (0,0);	// connect to any server
			}
*/		}

		mainwindow_retrieve_some_options_before_main_window_is_destroyed();
		write_gui_options();
		write_gui_uploadstats();
		downloadsattriblist_save();
	
	}

	/* every 15 minutes */
	if (minutes%15 == 0)
	{
		sync_ulstats_filenames ();
	}

	/* every 30 minutes */
	if (minutes%30 == 0 && CORE_IS_ALIVE)
	{
		send_get_shared_dirs();
		if (option_slowconnection==0)	// if slow do it only every 90 mins.
			send_get_shared_files();
	}

	/* every 45 minutes */
	if (minutes%45 == 0)
	{
		;
	}

	if (minutes%90 == 0)
	{
		if (option_slowconnection==1)
		{
			send_get_shared_files();
		}
	}

	/* server rotate feature */
	if ((option_server_rotate>0) && (!IS_OVERNET))
	{
		if (minutes>0 && minutes>=option_server_rotate
			 && (minutes%option_server_rotate)==0 && CORE_IS_ALIVE)
		{
			/* will set current server to low priority */
			servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_DO_SERVER_ROTATE);
		}
	}

	/* clear status log if user has set the option accordingly */
	if (option_clearstatuswindow>0 && minutes>=option_clearstatuswindow && minutes>0
		 && (minutes%option_clearstatuswindow)==0 && CORE_IS_ALIVE)
	{
		status_page_clear();
		status_msg (_("GUI: cleared status window. You will find all previous output in the gui_statuslog file.\n"));
	}


	/* workaround for core bug - core won't always recognise if it gets disconnected from server. We have to send
	 *  a search so the core can check if the socket is still open. Now, if the user hasn't clicked anything for
	 *  20 minutes, we'll just send some search to make sure we don't stay disconnected + w/o download all night...
	 */
	if ((!IS_OVERNET) && ((int)(time(NULL)-last_time_of_user_click) > (30*60)))
	{
		send_search(N_("plbrvqszt"));
		last_time_of_user_click = time(NULL);
	}

	/* check if the blacklistfile has been changed by hand */
	check_if_blacklistfile_changed ();
	check_if_filterlistfile_changed ();

	minutes++;
	return(1);	/* call us in the future */
}


/******************************************************************************
 *
 *   main_clean_up_connections_and_callbacks_before_quit
 *
 *   Remove timeouts and send some stuff to core before we go away.
 *   Now that all this stuff is done via glib/gnet, we need to do it
 *    before we call gtk_main_quit()
 *
 */

void
main_clean_up_connections_and_callbacks_before_quit (void)
{
	time_t	now;

	gtk_timeout_remove(general_update_timer_id);

	write_to_logfile (_("shutting down GUI: tidying up...\n"));

	/* will not be sent if core is not alive */
	send_stop_sending_dl_status();          
	send_stop_sending_ul_status();

	if (option_shutdown_core_on_exit == 1)
	{
		dns_lookup_remove_dns_servers ();
		send_logout();
	}

	/* We do this so core tries to send sth after we closed the
	 * socket and recognises that we're gone (Another Control
	 * Attempt problem on second connect...)
	 *
	 */
	send_get_shared_files();

	/* make sure our stuff gets sent out (but wait max. 1 second) */
	now = time(NULL);
	while ((gtk_events_pending()) && (time(NULL)-now)<1)
		(void) gtk_main_iteration();

	guicoreconn_close_connection ();
}


/******************************************************************************
 *
 *   main_send_shutdown_or_advanced_command_to_core
 *
 *   Only sends a shutdown message or an advanced
 *   command message to the core, nothing else.
 *
 *   Does not return.
 *
 */

static void
main_send_shutdown_or_advanced_command_to_core (void)
{
	if (!prefpath)
	{
	  find_and_init_pref_path();
	} /* else: error occurred and prefpath = $HOME */

	g_assert (prefpath!=NULL);

	read_gui_options();

	if (use_options_data_for_shutdown==1)
	{
		/* XXX - TODO: check if default options are set?! */
		fill_in_special_commandline_args(option_defaulthost,
						 option_defaultport,
						 option_defaultuser,
						 option_defaultpass,
						 advcmd_to_execute);

		if (!only_be_quiet)
		{
			if (only_send_shutdown)
			{
				printf (_(">>>>> shutdown core @ %s:%u (%s:%s)\n"),
					shutdown_host, shutdown_port,
					shutdown_user, shutdown_pass);
			}

			if (only_send_advcmd)
			{
				printf (_(">>>>> send advanced command to core @ %s:%u (%s:%s) - '%s'\n"),
					shutdown_host, shutdown_port,
					shutdown_user, shutdown_pass,
					shutdown_advcmd);
			}
		}
	}

	guicoreconn_do_shutdown_connection ();
	g_assert_not_reached();
}



/******************************************************************************
 * 
 *   main
 *
 */

int
main (int argc, char *argv[])
{
	/* just being hyper-careful... */
	g_return_val_if_fail (sizeof(gfloat)==4, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(gint  )==4, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(guint )==4, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(gchar )==1, EXIT_FAILURE);
	g_return_val_if_fail (sizeof(gshort)==2, EXIT_FAILURE);

	/* set our own error handler, so we can use g_printerr() our way */
	g_set_printerr_handler ((GPrintFunc)gui_print_error_function);

	/* init gettext:
	 *
	 * Set locale according to environment variables
	 * Set $LANG before executing the GUI to set your preferred language,
	 *   e.g. with "export LANG=fr" (sh)(if it is not already set that is)
	 *   
	 */
	(void) setlocale (LC_ALL, "");
	(void) bindtextdomain (PACKAGE, LOCALEPATH);
	(void) textdomain (PACKAGE);

	/* signal handling
	 *
	 * by Mario Sergio Fujikawa Ferreira <lioux@FreeBSD.org>
	 *
	 * XXX - [Tim:] Does SIGCHLD interfere with libgnet in any way?
	 * XXX - [Tim:] should we block SIGTRAP? (I get this sometimes when doing http fetch while running GUI in gdb)
	 *
	 */
	signal(SIGCHLD, sig_chld_handler);       	/* Prevents zombies */
	signal(SIGPIPE, sig_pipe_ignore_handler);	/* Handles post mortem conversations */
	/*sig_pipe_block(); */                   	/* Blocks signals from post morten conversations */

	/* GTK initialization (strips GTK commandline args if there are any) */
	(void) gtk_set_locale();
	gtk_init (&argc, &argv);

	main_parse_commandline_args(argc, argv);

	if (only_send_shutdown==1 || only_send_advcmd==1)
	{
		main_send_shutdown_or_advanced_command_to_core();
		g_assert_not_reached();
	}

	/* GNet initialization:
	 * from gnet library version 1.1.5 onwards we need to call gnet_init() from main()
	 * before we use the library (we need to call gnet_init() _before_ we call gtk_init()).
	 */
#if ( GNET_MAJOR_VERSION>1 \
		|| (GNET_MAJOR_VERSION==1 && GNET_MINOR_VERSION>1) \
		|| (GNET_MAJOR_VERSION==1 && GNET_MINOR_VERSION==1 && GNET_MICRO_VERSION>=5) )
	gnet_init ();
#endif

#if (GTK_MAJOR_VERSION==1 && GTK_MINOR_VERSION<=2)
	if (getenv(N_("ED2K_GUI_FONT")))
	{
        GtkStyle *defstyle = gtk_widget_get_default_style();
		if (defstyle)
		{
			GdkFont	 *font_gui = gdk_font_load (getenv(N_("ED2K_GUI_FONT")));
			if (font_gui)
			{
				defstyle->font = font_gui;
				gtk_widget_set_default_style(defstyle);
			} else fprintf (stderr, _("**** GUI: could not load font '%s' (as set in ED2K_GUI_FONT).\n"), getenv(N_("ED2K_GUI_FONT")));
		} 
	} 
#endif

	create_default_styles ();	// create some styles

	/* XXX: why do we need the cleanup_after_disconnect here? */
	cleanup_after_disconnect();

	if (!prefpath)	// meaning: it has not been set via the command line
	{
		(void) find_and_init_pref_path();
		g_assert (prefpath != NULL);
	}

	read_gui_options();
	uploadstats_read_from_file();
	downloadsattriblist_load();

	/* if we have been passed an ed2k-link as argument,
	 *  the linkhandler window might take over and not
	 *  the normal main GUI window.
	 */
	if (linkhandler_main(main_ed2k_link_passed_on_commandline))
	{
		g_free(main_ed2k_link_passed_on_commandline);
		exit(0);
	}
	
	write_to_logfile (N_("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv\n"));

/* XXX - if the freebsd pipe bug is fixed, remove error stuff in linkhandler.c */
#ifndef FREEBSD
	(void) pipe_create();           	/* create named pipe   */
#else
	fprintf (stderr, N_("*****\n"));
	fprintf (stderr, N_("***** NOTE: Disabled pipe functionality, as this is a FreeBSD system,\n"));
	fprintf (stderr, N_("*****       and the pipe code does not work with FreeBSD for some reason.\n"));
	fprintf (stderr, N_("*****\n"));
	fprintf (stderr, N_("*****       See here for a bug description:\n"));
	fprintf (stderr, N_("*****       http://sourceforge.net/tracker/index.php?func=detail&aid=656923&group_id=61752&atid=498271\n"));
	fprintf (stderr, N_("*****\n"));
	fprintf (stderr, N_("*****       If you are a developer, please have a look at the GIOChannel\n"));
	fprintf (stderr, N_("*****       callback in pipe.c and try to fix this bug.\n"));
	fprintf (stderr, N_("*****\n"));
#endif

	read_filters();                   	/* read search filters */
	check_if_blacklistfile_changed ();	/* read blacklist file */

	// add timer that takes care of stuff that needs to be done regularly but is
	// not important. called every minute
	general_update_timer_id = gtk_timeout_add (1000*60, general_update_timer, NULL);

	init_xpms();

	/* create main window */
	mainwindow_create();

	toolbar_set_not_connected_to_core();
	search_page_set_initial_extra_search_options_visibility();

	init_pixmaps();

	// if GUI option is set so, try to spawn a local donkey if it is not already running

	if (option_spawn_core_on_startup==1)
	{
		if ( donkey_process_running_check() <=0)
		{
			status_msg (_("GUI: auto-spawn: no local core running, we need to spawn one.\n"));
			(void) spawn_core_and_print_spawn_result_message(option_defaultdonkeypath);
		} else status_msg (_("GUI: auto-spawn: local core already running.\n"));
	}

	// Auto-connect to core on start-up if GUI option is set so
	if (1==1)
	{
		gboolean show_connect_dialog = TRUE;
		if (option_connect_to_core_on_startup == 1  ||  connect_right_away == 1 )
		{
			// connect_right_away is used in the case where we want to override
			//	the auto-connect on startup option but not change it
			//	(ie. through command line command)
			//	[used to be in connect_gui.c/cleanup_after_disconnect()]
			connect_right_away = 0;
			if (guicoreconn_set_connection (option_defaulthost, option_defaultport, option_defaultuser, option_defaultpass, NULL))
			{
				if (!guicoreconn_connect ())
				{
					g_printerr (_("Auto-connect on startup failed - connection structure hasn't been set up?!\n"));
				} else show_connect_dialog = FALSE;
			} else g_printerr (_("Auto-connect on startup failed - couldn't initiate valid connection structure with default values.\n"));
		}
		if (show_connect_dialog)
		{
			gui_connect_dialog ();
		}
	}

	gtk_main ();

	/* just making sure, so our printhandler stops outputting to the status widget */
	if (window)
		window = NULL;	

	// note: we saved the window options (position, width etc) with main win delete event
	write_gui_options();
	write_gui_uploadstats();
	downloadsattriblist_save();
	write_filters();
	free_filters ();
	cleanup_after_disconnect();

	write_to_logfile (N_("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"));

	G_FREE(prefpath);
	G_FREE(main_ed2k_link_passed_on_commandline);

#ifndef FREEBSD
	pipe_unlink();
#endif

	return(0);
}
