//
//	misc.c
//
//	contains miscellaneous functions (e.g. for dialog pop-up windows)
//
//  (c) 2001-2002 Tim-Philipp Muller <t.i.m@orange.net>
//
//

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "global.h"
#include "mainwindow.h"
#include "misc.h"
#include "misc_gtk.h"
#include "connect_gui.h"
#include "gui_options.h"
#include "statusbar.h"
#include "status_page.h"
#include "icons.h"
#include "uploads_page.h"
#include "downloads_page.h"
#include "gui_flags.h"
#include "guicoreconn.h"

#include <sys/param.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <libgen.h>	/* why do we need this header file? */
#include <limits.h>
#include <netdb.h>
#include <signal.h>	// for kill(pid,sig);
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

//gfloat		vvalue, hvalue;		/* save horizontal + vertical position of GtkAdjustments of scrolled window here */
					/* so when we do gtk_clist_clear + create new on update in filelist.c it stays   */
					/* at same position as it was before */


// functions

static GSList	*misc_get_directories_in_path (gboolean withcurdir, gboolean withhome);



static float
convert_to_MB (float n, char unit)
{
	switch (unit)
	{
		case 'G':	n = n*1024.0;
				break;
		case 'k':	n = n/1024.0;
				break;
		case 'b':	n = n/(1024.0*1024.0);
		case 'M':
		default:	;
	}
	return n;
}

gint compare_file_size_str (char *text1, char *text2)
{
	char  *dummy1, *dummy2;
	double n1, n2;

	if (text1[0]!='?')
	{
#ifdef HAVE_STRTOD
		n1 = strtod (text1, &dummy1);
#else
		n1 = atof(text1);
		dummy1 = text1 + strlen(text1) - 1;
#endif
		n1 = convert_to_MB (n1, *dummy1);
	} else n1 = 0;

	if (text2[0]!='?') {
#ifdef HAVE_STRTOD
		n2 = strtod (text2, &dummy2);
#else
		n2 = atof(text2);
		dummy2 = text2 + strlen(text2) - 1;
#endif
		n2 = convert_to_MB (n2, *dummy2);
	} else n2 = 0;

	if (n1>=n2)
	{
		return (gint) (n1>n2);
	}
	else return (-1);
}

// hash_to_hash_str
//
// takes 128-bit hash
// returns pointer to hash as hex string

const gchar *
hash_to_hash_str (const gchar *hash)
{
	gint 			j;
	const gchar		hexdigits[] = N_("0123456789abcdef");
	static gchar	m_hash_str[33];

	// initialise and terminate hash string
	memset (m_hash_str, 0x00, sizeof(m_hash_str)); 
	g_return_val_if_fail (hash!=NULL, m_hash_str);

	for (j=0; j<16; j++)
	{
		m_hash_str[(j<<1)  ] = hexdigits[(((hash[j]) & 0xf0) >> 4)];
		m_hash_str[(j<<1)+1] = hexdigits[(((hash[j]) & 0x0f)     )];
	}
	return m_hash_str;
}

const gchar *hash_str_to_hash (const gchar hash_str[35])
{
	char hexnum[3] = { 0x00, 0x00, 0x00 };
	int i;
        static char m_hash[16];

	for (i=0; i<16; i++) {
		strncpy (hexnum, hash_str+(i*2), 2);
		m_hash[i] = (char) strtol (hexnum, NULL, 16);
	}
	return m_hash;
}


// human_size
//
// returns the given filesize in bytes in 'human' format (e.g. '9.32M')

const gchar *
human_size (guint bytes)
{
	gfloat			size = (gfloat) bytes;
	static gchar	buf[128];

	if ( bytes >= 1024 )
	{
		if ( bytes >= (1024*1024))
		{
			if ( bytes >= (1024*1024*1024))
			{
				g_snprintf (buf,sizeof(buf), N_("%4.2fG"), size/(1024.0*1024.0*1024.0));	// Giga
			} else g_snprintf (buf,sizeof(buf), N_("%4.2fM"), size/(1024.0*1024.0)); 	// Mega
		} else g_snprintf (buf,sizeof(buf), N_("%4.0fk"), size/1024.0); // kilo
	} else g_snprintf (buf,sizeof(buf), N_("%db"), bytes); // bytes
	return buf;
}


// int2str
//
// returns the given integer in form of a string (useful for dealing with entry widgets)

const gchar *
int2str (gint i)
{
	static gchar istr[64];

	g_snprintf (istr, sizeof(istr), N_("%d"), i);
	return istr;
}



// misc_get_basename_from_pid
//
// Retrieves the basename of the program running under the given PID.
// If arg1 is not NULL, it will be set to the first command line argument of the prog.
// Returns basename or NULL if there is an error or the PID does not exist.

const gchar *
misc_get_basename_from_pid (pid_t pid, const gchar **arg1)
{
	FILE		*pidcheckfile;
	static gchar	 pidcheckpath [256];
	static gchar	 basefn [NAME_MAX+PATH_MAX+1];
	static gchar	 arg1str [NAME_MAX+1];	// XXX - what's the proper limit here?
	guint		 i=0;
	gchar		 c;

	g_snprintf (pidcheckpath, 64, N_("/proc/%u/cmdline"), pid);
	
	pidcheckfile = fopen (pidcheckpath, N_("r"));
	if (!pidcheckfile) return NULL;

	while (i<NAME_MAX && (c=fgetc(pidcheckfile))!=EOF && c)
		basefn[i++] = c;
	basefn[i++] = 0;
	i = 0;
	// and argv[1]
	while(i<NAME_MAX && (c=fgetc(pidcheckfile))!=EOF && c)
		arg1str[i++] = c;
	arg1str[i++] = 0;
	fclose(pidcheckfile);

	if (arg1) *arg1 = arg1str;
	return g_basename (basefn);
}


/* misc_is_process_is_a_running_donkey_core
 *
 * returns TRUE if /proc/<pid>/cmd shows that this is a donkey in GUI mode.
 *
 */

static gboolean
misc_is_process_is_a_running_donkey_core (const gchar *pidstr, const gchar *userbin, gint userbinlen)
{
	FILE		*fp;
	gchar		*basename = NULL;
	gchar	 	 buf[NAME_MAX];
	gchar		 argbuf[NAME_MAX];
	gboolean	 has_gui_mode_argument = FALSE;
	gint		 f, c;

	g_return_val_if_fail (pidstr != NULL, FALSE);

	/* Now read argv[0] and argv[x] */
	g_snprintf(buf, sizeof(buf), N_("/proc/%s/cmdline"), pidstr);

	fp = fopen(buf, N_("r"));
	if (!fp)
		return FALSE;

	/* read in program binary name = argv[0] */
	f = 0;
	while(f < sizeof(buf) && (c = fgetc(fp)) != EOF && c!=0x00)
		buf[f++] = c;
	buf[f++] = 0;

	/* read in further arguments argv[x] and check whether program
	 * was given a '-' command linen argument (which would put a core
	 * into GUI mode.
	 */
	f=0;
	while (c==0x00 && f<sizeof(argbuf))
	{
		f = 0;
		while(f < sizeof(argbuf) && (c = fgetc(fp)) != EOF && c!=0x00)
			argbuf[f++] = c;
		argbuf[f++] = 0;

		if (argbuf[0]=='-' && argbuf[1]==0x00)
			has_gui_mode_argument = TRUE;
	}

	fclose(fp);
	fp = NULL;

	/* Get a pointer to the basename */
	if ((basename = strrchr(buf, '/')) != NULL)
	{
		basename++;
	} else basename = buf;

	/* check if in gui mode (using the '-' option) */
	if (has_gui_mode_argument)
	{
		if ((userbin) && strncmp(basename, userbin, userbinlen) == 0)
			return TRUE;

		if (strncmp(basename, N_("donkey"), 6)==0)
			return TRUE;

		if (strncmp(basename, N_("edonkey"), 7)==0)
			return TRUE;

		if (strncmp(basename, N_("overnet"), 7)==0)
			return TRUE;

	}

	if (strncmp(basename, N_("lonkey"), 6)==0)
		return TRUE;

	if (strncmp(basename, N_("cdonkey"), 7)==0)
		return TRUE;

	return FALSE;
}


/*	donkey_process_running_check
 *
 * Read the proc filesystem and see if we have a donkey in gui mode (with '-' option) running
 * returns 0 if no donkey found, or PID of donkey found
 *
 */

guint
donkey_process_running_check (void)
{
	DIR				*dir;
	struct dirent	*d;
	gchar			*basename;
	gchar			*userbin = NULL;	/* basename of binary user specified in options file/dialog, if any */
	gint			 userbinlen = 0;
	gint			 pid;

	/* find out if user specified a binary */
	if ((option_defaultdonkeypath) && option_defaultdonkeypath[0]!=0x00)
	{
		if ((userbin=strrchr(option_defaultdonkeypath, '/')))
		{
			userbin++;	/* point to character after slash */
			userbinlen = strlen(userbin);
			if (!userbinlen) userbin=NULL;
		}
	}

	/* Open the /proc directory. */
	if ((dir = opendir(N_("/proc"))) == NULL)
		return -1;										/* error opening dir */

	/* Walk through the directory. */
	while ((d = readdir(dir)) != NULL)
	{
		/* See if this is a process (ie. a number) */
		/* if not, check next entry                */
		if ((pid = atoi(d->d_name)) == 0)
			continue;	

		/* check process: is this a donkey core?   */
		
		if (misc_is_process_is_a_running_donkey_core(d->d_name,userbin,userbinlen))
		{
			closedir(dir);
			return(pid);
		}
	}
	closedir(dir);

	return 0;	/* Done */
}


/* check directory given for a given file.
*  returns complete path if found, otherwise NULL
*/

const gchar *
check_dir_for_specified_binary (const gchar *dir, const gchar *filename)
{
	static char	  fullpath[PATH_MAX+NAME_MAX+2];
	DIR 		* directory;
	struct dirent	* d;

	g_return_val_if_fail (dir!=NULL,NULL);
	g_return_val_if_fail (filename!=NULL,NULL);

	if ((directory = opendir(dir)) == NULL) return NULL;	// open directory...

	while ((d = readdir(directory)) != NULL)			// Walk through the directory.
	{
		if (strcmp(d->d_name, filename)==0)			// check if directory entry matches filename
		{
			struct stat filestat;
			g_snprintf (fullpath, sizeof(fullpath), N_("%s%c%s"), dir, SLASH_CHAR, d->d_name);
			if ( stat (fullpath, &filestat) != 0 ) continue;
			if (S_ISDIR(filestat.st_mode)) continue;		// is it a directory? => not useful
			if (filestat.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH))	// is it executable?
			{
				closedir(directory);
				return fullpath;
			}
		}
	}
	closedir(directory);
	return NULL;
}

/* find_specified_binary_in_path
 *
 * checks PATH for a specific binary and returns full path to binary if found or NULL
 *
 */

static const gchar *
find_specified_binary_in_path (const gchar *binaryfn)
{
	GSList		*dirlist;
	const gchar	*ret = NULL;

	g_return_val_if_fail (binaryfn!=NULL,NULL);

	dirlist = misc_get_directories_in_path (TRUE,TRUE);
	PROCESS_GSLIST (dirlist, gchar,
		if (!ret)
		{
			ret = check_dir_for_specified_binary(item, binaryfn);
		}
		g_free(item);
	);
	g_slist_free(dirlist);

	return ret;
}

/* check directory given for file named 'donkey' and 'donkey_s'.
*  returns complete path if found, otherwise NULL
*/
char *check_dir_for_binary (char *dir)
{
	static char	  fullpath[1025];
	DIR 		* directory;
	struct dirent	* d;

	if (!dir) return NULL;
	if ((directory = opendir(dir)) == NULL) return NULL;	// open directory...

	while ((d = readdir(directory)) != NULL)			// Walk through the directory.
	{
		/* check if directory entry starts with 'donkey' or 'lonkey' */
		if (strncmp(d->d_name, N_("donkey"), 6)==0
			|| strncmp(d->d_name, N_("lonkey"), 6)==0)
		{
			struct stat filestat;
			g_snprintf (fullpath, 1024, N_("%s%s%s"), dir, (dir[strlen(dir)-1]==SLASH_CHAR) ? N_("") : SLASH_STRING, d->d_name);
			if ( stat (fullpath, &filestat) != 0 ) continue;
			if (S_ISDIR(filestat.st_mode)) continue;		// is it a directory? => not useful
			if (filestat.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH))	// is it executable?
			{
				closedir(directory);
				return fullpath;
			}
		}
	}
	closedir(directory);
	return NULL;
}

/* try to determine the location of the donkey binary
*       returns NULL if nothing is found
*/
gchar *
find_donkey_binary (void)
{
	gchar	*ret = NULL;
	GSList	*dirlist = misc_get_directories_in_path (TRUE,TRUE);

	PROCESS_GSLIST (dirlist, gchar,
		if (!ret)
		{
			ret = check_dir_for_binary(item);
		}
		g_free(item);
	);
	g_slist_free(dirlist);

	return ret;
}


// misc_get_directories_in_path
//
// returns a GSList with directories that are in the $PATH variable, and
//	the current dir and home if required.


static GSList *
misc_get_directories_in_path (gboolean withcurdir, gboolean withhome)
{
	GSList	*list = NULL;
	gchar	**dirarr, **onedir;
	gchar	*dirs;
	gchar	*path = NULL;
	gchar	*curdir = NULL;
	gchar	*homedir = NULL;

	if (withcurdir)
		curdir = g_get_current_dir();		// use g_strdup() here?
	if (!curdir)
		curdir = g_strdup(N_(""));
	if (withhome)
		homedir = g_get_home_dir();
	if (!homedir)
		homedir = g_strdup(N_(""));
	path = g_strdup(g_getenv(N_("PATH")));
	if (!path)
		path = g_strdup(N_(""));

	dirs = g_strdup_printf (N_("%s%s%s%s%s"),
								curdir, (*curdir!=0x00) ? N_(":") : N_(""),
								homedir, (*homedir!=0x00) ? N_(":") : N_(""),
								path);
	g_free(curdir);
	g_free(homedir);
	g_free(path);
	g_return_val_if_fail (dirs!=NULL,NULL);

	dirarr = g_strsplit	(dirs, N_(":"), 0);
	g_free(dirs);
	g_return_val_if_fail (dirarr!=NULL,NULL);

	onedir = dirarr;
	while (*onedir)
	{
		if (**onedir)
			list = g_slist_append (list, (gpointer) g_strdup(*onedir));
		onedir++;
	}

	g_strfreev(dirarr);

	return list;	
}

// whereis_binary
//
// searches for a program file in the current directory, the HOME directory and the
//	PATH environment and returns the full path found or NULL if not found
//
// CALLER MUST FREE RETURNED STRING

static gchar *
whereis_binary (const gchar *progname, gboolean searchcurdir, gboolean searchhome)
{
	GSList	*dirlist;
	gchar	*foundpath = NULL;

	g_return_val_if_fail (progname!=NULL, NULL);

	dirlist = misc_get_directories_in_path (searchcurdir, searchhome);
	PROCESS_GSLIST (dirlist, gchar,
		if ((item)&&(!foundpath))
		{
			gchar *fullpath = g_strdup_printf (N_("%s%s%s"), item,
								(item[strlen(item)-1]==SLASH_CHAR) ? N_("") : SLASH_STRING,
								progname);
			if (fullpath)
			{
//				g_print ("checking '%s'...\n", fullpath);
				if (access(fullpath, F_OK)==0)
					foundpath = fullpath;
				else g_free(fullpath);
			}
			g_free(item);
		} else if (item) g_free(item);
	);
	g_slist_free(dirlist);	
	return foundpath;
}



// invoke_browser_with_url
//
// invokes a browser to display the given URL.

void
invoke_browser_with_url (const gchar *url)
{
	gchar	*browserpath = NULL;

	g_return_if_fail (url!=NULL);

	// check whether $BROWSER environment variable is set (thanks to Yorzik for this)
	browserpath = g_strdup(g_getenv(N_("BROWSER")));

	statusbar_extra_message (_("invoking browser... (environment variable $BROWSER is %s)\n"),
								(browserpath!=NULL) ? _("set") : _("not set"));

	// if it is set, check whether the file actually exists	
	if (browserpath)
	{
		// $BROWSER is not a full path? => find full path
		if (!strchr(browserpath,SLASH_CHAR))
		{
			gchar *browser = browserpath;
			browserpath = whereis_binary (browser, FALSE, FALSE);
			g_free(browser);
		}
		if (((browserpath)&&(access(browserpath, F_OK)!=0))||(!browserpath))
		{
			g_printerr (_("Environment variable $BROWSER contains invalid or incomplete path to browser binary.\n"));
			if (browserpath)
				g_free(browserpath);
			browserpath = NULL;
		}
	} 

	// $BROWSER not set or invalid? => find us a browser
	if (!browserpath)
	{
		const gchar	*browsers[] = { N_("galeon"), N_("konqueror"), N_("mozilla-bin"), N_("netscape"), N_("opera"), NULL };
		const gchar **check = browsers;

		// find us a browser
		while ((*check) && (!browserpath))
		{
			gchar *found = whereis_binary (*check, FALSE, FALSE);
			if (found)
				browserpath = found;
			check++;
		}
	}

	// display error message if we didn't find a browser
	if (!browserpath)
	{
		ok_dialog (	_("Help"), _("I tried to invoke a browser,\n"
					"but couldn't find one.\n\n"
					"Please set the environment variable $BROWSER\n"
					"to a full path to a browser binary of your choice.\n"));
		statusbar_extra_message (N_("    "));
 	} else {
		// invoke it...
		switch (fork())
		{
			case -1:	statusbar_extra_message (_("fork() to start browser failed!\n"));
			break;

			case 0:		// we are the child - invoke the browser
			{
				struct rlimit	rl;
				int				i;
				// close all descriptors 
				// this is very expensive since it can close as much as
				// thousand descriptors
				getrlimit(RLIMIT_NOFILE, &rl);
				for (i = rl.rlim_max - 1; i >= 3; i--)
				{
					(void)close(i);
				}

//				fprintf(stderr, ">>>>>>>> calling '%s' with url '%s' \n", browserpath, url);
				execl (browserpath, g_basename(browserpath), url, NULL);
				fprintf(stderr, _("**** execl() to start browser failed!\n"));
				_exit(0);
			}
			break;

			default:	// we are the parent - do nothing
			break;
		}
		g_free (browserpath);
	}
}

// get time stamp

gchar *
get_timestamp (void)
{
	static gchar	 stamp[256];
	time_t			 timenow = time(NULL);
	struct tm		*tmnow = localtime (&timenow);

	g_snprintf (stamp, sizeof(stamp), N_("%02d%02d%02d %02d.%02d.%02dh - "), tmnow->tm_mday, tmnow->tm_mon+1,
			 tmnow->tm_year-100, tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec);
	return stamp;
}


/* misc_rotate_logfiles
 *
 *
 *
 */

static void
misc_rotate_logfiles (const gchar *logfilename)
{
	static gchar	*compress_program = NULL;	
	static gchar	*compressed_extension = NULL;
	const gchar		*ret = NULL;
	gint			 number;

	g_return_if_fail (logfilename!=NULL);

	status_message_blue (_("GUI: Logfile 'gui_statuslog' is getting big. Rotating log files...\n"));

	/* if we haven't done so before, check for a compress program on this system
	 * (note: if we have found one, compress_program and compressed_extension are set,
	 *  if we checked and haven't found one, compress_program will be NULL and
	 *  compressed_extension will be set (to "")
	 */
	if (!compressed_extension)
	{
		/* check whether we have bzip2 on this system (and in our PATH) */
		ret = find_specified_binary_in_path(N_("bzip2"));
		/* found bzip2 */
		if (ret)
		{
			compress_program = g_strdup (ret);	/* this leaks once, but who cares? */
			g_return_if_fail (compress_program!=NULL);
			compressed_extension = N_(".bz2");
		}

		/* if not, check whether we have gzip on this system (and in our PATH) */
		if (!ret)
		{
			ret = find_specified_binary_in_path(N_("gzip"));
			/* found gzip */
			if (ret)
			{
				compress_program = g_strdup (ret);	/* this leaks once, but who cares? */
				g_return_if_fail (compress_program!=NULL);
				compressed_extension = N_(".gz");
			}
		}
	}

	/* neither bzip2 nor gzip found? => don't compress */
	if (!compress_program)
	{
		compressed_extension = N_("");
	} else status_message_blue (_("GUI: Using '%s' for compression\n"), compress_program);

	g_return_if_fail (compressed_extension!=NULL);

	/* rotate old (possibly compressed) statuslogs (max. 999) */
	for (number=999; number>=0; number--)
	{
		gchar *oldlogfn = g_strdup_printf (N_("%s-old.%d%s"), logfilename, number, compressed_extension);
		g_return_if_fail (oldlogfn!=NULL);
		/* check whether old log file exists in the first place */
		if (access(oldlogfn,F_OK)==0)
		{
			gchar *oldlogfn2 = g_strdup_printf (N_("%s-old.%d%s"), logfilename, number+1, compressed_extension);
			g_return_if_fail (oldlogfn2!=NULL);
			/* shift old log file up one number */
			status_message_blue (_("GUI: Renaming old logfile '%s' -> '%s'\n"), g_basename(oldlogfn), g_basename(oldlogfn2));
			if (rename(oldlogfn, oldlogfn2)<0)
				status_system_error_msg (_("Couldn't rename() old status logfile.\n"));
			g_free(oldlogfn2);
		}
		g_free(oldlogfn);
	}

	/* now rename current logfile into gui_statuslog<.instance>.0 and compress it */
	if (1)
	{
		gchar *newlogfn = g_strdup_printf (N_("%s-old.0"), logfilename);
		g_return_if_fail (newlogfn!=NULL);
		status_message_blue (_("GUI: Renaming current logfile '%s' -> '%s'\n"), g_basename(logfilename), g_basename(newlogfn));
		if (rename(logfilename,newlogfn)==0)
		{
			/* now compress it if we found a compress program */
			if (compress_program)
			{
				gchar *sysmsg = g_strdup_printf (N_("%s \"%s\""), compress_program, newlogfn);
				g_return_if_fail (sysmsg!=NULL);
				status_message_blue (_("GUI: Compressing logfile '%s' -> '%s%s'\n"), g_basename(newlogfn), g_basename(newlogfn), compressed_extension);
				/* XXX - would there be any point in checking the return value here? */
				system (sysmsg);
				g_free(sysmsg);
			}
		} else status_system_error_msg (_("Couldn't rename() current status logfile.\n"));
		g_free(newlogfn);
	}

	status_message_blue (_("GUI: Log rotation complete.\n"));
}

// append string logstr to logfile

static void
write_to_logfile_real (const gchar *logstr)
{
	FILE				*lf;
	static gchar		*fn = NULL;

	/* lock, so the status log output that the log rotate function creates will not
	 * recursively lead to doom if max logsize is set smaller than the output created
	 * by the status log rotate function...
	 */
	static gboolean		lock = FALSE;

	if (lock)
		return;

	if (!prefpath)
		return;

	if (!fn)
	{
		const gchar *fn2 = complete_pref_path (N_("gui_statuslog"), GUI_CONFIG_USE_INSTANCE);
		g_return_if_fail (fn2!=NULL);
		fn = g_strdup (fn2);
		g_return_if_fail (fn!=NULL);
	}
	
	if (logstr==NULL || *logstr==0x00)
		return;

	if ((option_nolinkstolog) && strncmp (logstr, N_("ed2k://"), 6)==0)
		return;

	if (logstr[0]==' ' && logstr[1]==0x00)
		return;

	if (strncmp(logstr, N_("No more servers to extend to"),10)==0)
		return;

	if ((lf=fopen(fn, N_("a")))!=NULL)
	{
		static time_t	last_checked_logfilesize = (time_t) 0;
		struct stat		logstat;

		fprintf (lf, N_("%s%s"), get_timestamp(), logstr);

		/* make sure logfile doesn't get too big, but don't
		 * check size more often than every minute
		 */
		if (time(NULL) - last_checked_logfilesize > 10)
		{
//			g_print ("STAT'ing\n");
			last_checked_logfilesize = time(NULL);
			if (fstat (fileno(lf), &logstat)==0)
			{
//				g_print ("STAT OK - size = %u\n", logstat.st_size);
				if (logstat.st_size > (option_maxlogfilesize*1024))
				{
					fclose(lf);		/* close file before we move/compress it */
					lf = NULL;
					lock = TRUE;	/* see above why */
					misc_rotate_logfiles(fn);
					lock = FALSE;
				}
			} else status_system_error_msg (_("Couldn't fstat() logfile gui_statuslog\n"));
		}

		if (lf) fclose(lf);
	}
}

/* write_to_logfile
 *
 */

void
write_to_logfile (const gchar *format, ...)
{
	GString	*output = NULL;
	gchar	*s_alloc = NULL;
	va_list	 ap;

	g_return_if_fail (format!=NULL);

	output = g_string_new("");
	g_return_if_fail (output!=NULL);

	va_start(ap, format);
	if (vasprintf(&s_alloc, format, ap)>=0)
	{
		output = g_string_append (output, s_alloc);
		free(s_alloc);	/* as vasprintf allocs it, we need to use free() and not g_free() */
		s_alloc = NULL;
	}
	va_end(ap);

	write_to_logfile_real(output->str);
	g_string_free(output,TRUE);
}



// add_menuitem
//
//

GtkWidget *
add_menuitem (	GtkWidget *menu,
				const gchar *text,
				void *function,
				gpointer data,
				gchar **icon)
{
	GtkWidget	*item;

	item = gtk_menu_item_new();
	if (text)
	{
		gtk_signal_connect (GTK_OBJECT(item), N_("activate"), GTK_SIGNAL_FUNC(function), data);
                gtk_container_add (GTK_CONTAINER(item), make_icon_and_label_hbox(text,icon));
	}
	gtk_menu_append (GTK_MENU(menu), item);
	gtk_widget_show (item);
	return (item);
}


// seconds_into_human_time
//
// takes a span of time in units of secons and returns it in 'human' format (e.g. '5.4h')

const gchar *
seconds_into_human_time (guint seconds)
{
	static gchar	 output[128];
	gchar 			*timeunit = NULL;
	gfloat 			 actual_time = (gfloat) seconds;

	if (seconds==0)
		return _("now");

	if (actual_time>=60.0)
	{
		actual_time = actual_time/60.0;	// now it's in minutes
		if (actual_time>=60.0)
		{
			actual_time = actual_time/60.0;	// now it's in hours
			if (actual_time>=24.0)
			{
				actual_time = actual_time/24.0;	// now it's in days
				if (actual_time>=365.25)
				{
					actual_time = actual_time/365.25; // now it's in years
					timeunit = _("y");
					// assume no one keeps a file in the queue longer than 2yrs
					if (actual_time>=2.0)
					{
						return _("never");
					}
				} else timeunit = _("d");
			} else timeunit = _("h");
		} else timeunit = _("m");
	} else timeunit = _("s");

	g_snprintf (output, sizeof(output), N_("%.1f%s"), actual_time, timeunit);
	return output;
}

// produce_uldlstat_line
//
// Takes some statistics and calculates some more stats from those and prints out
//	those stats in the statusbar
//
// which = "upload" or "download"
// total = total ul/dl in kB
// seconds_nonzero = number of seconds upload/download was >0 (=actually down/uploading)

static void
produce_uldlstat_line (const char *which, float total, time_t seconds_nonzero)
{
	// secs since stat last cleared
	float secs_since = (float)(time(NULL)-stats_last_cleared);

	// days since stat last cleared
	float days_since = secs_since/(24.0*3600.0);

	// total ul/dl in GB instead of kB
	float total_GB   = total/(1024.0*1024.0);	

	// we don't want to divide by 0 if not uploaded/downloaded yet, so +0.0000001
	// secs downloading since stats last cleared
	float actual_kBs = total/(seconds_nonzero+0.00000001);		

	// percent of the time we spent *loading
	float percent    = ((float) seconds_nonzero) / secs_since * 100.0;

	if (percent > 100.0) percent = 100.0;
	statusbar_extra_message (_(" Total %s: ~%.2fG in %.1f days (~%.1fM/day) - actual: ~%.1fkB/s, %sing %.1f%% of the time (%s)"),
		 which, total_GB , days_since, (total/1024.0)/days_since, actual_kBs, which, percent, seconds_into_human_time (seconds_nonzero));
}

void
print_download_stats_line (void)
{
	produce_uldlstat_line (_("download"), total_downloaded, seconds_download_was_nonzero);
}

void
print_upload_stats_line (void)
{
	produce_uldlstat_line (_("upload"), total_uploaded, seconds_upload_was_nonzero);
}



// exec_command_on_complete_download
//
// takes the filename of the file that is complete and runs the command specified in options
//	when a download is complete (if a command is specified that is)

void
exec_command_on_complete_download (const gchar *filename)
{
	g_return_if_fail (filename!=NULL);
	g_return_if_fail (option_exec_on_complete!=NULL);
	g_return_if_fail (option_incoming!=NULL);

	// is a command specified in the options?
	if (*option_exec_on_complete)
	{
		gchar *cmdlinestr = g_strdup_printf (N_("%s%s"),
								option_exec_on_complete,
								(option_exec_on_complete[strlen(option_exec_on_complete)-1]=='&') ? N_("") : N_(" &"));

		g_return_if_fail (cmdlinestr!=NULL);
	
		// set environment variables ED2K_FN and ED2K_IN
		(void) setenv (N_("ED2K_FN"), filename,1);			// 1 = overwrite old value
		(void) setenv (N_("ED2K_IN"), option_incoming,1);	// 1 = overwrite old value

		// now run the specified command line in a shell in the background (we've added the '&' at the end)
		if (system (cmdlinestr)<0)
		{
			g_printerr (_("exec shell command on complete failed for some reason:\n"));
			g_printerr (N_("%s\n"), cmdlinestr);
			status_system_error_msg (_("Error:"));
		}
		
		g_free(cmdlinestr);
	}
}


// remove_junk_chunk
//
// remove chunk in string INCLUDING both from and to - no validity tests made for parameters!

void
remove_junk_chunk (gchar *str, gchar *from, gchar *to)
{
	if ((!str) || (!from) || (!to))
		return;

	// remove spaces before bracket if sth else than a letter or a number follows the expression
	if(!isalnum(to[1]))
	{
		while (from>str &&  from[-1]==' ')
			from--;
	}

	// and remove chunk

	/* if there is nothing after 'to', then it suffices to terminate string at beginning of section to remove */
	if (to[0] == 0x00)
	{
		from[0] = 0x00;
		return;
	}

	/* we can't use strcpy(from,to+1) here because 'man strcpy' says the "strings may not overlap" (?!) */
	while (*to)
	{
		*from = *(to+1);	/* +1 as we want to delete the string INCLUDING the character at *to  */
		from++;
		to++;
	}
}


// remove_word_chunks
//
// remove all chunks which match a given word case-insensitive
// original word must be lower case!

static void
remove_word_chunks (gchar *str, const gchar *w)
{
	gint	n, w1len;

	if ((!w) || (!str))
		return;

	n = (w[0]==' ') ? 1 : 0;
	w1len = strlen(w) - 1 - n;

	if (w1len != -1)
	{
		gchar	*word = g_strdup(w);	/* to contain the word chunk to remove in lower case */
		gchar	*ostr = g_strdup(str);	/* to contain the original string in lower case */
		gchar	*match;

		g_strdown(word);
		g_strdown(ostr);

		while ((match = strstr(ostr,word))!=NULL)
		{
			guint idx = (match - ostr) + n;
			remove_junk_chunk (ostr, ostr + idx, ostr + idx + w1len);
			remove_junk_chunk (str, str + idx, str + idx + w1len);
				
/* original:
			remove_junk_chunk (str, match+n, match+strlen(w)-1);
			remove_junk_chunk (str, str+(match-ostr)+n, str+(match+strlen(w)-ostr-1));
*/
		}

		g_free (word);
		g_free (ostr);
	}
}

// remove_word_chunk_until_end
//
//

static void
remove_word_chunk_until_end (gchar *str, gchar *phrase)
{
	gchar *d, *e;
	if ((!str) || (!phrase)) return;
	d = strstr (str, phrase);
	if (d!=NULL)
	{
		e=strchr(d+strlen(phrase)+1,' ');
		if (e==NULL) e=strchr(d+strlen(phrase)+1,'.');
		if (e==NULL) e=d+strlen(d)-1;
		if (e>d) remove_junk_chunk (str,d,e);
	}
}

// remove_word_chunk_and_previous_word
//
// note: this will only remove one instance of the word

static void
remove_word_chunk_and_previous_word (gchar *str, gchar *word)
{
	gchar *s, *match, c;

	if ((!str) || (!word))
		return;

	match = strstr (str,word);

	if ((!match) || (match<str+2))			// word not found
		return;

//	printf ("match = '%s'\n", match);
	c = match[-1];						// -1 as we don't want the space
	match[-1] = 0x00;
	s = strrchr (str,' ');
	match[-1] = c;
	if (!s) return;				// only one or none word before
//	printf ("previous word = '%s'\n", s);
	remove_junk_chunk (str,s+1,match+strlen(word)-1);
}


// remove_bracket_junks
//
// remove all chunks of a certain bracket type

void
remove_bracket_junks (gchar *str, gchar open, gchar close)
{
	gchar *s, *e;	// start, end of junk chunk
	// remove brackets - assuming that if we have two brackets there's nothing useful in between both
	while ((s=strchr(str, open))!=NULL)
	{
		if ((e=strchr(s+1, close))==NULL) return;	// no closing bracket found -> finito
		remove_junk_chunk (str,s,e);
	}
}

static const char *junkwords[] =
{
 N_("german"), N_("english"), N_("french"), N_("spanish"), N_("dvdrip"), N_("svcd"), N_("dvd"),
 N_("sddivx"), N_("dubbed"), N_("subtitled"), N_("subtitles"), N_("tgsc"), N_("deutsch"),
 N_("divx"), N_("shared"), N_("vcd"), N_("vhsrip"), N_("vhs-rip"), N_("vhs"), N_("scr "),
 N_("shared by"), N_("released by"), N_("release by"), N_("release"), N_("slicer"),
 N_("vbr"), N_("stereo"), N_("mono "), N_("128kbs"), N_("160kbs"), N_("192kbs"), N_("256kbs"),
 N_("&amp;"), N_("#tradeftp"), N_("kein fake"), N_("no fake"), N_("25fps"), N_("engl "),
 N_("engl. "), N_(" dt "), N_("dt."), N_("uncut"), N_("high quality"), N_("CloneCD image"),
 N_("surround "), N_("ppv-rip"), N_("directors cut"), N_("director's cut"), N_(" cam "),
 NULL
};

/*
// remove_multiple_characters
//
//

static void
remove_multiple_characters (gchar *str, gchar c)
{
	gint	i;
	gchar	dummy[10];

	memset (dummy, c, 10);
	for (i=9; i>1; i--)
	{
		dummy[i]=0x00;
		remove_word_chunks (str, dummy);	// max. five occurences of each len
		remove_word_chunks (str, dummy);
		remove_word_chunks (str, dummy);
		remove_word_chunks (str, dummy);
		remove_word_chunks (str, dummy);
	}
}
*/

// remove_junk_from_filename
//
//

void
remove_junk_from_filename (char *str)
{
	char *d, *e;
	int i;
//	int debug = (str[0]=='M');
//	if (debug) printf ("1 - %s\n", str);
	remove_bracket_junks (str, '(', ')');
	remove_bracket_junks (str, '{', '}');
	remove_bracket_junks (str, '[', ']');
	remove_bracket_junks (str, '=', '=');
//	if (debug) printf ("2 - %s\n", str);

	// substitute '_' with spaces
	for (d=str; d[0]!=0x00; d++)
	{
		switch (d[0])
		{
			case '-':
			case '=':
			case '_':
			case ':':
			case '+':
			case '%':
			case '$':
			case '#':
			case '^':
			case '*': d[0]=' ';
			default:
				break;
		}
	}

//	if (debug) printf ("3 - %s\n", str);

	// substitute all dots but the last with spaces
	e = strrchr (str, '.');
	if (e)
	{
		while ((d = strchr(str,'.')) && d<e) d[0]=' ';
	}
//	if (debug) printf ("4 - %s\n", str);

	remove_word_chunk_until_end (str, N_("shared by"));
	remove_word_chunk_until_end (str, N_("release by"));
	remove_word_chunk_until_end (str, N_("released by"));
	remove_word_chunk_until_end (str, N_("realase by"));
	remove_word_chunk_until_end (str, N_("ripped by"));
	remove_word_chunk_until_end (str, N_("rip by"));
	remove_word_chunk_until_end (str, N_("capture by"));
	remove_word_chunk_until_end (str, N_("captured by"));
	remove_word_chunk_until_end (str, N_("encoded by"));
	remove_word_chunk_until_end (str, N_("encoder by"));
	remove_word_chunk_until_end (str, N_("resized by"));
	remove_word_chunk_and_previous_word (str, N_("crew"));
//	if (debug) printf ("5 - %s\n", str);

	// remove file extensions if not too long
	if (option_keep_extensions==0)
	{
		d = strrchr (str,'.');
		if (((str+strlen(str))-d)<=6)
		{
			d[0]=0x00;
		}
//		if (debug) printf ("6 - %s\n", str);
	}

	// remove all chunks with words we don't want
	i=0;
	while (junkwords[i])
	{
		remove_word_chunks (str, junkwords[i]);
		i++;
	}
//	if (debug) printf ("7 - %s\n", str);

//	we've turned these into spaces now anyway
//	remove_multiple_characters (str, '-');	// twice so we get encapsulations
//	remove_multiple_characters (str, '=');
//	remove_multiple_characters (str, '-');
//	remove_multiple_characters (str, '=');

	// if double spaces, we only want to remove one, not both, do we?
	if (str[0]!=0x00)
	{
		d = str+1;
		while (d[0]!=0x00)
		{
			if (d[-1]==' ' && d[0]==' ')
			{
				remove_junk_chunk (str,d,d);
			} else d++;
		}
	}
//	if (debug) printf ("8 - %s\n", str);

	// remove initial spaces
	while (str[0]==' ') remove_junk_chunk (str,str,str);
//	if (debug) printf ("9 - %s\n", str);

	// remove spaces before file extension dot - looks uncool
	if (option_keep_extensions)
	{
		d = strrchr (str,'.');
		if (d>str)
		{
			if (d[-1]==' ') remove_junk_chunk (str,d-1,d-1);
		}
	}
}

// function for all those clist compare functions

void
get_text_from_gtk_clist_rows (GtkCListRow *row1, GtkCListRow *row2, guint sort_column, gchar **text1, gchar **text2)
{
	g_return_if_fail (text1!=NULL);
	g_return_if_fail (text2!=NULL);
	g_return_if_fail (row1!=NULL);
	g_return_if_fail (row2!=NULL);

  switch (row1->cell[sort_column].type)
    {
    case GTK_CELL_TEXT:
      *text1 = GTK_CELL_TEXT (row1->cell[sort_column])->text;
      break;
    case GTK_CELL_PIXTEXT:
      *text1 = GTK_CELL_PIXTEXT (row1->cell[sort_column])->text;
      break;
    default:
      break;
    }

  switch (row2->cell[sort_column].type)
    {
    case GTK_CELL_TEXT:
      *text2 = GTK_CELL_TEXT (row2->cell[sort_column])->text;
      break;
    case GTK_CELL_PIXTEXT:
      *text2 = GTK_CELL_PIXTEXT (row2->cell[sort_column])->text;
      break;
    default:
      break;
    }
}

// reallocs the char* pointer variable *dest with the space needed by the string
//  and fills it with the source string src
// returns 0 on success
// returns -1 on errors

int strrealloccpy (char **dest, char *src)
{
	guint srclen;
	g_return_val_if_fail (dest!=NULL, -1);
	g_return_val_if_fail (src!=NULL, -1);
	if (*dest == src) return 0;	// nothing to do
	srclen = strlen(src);
	*dest = realloc (*dest, srclen+1);
	if (*dest==NULL) return (-1);	// memory alloc problem
//	printf ("*dest = %10x - srclen=%d\n", (unsigned int)(*dest), srclen);
	memset (*dest, 0x00, srclen+1);
	strncpy (*dest, src, srclen);
	return 0;
}


// choplast
//
// chops off last character of a string
// returns: pointer to string
//

gchar *
choplast (gchar *str)
{
	g_return_val_if_fail (str!=NULL,NULL);
	str[strlen(str)-1]=0x00;
	return str;
}



/* spawn_file_preview
 *
 * returns 0 if everything goes ok, otherwise something <0
 *
 */

gint
spawn_file_preview (const gchar *filename, const gchar *extension)
{
	const char	 program_base[] = N_("xterm");
	const char	*program_preview = complete_pref_path (N_("ed2k_gui_preview"), FALSE);
	char	 	 program_preview2[PATH_MAX+NAME_MAX+1];
	char *const	 parameter[] = { (char*) program_base, N_("-e"), program_preview2, (char*) filename, NULL } ;
	char 		 resolved_path[PATH_MAX], dir_path[PATH_MAX];
	int i; 
	pid_t pid;
	struct rlimit rl;

	dir_path[0] = resolved_path[0] = '\0';

	g_return_val_if_fail (filename!=NULL, -2);
	g_return_val_if_fail (program_preview!=NULL, -5);

	strcpy (program_preview2, program_preview);

	if (realpath(filename, resolved_path) == NULL)
	{
		g_printerr (_("Preview: could not resolve resolved_path from filename '%s'!\n"), filename);
		return(-3);
	}

	// If you thread this program, g_dirname is not thread safe
	// in most OSes
	// strcpy is safe here because g_dirname never overflows
	// MAXPATHLEN

	if (strcpy(dir_path, g_dirname(resolved_path)) == NULL)	// strdup->leaks
	{
		g_printerr (_("Preview: could not get dir_path from resolved_path '%s'!\n"), resolved_path);
		return(-3);		// low memory
	}

	/* check whether directory with temp files is accessible and readable */
	if (access(dir_path, R_OK | X_OK) != 0)
	{
		g_printerr (_("Preview: cannot access dir_path '%s'!\n"), dir_path);
		return(-4);		// directory is not accessible
	}

	/* check whether preview script exists in config folder */
	if (access(program_preview2, R_OK | X_OK) != 0)
	{
		/* preview script does not exist yet. Create it. */
		FILE *of;
		of = fopen (program_preview2, N_("w"));
		if (!of)
		{
			gchar *msg = g_strdup_printf (_("Could not create ed2k_gui_preview script '%s'!?\n"), program_preview);
			status_system_error_msg (msg);
			g_free(msg);
			return (-6);
		}
		status_message_blue ("GUI: Creating GUI preview skript in '%s'...\n", program_preview2);

		fprintf (of, N_("#!/bin/sh\n\n"));

		fprintf (of, N_("PREVIEW_BIN=\"\"\n"));
		fprintf (of, N_("MP3PLAYER=\"mpg123 -v\"\n"));
		fprintf (of, N_("OGGPLAYER=\"ogg123\"\n"));
		fprintf (of, N_("VIDPLAYER=\"mplayer -idx\"\n"));

		fprintf (of, N_("\n"));
		fprintf (of, _("echo \"Prototype of ed2k_gui_preview !\"\n"));
		fprintf (of, _("echo \"Got the following parameter: $*\"\n"));
		fprintf (of, _("echo \"File extension             : $ED2K_PREVIEW_EXTENSION\"\n"));
		fprintf (of, N_("echo\n\n"));
		fprintf (of, N_("\n"));

		fprintf (of, N_("case \"$ED2K_PREVIEW_EXTENSION\" in\n"));

		fprintf (of, N_("mp3)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$MP3PLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("ogg)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$OGGPLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("avi)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$VIDPLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("divx)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$VIDPLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("mpg)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$VIDPLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("mpeg)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$VIDPLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("# .bin files are usually mpeg with some format junk.\n"));
		fprintf (of, N_("# mplayer can play .bin files\n"));
		fprintf (of, N_("bin)\n"));
		fprintf (of, N_("\tPREVIEW_BIN=\"$VIDPLAYER\"\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("*)\n"));
		fprintf (of, _("\techo \"Don't know which binary to use with this format.\"\n"));
		fprintf (of, _("\techo \"Please configure this script (it's easy, don't worry)\"\n"));
		fprintf (of, N_("\techo\n"));
		fprintf (of, _("\techo \"Hit <enter> to close this window.\"\n"));
		fprintf (of, N_("\tread\n"));
        fprintf (of, N_("\texit 0\n"));
		fprintf (of, N_("\t;;\n"));

		fprintf (of, N_("esac\n"));

		fprintf (of, N_("$PREVIEW_BIN \"$*\"\n\n"));
		fprintf (of, _("echo \"Finished. Hit <enter> to close this window\"\n"));
		fprintf (of, N_("read\n\n"));
		/* make script executable */
		if (fchmod(fileno(of), S_IRUSR|S_IWUSR|S_IXUSR)<0)
		{
			status_system_error_msg (_("Couldn't make script executable!\n"));
		}
		fclose(of);
		status_message_blue (_("GUI: preview script '%s' created.\n"), program_preview2);

		ok_dialog (_("preview script created"),
				   _("I have created a preview script in the GUI configuration file\n"
					 "directory (probably $HOME/.ed2k_gui/ed2k_gui_preview\n"
					 "unless you specified a different directory on the command line).\n\n"
					 "You can customise this script to use the player of you choice\n"
					 "to preview files depending on their file extension...\n"));
	}

	status_message_blue ("GUI: preview - parameter used: %s %s\n", program_preview2, filename);

	if ((pid = fork()) == -1)
		return(-1);		// cannot fork

	// parent continues
	if (pid != 0)
	{
		status_message_blue(_("GUI: preview: got a new pid: %u\n"), pid);
		sleep(1);		// sleep a bit to avoid hammering
		return(0);		// forked fine
	}

	// close all file descriptors (apart from the 3 standard ones)
	getrlimit(RLIMIT_NOFILE, &rl);
	for (i = rl.rlim_max - 1; i >= 3; i--)
	{
		(void)close(i);
	}

	/* export file extension in lower case */
	if (extension)
	{
		gchar *ext_lowercase = g_strdup(extension);
		if (ext_lowercase)
		{
			g_strdown(ext_lowercase);
			setenv(N_("ED2K_PREVIEW_EXTENSION"), ext_lowercase, 1);
			g_free(ext_lowercase);
		} else setenv(N_("ED2K_PREVIEW_EXTENSION"), extension, 1);
	}

	// child continues
	execvp(program_base, parameter);
	perror (_("bad error in execvp()"));

	_exit(0);
}



// misc_get_array_of_lines_from_textfile
//
// Reads in a text file and returns a NULL-terminated array of strings
//  each of which represents one line. The lines are returned without the
//  final '\n' character at the end of the line.
//
// The returned string array must be freed by the caller using g_strfreev()
//
// returns NULL if the file couldn't be read or some other grave error occured

gchar **
misc_get_array_of_lines_from_textfile (const gchar *filename)
{
	gint	  lines = 0;
	FILE	 *file;
	GString	 *linestr = NULL;
	gchar	**arr = NULL;

	g_return_val_if_fail (filename!=NULL, NULL);
	file = fopen(filename,N_("r"));
	if (!file) return NULL;

	while (!(feof(file)||ferror(file)))
	{
		gint	c;
		if (!linestr)
		{
			linestr = g_string_sized_new(127);
			g_return_val_if_fail (linestr!=NULL, NULL);
			arr = g_renew (gchar*, arr, lines+1);
			g_return_val_if_fail (arr!=NULL, NULL);
		}

		c = fgetc (file);
		if ((gchar)c!='\r' && (gchar)c!='\n' && (gchar)c!=EOF)	// ignore line breaks
		{
			g_string_append_c (linestr, (gchar)c);
		}
		if ((gchar)c=='\n' || (gchar)c==EOF)	// do we have a complete line?
		{
			arr[lines] = linestr->str;	// yes: add it to our array of lines
			g_string_free (linestr,FALSE);	// free string structure, but not chars
			linestr = NULL;			// so we get a new one in the next round
			lines++;
		}
	}
	if (linestr) g_string_free (linestr,TRUE);
	arr = g_renew (gchar*, arr, lines+1);
	g_return_val_if_fail (arr!=NULL, NULL);
	arr[lines] = NULL;	// terminate it
	fclose(file);
	return arr;
}


// misc_gnet_inetaddr_get_ip
//
// Gets the IP Address [XXX - in which order] from a GInetAddr

guint32	
misc_gnet_inetaddr_get_ip (const GInetAddr *ia)
{
	gchar	*name;
	guint32	n1,n2,n3,n4;
	g_return_val_if_fail (ia!=NULL,0);
	name = gnet_inetaddr_get_canonical_name (ia);
	g_return_val_if_fail (name!=NULL,0);
	g_return_val_if_fail (sscanf(name,N_("%u.%u.%u.%u"),&n1,&n2,&n3,&n4)==4,0);
	g_return_val_if_fail ((n1<256 && n2<256 && n3<256 && n4<256), 0);
	// XXX - shifts are byte order independent, aren't they?
	if (G_BYTE_ORDER!=G_LITTLE_ENDIAN)
		g_print (N_("CHECK ME: %s\n"), __FUNCTION__);

	g_free(name);
	return (n1 | (n2<<8) | (n3<<16) | (n4<<24));
}


// misc_get_extension_from_filename
//
// Returns the extension of the filename in lower case,
//   or a pointer to an empty string if not found

const gchar *
misc_get_extension_from_filename (const gchar *name)
{
	static gchar	 ext[15];
	gchar		*dot;
	guint		 i = 0;

	memset (ext, 0x00, sizeof(ext));
	g_return_val_if_fail (name!=NULL, ext);
	dot = strrchr (name, '.');
	if (!dot) return ext;
	if (strlen(++dot)>5) return ext;
	while (dot[i])
	{
		ext[i] = tolower (dot[i]);
		i++;
	}
	return ext;
}


// misc_check_if_extension_is_in_string_array
//
// Returns TRUE if the extension ext matches an extension in the
//	NULL-terminated string array 'suffices', otherwise FALSE;

static gboolean
misc_check_if_extension_is_in_string_array (const gchar *ext, const gchar **suffices)
{
	const gchar	**suffix = suffices;

	if ((suffices) && (ext))
	{
		while ((*suffix) && (**suffix))
		{
			if (g_strcasecmp(*suffix,ext)==0) return TRUE;
			suffix++;
		}
	}
	return FALSE;
}

// misc_get_filetype_from_extension
//
// Tries to determine the file type (Audio/Video/etc.) from
//	the given extension and returns it as a string.
// Returns empty string if type couldn't be determined.

const gchar *
misc_get_filetype_from_extension (const gchar *ext)
{
	const gchar *audio[] = { N_("mp3"), N_("wav"), N_("au"), N_("ogg"), NULL };
	const gchar *video[] = { N_("rm"), N_("mpg"), N_("mpeg"), N_("avi"), N_("asf"), N_("mov"), N_("divx"), NULL };
	const gchar *prog[] = { N_("exe"), N_("zip"), N_("gz"), NULL };
	const gchar *doc[] = { N_("pdf"), N_("html"), N_("txt"), N_("doc"), N_("abw"), N_("rtf"), N_("asc"), N_("xml"), NULL };
	const gchar *imgs[] = { N_("jpg"), N_("jpeg"), N_("gif"), N_("png"), N_("bmp"), N_("xpm"), NULL };
	
	if (ext)
	{
		if (misc_check_if_extension_is_in_string_array (ext, audio)) return _("Audio");
		if (misc_check_if_extension_is_in_string_array (ext, video)) return _("Video");
		if (misc_check_if_extension_is_in_string_array (ext, prog)) return _("Pro");
		if (misc_check_if_extension_is_in_string_array (ext, doc)) return _("Document");
		if (misc_check_if_extension_is_in_string_array (ext, imgs)) return _("Image");
	}
	return N_("");
}


/* misc_gtk_entry_insert_text_handler
 *
 * special handler for a GtkEntry 'insert_text' event, so that the 'Enter'
 * key on the num block behaves just like the real enter key in gtk
 *
 * found on the gtk-list@gnome.org mailing list:
 * http://mail.gnome.org/archives/gtk-list/2002-April/msg00251.html
 * submitted by Jean-Christophe Berthon
 *
 */

void
misc_gtk_entry_insert_text_handler (GtkEntry *entry, const gchar *text, gint length, gint *position, gpointer pxData_E)
{
	GtkEditable			*editable = GTK_EDITABLE(entry);
	gchar				*result = NULL;

	/* That's what the numkeypad enter is producing for invisible caractere */
	if (text[length-1] == 0x0d)
	{
		/* remove bogus character produced by num block enter key at end */
		if (length > 1)
		{
			result = calloc(length - 1, sizeof(gchar));
			strncpy(result, text, length - 1);
		}
		/* make sure we're not recursively calling ourself when we change the text in the widget
		 * to the new text (ie. w/o the bogus character at the end. Thefore: block the signal function
		 */
		gtk_signal_handler_block_by_func (GTK_OBJECT(editable), GTK_SIGNAL_FUNC(misc_gtk_entry_insert_text_handler), pxData_E);
		/* insert text w/o bogus character */
		gtk_editable_insert_text (editable, result, length-1, position);
		gtk_signal_handler_unblock_by_func (GTK_OBJECT(editable), GTK_SIGNAL_FUNC(misc_gtk_entry_insert_text_handler), pxData_E);
		/* prevent all other widgets from handling this event any further */
		gtk_signal_emit_stop_by_name(GTK_OBJECT(editable), N_("insert_text"));
		/* simulate 'return' key event */
		gtk_signal_emit_by_name(GTK_OBJECT(editable), N_("activate"));
		free(result);
	} 
}


/* misc_get_ncpus
 *
 * returns the number of CPUs the current machine has from the
 * information in /proc/cpuinfo
 *
 * (we need to do this at least for the dns_lookup routines,
 *  because due to a bug in libgnet or glib the async dns lookup
 *  is prone to crash on SMP machines.)
 *
 * written by Roman Hodek.
 *
 * returns the number of CPUs or 0 in case of an error.
 *
 */

#define PROC_CPUINFO_FILE "/proc/cpuinfo"

static gint
misc_get_ncpus_check (void)
{
	FILE *f;
	gchar line[1024];
	gint n = 0;

	f = fopen(PROC_CPUINFO_FILE, "r");
	if (!f)
	{
		perror(PROC_CPUINFO_FILE);
		return 0;
	}

	while(fgets(line, sizeof(line), f))
	{
		if (g_strncasecmp(line, "processor", 9) == 0)
			 ++n;
	}

	fclose(f);

	return n;
}

gint
misc_get_ncpus (void)
{
	static gint num_cpus = -1;

	/* we only need to check once */
	if (num_cpus == -1)
	{
		num_cpus = misc_get_ncpus_check();
	}
	return num_cpus;
}


/* misc_ip_is_local
 *
 * IP is in 'edonkey-format' (ie. little endian, not network byte order)
 * returns TRUE if this IP is local
 *
 */

gboolean
misc_ip_is_local (guint32 ip)
{
	guint8 n4 = *((guint8*)&ip);
	guint8 n3 = *(((guint8*)&ip)+1);
	guint8 n2 = *(((guint8*)&ip)+2);
	guint8 n1 = *(((guint8*)&ip)+3);

	if (n4==10)
		return TRUE;

	if (n4==172 && n3>=16 && n3<=31)
		return TRUE;

	if (n4==192 && n3==168)
		return TRUE;	

	if (n4==127 && n3==0 && n2==0)
		return TRUE;

	return FALSE;
}


/******************************************************************************
 *
 *   misc_check_if_ed2klink_is_valid
 *
 *   returns TRUE if the string passed is a valid ed2klink, otherwise FALSE
 *   (note: it does not check for 'bad characters' in filename)
 *
 *   If fn, size and hashstring are set, and the link is valid, 
 *    memory is allocated and these vars are set to contain the fields.
 *    Caller must free this allocated memory!
 *
 */

gboolean
misc_check_if_ed2klink_is_valid (const gchar *ed2klink, gchar **fn, guint *size, gchar **hashstring)
{
	const gchar *checkpos = ed2klink;
	const gchar *fnpos;
	const gchar *sizepos;
	const gchar *hashpos;
	guint        filenamelength = 0;
	guint        hashlength = 0;
	guint        sizelength = 0;

	g_return_val_if_fail (ed2klink!=NULL, FALSE);

	if (g_strncasecmp(checkpos,N_("ed2k:"),5)!=0)
		return FALSE;

	checkpos += 5;

	while (*checkpos == '/')
		checkpos++;

	if (g_strncasecmp(checkpos,N_("|file|"),6)!=0)
		return FALSE;

	checkpos +=6;

	fnpos = checkpos;
	while (*checkpos != '|')
	{
		if (!isprint(*checkpos))	/* control character or sth like that? */
			return FALSE;
		checkpos++;
		filenamelength++;
	}

	if (filenamelength==0 && filenamelength > NAME_MAX)
		return FALSE;

	checkpos++;

	sizepos = checkpos;
	while (*checkpos != '|')
	{
		if (!isdigit(*checkpos))
			return FALSE;
		checkpos++;
		sizelength++;
	}

	if (sizelength==0 && sizelength > 10)
		return FALSE;

	checkpos++;

	hashpos = checkpos;
	while (*checkpos != '|')
	{
		if (!isxdigit(*checkpos))
			return FALSE;
		checkpos++;
		hashlength++;
	}

	if (hashlength!=32)
		return FALSE;

	if (fn)
	{
		*fn = g_new(gchar, filenamelength+1);
		g_return_val_if_fail(fn!=NULL,FALSE);
		memcpy (*fn, fnpos, filenamelength*sizeof(gchar));
		*((*fn)+filenamelength) = 0x00;
	}

	if (size)
	{
		*size = atoi(sizepos);
	}

	if (hashstring)
	{
		*hashstring = g_new(gchar, 33);
		g_return_val_if_fail (hashstring!=NULL,FALSE);
		memcpy (*hashstring, hashpos, 32*sizeof(gchar));
		*((*hashstring)+32) = 0x00;
	}

	return TRUE;
}


/******************************************************************************
 *
 *   misc_get_sorted_records_list_from_clist_selection
 *
 *   Returns a GSList with the data records belonging to
 *    the selected rows in the specified clist. The list
 *    is returned in REVERSED order, ie. records at the
 *    end of the data list(array) come first.
 *
 *   The sort function should sort for normal order though.
 *
 */

GSList *
misc_get_sorted_records_list_from_clist_selection (GList *selection, GtkCList *clist, GCompareFunc func)
{
	GSList	*recs = NULL;
	GList	*node = selection;

	g_return_val_if_fail (clist!=NULL, NULL);
	g_return_val_if_fail (func!=NULL, NULL);

	while (node)
	{
		guint row = GPOINTER_TO_UINT(node->data);
		gpointer rec = gtk_clist_get_row_data (clist,row);
		if (rec)
		{
			recs = g_slist_insert_sorted (recs, rec, func);
		}
		node = node->next;
	}

	return g_slist_reverse(recs);
}


