//
//       notebook.c              (part of the edonkey2000 linux gui)
//
//       (c) 2001-2002 Tim-Philipp Muller <t.i.m@orange.net>
//
//
//       functions to
//         - create a notebook window
//         - add a page to the notebook
//
//


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <time.h>

#include "global.h"
#include "status_page.h"
#include "downloads_data.h"
#include "downloads_page.h"
#include "servers_page.h"
#include "options_page.h"
#include "search_page.h"
#include "shared_dirs.h"
#include "shared_files.h"
#include "shared_page.h"
#include "stats.h"
#include "uploads_page.h"
#include "gui_options.h"
#include "notebook.h"
#include "colors.h"
#include "icons.h"
#include "misc.h"
#include "gui_flags.h"
#include "guicoreconn.h"
#include "mainwindow.h"


// global variables

int current_page = 0;

GtkWidget *notebook;		// the notebook

// a list of the text label widgets, so we can switch them on/off
static GSList	*notebook_tab_labels = NULL;


// notebook_show_hide_tab_label_text
//
// show/hide tab text according to 'option_main_notebook_tabs_show_text'

void
notebook_show_hide_tab_label_text (void)
{
	if (notebook_tab_labels)
	{
			PROCESS_GSLIST( notebook_tab_labels, GtkWidget,
				if ((item) && GTK_IS_WIDGET(item))
				{
					if (option_main_notebook_tabs_hide_text==1
						&& (option_main_notebook_tabs_pos==GTK_POS_LEFT
							|| option_main_notebook_tabs_pos==GTK_POS_RIGHT))
					{
						gtk_widget_hide(item);
					} else gtk_widget_show(item);
					if (item->parent)	// parent is the hbox
						gtk_container_check_resize(GTK_CONTAINER(item->parent));
				}
			);
			// make sure tabs are really re-sized
			gtk_notebook_set_page (GTK_NOTEBOOK(notebook),-1);	// last page
			gtk_notebook_set_page (GTK_NOTEBOOK(notebook),NOTEBOOK_PAGE_OPTIONS); // and back
	}
}


// notebook_page_switched
//
// Event that occurs when a different page is now the focus
//
// (don't use 'widget' or 'page' here - if you do, you need to change receive_options())

void
notebook_page_switched (GtkWidget *widget, GtkNotebookPage *page, gint page_num)
{

	last_time_of_user_click = time(NULL);

	current_page = page_num;

//	g_print ("DEBUG: current_page = page_num = %u\n", page_num);

	switch (current_page)
	{
		case NOTEBOOK_PAGE_DOWNLOADS:
		{
			if (clist_downloads)	/* ... not on page creation */
			{
				downloadsclist_update_status_all ();
			}
			print_download_stats_line ();
		}
		break;

		case NOTEBOOK_PAGE_UPLOADS:
		{
			uploadsclist_update();
			uploadstatsclist_update();
			print_upload_stats_line ();
		}
		break;

		case NOTEBOOK_PAGE_OPTIONS:
		{
			/* make sure that if fields are changed, but not
			 * confirmed with OK, the old values reappear
			 */
			options_update_option_widgets();
		}
		break;

		case NOTEBOOK_PAGE_SERVERS:
		{
			if (CORE_IS_ALIVE)
				servers_show_dead_alive_stats();
		}
		break;

		case NOTEBOOK_PAGE_SHARED:
		{
			/* hide path entry widget hbox at bottom if on local host (then we have dialogs) */
			if (CORE_IS_LOCAL)
			{
				gtk_widget_hide (shared_hbox);
			} else gtk_widget_show (shared_hbox);

			/* we don't want to replace the error message that's already there */
			if (CORE_IS_ALIVE)
			{
				statusbar_extra_message (_(" You are sharing %d files in %d directories."),
						(sharedfiles) ? sharedfiles->len : 0,
						(shareddirs) ? shareddirs->len : 0);
			}
			gtk_widget_grab_focus (clist_files);
		}
		break;

		default:
		break;
	}
}


// notebook_add_page
//
// Add a page to the notebook
//
// notebook - existing notebook
// szName - name to give to the new page (NULL if we don't want a frame)
//

static void
notebook_add_page (GtkWidget *notebook, GtkWidget *tabwidget, GtkWidget *pagewidget)
{
	g_return_if_fail (notebook   != NULL);
	g_return_if_fail (tabwidget  != NULL);
	g_return_if_fail (pagewidget != NULL);

	if (option_nogreen == 0)
	{
		gtk_widget_set_style (pagewidget, style_notebook_page);
		gtk_widget_set_style (notebook, style_notebook_page);
	}

	gtk_notebook_append_page (GTK_NOTEBOOK (notebook), pagewidget, tabwidget);
}



/* notebook_create_tab_content
 *
 * Creates a hbox with an icon and a label
 *  and saves the label widget, so we can hide
 *  the labels when the tabs are on the side of
 *  the notebook.
 *
 */

static GtkWidget *
notebook_create_tab_content (gchar *tabtext, gchar **icon)
{
	GtkWidget *hbox;

	g_return_val_if_fail (icon!=NULL,NULL);
	g_return_val_if_fail (tabtext!=NULL,NULL);

	hbox = gtk_hbox_new(FALSE,0);
	if (hbox)
	{
		GtkWidget *pix, *label;

		pix = create_widget_from_xpm (window, icon);
		gtk_widget_show(pix);
		gtk_box_pack_start (GTK_BOX(hbox), pix, FALSE,FALSE,0);

		label = gtk_label_new (tabtext);
		gtk_widget_show(label);
		gtk_box_pack_start (GTK_BOX(hbox), label, FALSE,FALSE,0);

		/* We want to save the notebook tab label widgets so
		 *	we can switch them on or off when the notebook tabs are rotated
		 */

		notebook_tab_labels = g_slist_append (notebook_tab_labels, (gpointer)label);
	}

	if (option_nogreen == 0)
		set_style_recursively (hbox, style_notebook_label);

	return hbox;
}



/*
 * notebook_create
 *
 * Create a new notebook and add pages to it.
 *
 * window - window to create the notebook in.
 */

GtkWidget *
notebook_create (GtkWidget *window, GtkPositionType pos)
{
	GtkWidget	*tab, *page;

	notebook = gtk_notebook_new ();
	gtk_notebook_set_scrollable (GTK_NOTEBOOK(notebook),TRUE);

	/* listen for the switch page event */
	gtk_signal_connect (GTK_OBJECT (notebook), N_("switch_page"),
			  GTK_SIGNAL_FUNC (notebook_page_switched), NULL);

	/* set position of tabs */
	gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), pos);

	/* give notebook a border */
	gtk_container_border_width (GTK_CONTAINER (notebook), 2);

	/* fill notebook with pages */
	page = downloads_page_create();
	tab  = notebook_create_tab_content (_(" Downloads "), icon_download);
	notebook_add_page (notebook, tab, page);

	page = uploads_page_create();
	tab  = notebook_create_tab_content (_(" Uploads "), icon_upload);
	notebook_add_page (notebook, tab, page);

	page = search_page_create();
	tab  = notebook_create_tab_content (_(" Search "), icon_search);
	notebook_add_page (notebook, tab, page);

	page = servers_page_create();
	tab  = notebook_create_tab_content (_(" Servers "), icon_servers);
	notebook_add_page (notebook, tab, page);

	page = status_page_create();
	tab  = notebook_create_tab_content (_(" Status "), icon_status);
	notebook_add_page (notebook, tab, page);

	page = options_page_create();
	tab  = notebook_create_tab_content (_(" Options "), icon_options);
	notebook_add_page (notebook, tab, page);

	page = shared_page_create();
	tab  = notebook_create_tab_content (_(" Shared "), icon_folder);
	notebook_add_page (notebook, tab, page);

	page = stats_create_page();
	tab  = notebook_create_tab_content (_(" Statistics "), icon_stats);
	notebook_add_page (notebook, tab, page);

	notebook_show_hide_tab_label_text ();

	// shortly switch to status page once in order to avoid gdkwindow assertion
	// 	failure on first ed2k-link write to it (don't ask me why), then switch
	//	back to first page
	gtk_notebook_set_page (GTK_NOTEBOOK(notebook), NOTEBOOK_PAGE_STATUS);
	gtk_notebook_set_page (GTK_NOTEBOOK(notebook), 0);
	return notebook;
}

