/***************************************************************************
         parse_message.c  -  parses the message packets from the core

                             -------------------
    begin                : June 2001
    copyright            : (C) 2001-2002 by Tim-Philipp Mller
    email                : t.i.m at orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// TODO:
//
// * If 'download corrupted' message comes up, find the file record from the filename
//	and set the 'corrupted' flag instead of the way we do it now.

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include "guicoreconn.h"
#include "parse_message.h"
#include "packet.h"
#include "gui_core_protocol.h"
#include "send_command.h"
#include "http_get.h"
#include "tag.h"

#include "blacklist.h"
#include "connect_gui.h"
#include "gaps.h"
#include "mainwindow.h"
#include "misc_gtk.h"
#include "notebook.h"
#include "search_page.h"
#include "servers_data.h"
#include "servers_page.h"
#include "shared_dirs.h"
#include "shared_files.h"
#include "shared_page.h"
#include "stats.h"
#include "statusbar.h"
#include "status_page.h"
#include "toolbar.h"
#include "uploads_page.h"
#include "uploads_data.h"
#include "uploadstats_data.h"
#include "downloads_attrib.h"
#include "downloads_data.h"
#include "downloads_page.h"
#include "options_page.h"
#include "gui_options.h"
#include "global.h"
#include "misc.h"
#include "notebook.h"
#include "gui_flags.h"

/* global variables */

//gint timer_avoid_deadlock_on_connect = 0;	// see do_message / core_status_msg

int  update_space_stat = 0;		// update after next status packet
int  just_asked_for_space_stat = 0;	// flag we set to avoid message output when we send 'space' to get space stats
int  just_asked_for_peers_stat = 0;	// flag we set to avoid message output when we send 'vp' to get connection stats


// local variables

// so we don't redraw the clist after every result => faster on old comps when 1000s of results come in
//static time_t	last_extended_search_result_received = (time_t) 0;
//static guint	no_of_extended_search_results_received = 0;

static guint	print_next_clientID = 0;	// only the first after connect to status log

static time_t	 last_connecting = 0;

//char *last_connecting_attempt = NULL;	// string with last 'connecting to...' message

/* last time we RECEIVED a client status packet with the space free/needed status */
static time_t 		space_stats_last_updated = 0;

/* The core doesn't tell us other than by sending a log message (grrrrrr @ Jed) */
static gboolean		last_status_message_was_download_corrupted = FALSE;

/* We need this outside of the function, so the p_m_receive_dl_status_list()
 *  routine can call the stats register function with the up-to-date UL speed
 *  value at hand. We need to pass that value to the DL status function somehow.
 */
static gfloat		total_ul_speed = 0.0;


/* a flag that indicates that we've received options correctly
 * after detecting that we are dealing with overnet */
static gboolean		parse_message_received_overnet_options = FALSE;

/* locoal functions */

static void			parse_message_receive_dl_status_list (void);
static void			parse_message_receive_ul_status_list (void);
static gfloat		parse_message_receive_dl_status_items (guint items);
static void			parse_message_check_whether_gaps_need_updating (void);
static void			parse_message_receive_options (void);


/******************************************************************************
 *
 *  parse_message_set_overnet_mode
 *
 */

void
parse_message_set_overnet_mode (gboolean overnet)
{
	if (!overnet)
	{
		parse_message_received_overnet_options = FALSE;
		return;
	}

	/* If we haven't received the options properly,
	 * get the options again, this time correctly
	 * (knowing that we are dealing with overnet)
	 * (overnet sends options slighty different)
	 *
	 * parse_message_receive_options() will set this
	 * flag to TRUE once it received options while
	 * in overnet mode.
	 */
	if (!parse_message_received_overnet_options)
	{
		send_get_options();
	}
}


/******************************************************************************
 *
 *  parse_message_check_whether_gaps_need_updating
 *
 *  gets new progress bar stats every x calls, with
 *  x depending on whether the GUI-core connection
 *  is configured to be a slow connection.
 */

static void
parse_message_check_whether_gaps_need_updating (void)
{
	static guint callcounter = 0;  
	guint        everyx;

	everyx = (option_slowconnection == 1) ? 10: 5;

	if (callcounter % everyx == 0)
	{
		get_gaps_for_visible_downloads();
	}

	callcounter++;
}


/******************************************************************************
 *
 *  parse_message_receive_dl_status_items
 *
 *  returns total download speed
 *
 */

static gfloat
parse_message_receive_dl_status_items (guint items)
{
	guint  i;
	gfloat total_dl_speed = 0.0;

	for (i=0; i<items; i++)
	{
		guint16		 index   = read_16();
		guint8		 status  = read_8();
		gfloat		 speed   = read_float();
		guint32		 trans   = read_32();
		guint8		 avail   = read_8();
		guint8		 sources = read_8();

		download_set_stats (index, status, speed, trans, avail, sources);

		/* Note: when a corruption occurs, we might get astronomically high
		 * speed values for a couple of seconds. ignore those for the total */

		if (speed>2048.0)
		{
			g_printerr (_("ignored out-of-bounds total download speed (>2MB/s) for statistics\n"));
		}
		else total_dl_speed += speed;
	}

	return total_dl_speed;
}


/******************************************************************************
 *
 *  parse_message_receive_dl_status
 *
 *  Receives an parses a download status packet from the core.
 *
 *  Note: we have to keep track of any file priority changes
 *        ourselves. With CORE_NEW_DOWNLOAD we only receive the
 *        'initial' priority state, but no updates after that.
 *
 *  Note: It is possible that the core sends a status packet on connect by
 *        the GUI before we have received the CORE_NEW_DOWNLOAD messages.
 *        (those are being ignored then) 
 */

static void
parse_message_receive_dl_status_list (void)
{
	static time_t	last_status_packet = 0;
	gint         	timediff = UPLOAD_STAT_PACKET_INTERVAL_IN_S;
	gfloat       	total_dl_speed = 0.0;

	if (last_status_packet > 0)
	{
		timediff = CLAMP ((time(NULL)-last_status_packet), 2, 4);
	}

	last_status_packet = time(NULL);

	if (downloads_set_all_to_no_stats_received())
	{
		total_dl_speed = parse_message_receive_dl_status_items( read_16() );

		if (total_dl_speed > 0.0)
		{
			seconds_download_was_nonzero += timediff;
			total_downloaded +=  total_dl_speed * ((gfloat) timediff);
/*XXX*/
			if (total_dl_speed>2048.0)
			{
				g_printerr (N_("DEBUG: %s:%u - we shouldn't be called.\n"), __FUNCTION__,__LINE__);
			}
		}

		downloads_check_for_stats_received();
		parse_message_check_whether_gaps_need_updating();
	} 

	/* Upload stat packet is sent by core before download 
	 *  stat packet, that's why we call the function here */
	stats_register_upload_download_stats_packet ( time(NULL),
	                                              total_ul_speed,
	                                              total_dl_speed,
	                                              0, 0);

	if (update_space_stat == 1)
	{
		/* so if we delete/pause etc. and have many downloads
		 * we don't want this too often in a row, as we
		 * get a space needed line for every download
		 * with older cores */
		if ((time(NULL)-space_stats_last_updated)>30
				|| downloads_get_num()<50
				|| core_build_date >= CORE_BUILD_1)
		{
			send_get_space_stats();
		}
	}

	downloadsattriblist_update_from_downloadlist();
	downloadsclist_resort ();
	free_oldpixmaps ();
	toolbar_set_dl_speed (total_dl_speed);
}



/* parse_message_receive_ul_status_list
 *
 *
 *
 */

static void
parse_message_receive_ul_status_list (void)
{
	static time_t	lastpacket = 0;
	gfloat			total_ul_inc;
	guint			timediff;

	timediff = (guint)(time(NULL)-lastpacket);
	if (timediff>5)
		timediff = 3;

	total_ul_speed = 0.0;

	if (uploads)
	{
		guint i, num = read_16();
		for (i=0; i<num; i++)
		{
			guint indexpos = read_16();	// item number
			if (indexpos < uploads->len)
			{
				uploadRecord	*ul = g_ptr_array_index (uploads, indexpos);
				gfloat			 ulspeed = read_float();
				if (ul)
				{
					ul->speed = ulspeed;
					// if we have a hash for this upload record, update total uploaded
					//	value (otherwise it doesn't make sense, because we couldn't
					//	find the upload stats record).
					if (uploads_check_record_completeness(ul))
					{
						uploadstats_add_or_increase_record (ul, timediff);
					}
				}
				total_ul_speed += ul->speed;
			} else g_printerr (N_("%s: %s"),  __FUNCTION__, _("received UL status for a file which is not in list (bug?)\n"));
		}
	}

	if (total_ul_speed > 0.0)
		seconds_upload_was_nonzero += timediff;

	/* save cpu power - don't refresh list if it's not visible */
	if (current_page == NOTEBOOK_PAGE_UPLOADS)
	{
		uploadsclist_update();
		uploadstatsclist_update ();
	}

	lastpacket = time(NULL);

	toolbar_set_ul_speed (total_ul_speed);

	total_ul_inc = (timediff*1.0) * total_ul_speed;

	/* more than 10000 kB/s = ca. 10 MB/s upload? unlikely */
	g_return_if_fail (total_ul_inc < 10000.0);

	total_uploaded += total_ul_inc;
}


/******************************************************************************
 *
 *  parse_message_receive_options
 *
 */

static void
parse_message_receive_options (void)
{
	gint   dummy;
	gchar  cdummy[1024];
	gchar  cbdstr[128];	// core build date string
	gchar *dummystr;

	/* we set this to 0 again here, because we set it
	 * to 1 when we try to do autoconnect at startup,
	 * so we don't get two donkeys if initial spawn
	 * worked, but not the connect. */
	connection_closed_deliberately = 0;

	coreversion = read_16();				/* core client version 	*/

	option_max_dl_speed = 1.0 * int_from_float_read();
	option_max_ul_speed = 1.0 * int_from_float_read();

	option_doorport = read_16();

	dummy = read_16();						/* max connections (ooops, double)*/

	dummystr = read_i4str();
	g_snprintf(option_nickname, sizeof(option_nickname), N_("%s"), dummystr);
	g_free(dummystr);

	dummystr = read_i4str();
	g_snprintf(option_temporary, sizeof(option_temporary), N_("%s"), dummystr);
	g_free(dummystr);

	dummystr = read_i4str();
	g_snprintf(option_incoming, sizeof(option_incoming), N_("%s"), dummystr);
	g_free(dummystr);

	if (!IS_OVERNET)
	{
		guint coreversion_major = coreversion%100;
		option_always_connected = read_8();
		option_autoremove       = read_8();

		/* new cores (v43/v44) have an extra byte here. And we will always get
		 * 0 for the above options (bug in guicontroller.cpp)
		 */
		if (coreversion_major == 43 || coreversion_major == 44)
		{
			static gboolean msgflag = FALSE;
			if (!msgflag)
			{
				status_message_blue(_("GUI: For you information: this core does not tell us the\n"
				                      "'always stay connected' and 'automatically remove dead servers'\n"
				                      "options properly (core bug), therefore there will be harmless\n"
				                      "GUI-core inconsistencies.\n"));
				msgflag = TRUE;
			}
			dummy = read_8();
			
			/* if we don't get them from the core, set them to the most useful values */
			option_always_connected = 1;
			option_autoremove = 1;
		}
	}
	else
	{
		dummy = read_8();
		dummy = read_8();
		dummy = read_8();
		parse_message_received_overnet_options = TRUE;
	}

	option_allowpm         = read_8();
	option_save_corrupted  = read_8();
	dummy                  = read_8();	/* verify cancel - we ignore this and use our own */
	option_adminport       = read_16();
	option_max_connections = read_32();

	core_build_date = (gint) read_32();	/* core build date */

	/* old core or new core, that is the question */
	if (core_build_date == -1)					
	{
		core_build_date = 0;
		cbdstr[0] = 0x00;
	}
	else
	{
		g_snprintf (cbdstr, sizeof(cbdstr), N_(" (build %02u%02u%04u)"), core_build_date&0xff, (core_build_date&0xff00)>>8, ((core_build_date&0xff0000)>>16)+2000);
		/* if newer core with extra options */
		option_line_down_speed = int_from_float_read();		// line down speed
	}

	dummy = read_32();					// PID
	if (dummy>1)
	{
		local_donkey_pid = dummy;			// user real PID - but only if core supports this
	}							// otherwise use the one guessed via /proc system or none


	/* Return here if this is not the very first time that we got the options.
     *  After all, the core version won't change in one and the same GUI-core session, etc.
	 *
	 */
	if (first_time_we_get_options != 1)
	{
		return;
	}

	if ((IS_OVERNET) && core_build_date>0 && (core_build_date<=OVERNET_CORE_TOO_OLD || core_build_date>=CORE_TOO_FAR_IN_THE_FUTURE))
	{
		ok_dialog (N_("overnet version too old"),
		           N_("This version of the overnet core won't work properly\n"
				      "with this GUI. It will not return any search results\n"
			          "to the GUI and is as such pretty useless.\n\n"
					  "I'm afraid you will need to wait for the next overnet\n"
					  "version before you can use this GUI with it.\n\n"
					  "You should have no problems using this overnet core\n"
					  "in command line mode without a GUI though."));
	}

	if (1==1)
	{
		gchar *verstr = NULL;	/* all version stuff together        */
		gchar *cpidstr = NULL;	/* core pid string (if on localhost) */

		if (local_donkey_pid>0)
		{
			const gchar *bn =(CORE_IS_LOCAL) ? misc_get_basename_from_pid(local_donkey_pid,NULL) : N_("");
			gchar *binname = (CORE_IS_LOCAL) ? g_strdup_printf (N_(" - '%s'"), (bn!=NULL) ? bn : N_("???")) : g_strdup(N_(""));
			cpidstr = g_strdup_printf (N_(" (%spid=%u%s)"),
										(CORE_IS_LOCAL) ? N_("") : _("remote "),
										local_donkey_pid, binname);
			g_free(binname);
		} else cpidstr = g_strdup_printf (_(" (%s, pid not available)"), (CORE_IS_LOCAL) ? _("local") : _("remote"));

		verstr = g_strdup_printf (N_("%s: %u%s%s"), _("Core version"),
									coreversion, cbdstr, cpidstr);

		toolbar_set_core_version_string (verstr);
		
		if ( coreversion < MIN_CORE_VERSION)
		{
			ok_dialog (_("notification"), _("The core you are currently connected to\n"
						   "seems to be an older version. You will\n"
						   "not be able to use all features of this\n"
						   "GUI with this core.\n"
						   "Go to the GUI homepage\n"
						   "http://ed2k-gtk-gui.sourceforge.net\n"
						   "to get a newer version.\n"));
		}

		first_time_we_get_options = 0;

		linkhandler_check_for_pending_links();
		
		if (cpidstr)
			g_free(cpidstr);
		if (verstr);
			g_free(verstr);
	}
}

int timer_get_new_serverlist_10s_after_connect (gpointer data)
{
        send_get_serverlist();
        return 0;       /* we want to be called only once */
}


// we want to move all error items to the back, because the core doesn't send stats for them (changing the item
// numbering!)
// hmm, I'm not sure if this is true, let's not do it for now.

void downloadlist_move_error_items_to_the_back(void)
{
	;
}



// parse_status_message_from_core
//
// Often the core only tells us certain things via a status text message, in other
//	cases we want to explain some messages or act on certain messages. Do all
//	that here.
//
// Returns TRUE if the message is to be shown to the user, otherwise FALSE

enum
{
CORE_STATUS_MSG_VP_CONNECTIONS=0,
CORE_STATUS_MSG_CANTCONNECT,
CORE_STATUS_MSG_CONNECTINGTO,
CORE_STATUS_MSG_PLACEDONCONNQ,
CORE_STATUS_MSG_DONELOADINGFILES,
CORE_STATUS_MSG_CORRUPTEDDOWNLOAD,
CORE_STATUS_MSG_IMPORTFAILED,
CORE_STATUS_MSG_OUTOFDATE,
CORE_STATUS_MSG_CLIENTID,
CORE_STATUS_MSG_OVERNETID,			/* this must come after the client ID */
CORE_STATUS_MSG_USERS,
CORE_STATUS_MSG_SPACE_TEMP,
CORE_STATUS_MSG_SPACE_IN,
CORE_STATUS_MSG_SPACE_NEEDED,
CORE_STATUS_MSG_NOTALLWRITTEN,
CORE_STATUS_MSG_NOMORETOEXTENDTO,
CORE_STATUS_MSG_YOURIDHASCHANGED,
CORE_STATUS_MSG_LOADINGSHAREDFILES,
CORE_STATUS_MSG_NOTCONNECTEDTOASERVER,
CORE_STATUS_MSG_ALREADYCONNECTED,
CORE_STATUS_MSG_ERRORLOWID,
CORE_STATUS_MSG_OVERNETOPEN,
CORE_STATUS_MSG_OVERNETFW,
CORE_STATUS_MSG_OVERNETUNCLEAR,
CORE_STATUS_MSG_BRACKET, // should be LAST!
CORE_STATUS_MSG_LAST,
} ;

static const gchar *coreStatMsgs[CORE_STATUS_MSG_LAST] =
{
N_("Connections"),
N_("Can't connect to"),
N_("Connecting to"),
N_("Placed on connection Queue"),
N_("Done loading shared"),
N_("Corrupted Download:"),
N_("Import failed"),
N_("server is out of date"),	/* Roman Hodek: my core says "your client" not "the client" */
N_("ClientID"),
N_("ID: "),						/* this must come after the client ID (substring) */
N_("Users:"),	// from 'g' command - we want to suppress output of that (only need client ID)
N_("on temp part"),
N_("on incoming part"),
N_("needed by downl"),
N_("ot all written"),
N_("No more servers to extend to"),
N_("Your ID has changed"),
N_("Loading shared files"),
N_("Not connected to a server"),
N_("Already Connected"),
N_("not reachable. You have a LOWID"),
N_("(open)"),
N_("(firewalled)"),
N_("(unclear)"),
N_("(")		// the opening bracket of a peer from 'vp' or of individual DL from 'space'
} ;


/* parse_status_message_from_core
 *
 * returns FALSE if message should NOT be shown to user/put in statuslog
 *
 */

static gboolean
parse_status_message_from_core (const gchar *msg)
{
	guint			 msgnum;
	gchar			*tab;
	static gchar	*last_overnet_status = NULL;
	static gchar	*incomingSpaceStr = NULL, *neededSpaceStr = NULL, *tempSpaceStr = NULL;
	static gboolean	 lastMsgWasLoadingSharedFiles = FALSE;

	g_return_val_if_fail (msg!=NULL, FALSE);

	// For space needed, free, etc. messages:
	tab = strrchr (msg, '\t');
	if (!tab) tab = strrchr (msg, ' ');

	for (msgnum=0; msgnum < (sizeof(coreStatMsgs)/sizeof(gchar*)); msgnum++)
	{
		if (strstr(msg, coreStatMsgs[msgnum])!=NULL)
			break;
	}

	switch (msgnum)
	{
		case CORE_STATUS_MSG_VP_CONNECTIONS:
		{
			// Note: just_asked_for_peers_stat is not boolean, but 0 if not,
			//	1 if yes and not gotten first line, and >1 if yes and
			//	gotten first line (so we can ignore the peers listed)
			if (just_asked_for_peers_stat==1)
			{
				guint cons_now = atoi(msg);
				(void) toolbar_set_cons (cons_now);
				return FALSE;
			}
		}
		break;
		case CORE_STATUS_MSG_BRACKET:
		{
			// Note: just_asked_for_peers_stat is not boolean, but 0 if not,
			//	1 if yes and not gotten first line, and >1 if yes and
			//	gotten first line (so we can ignore the peers listed)
			// (Also, we only have a peer if the FIRST character is a '('
			if (*msg=='(' && just_asked_for_peers_stat>0)
			{
				just_asked_for_peers_stat++;
				return FALSE;	// ignore peers output
			}
			if (*msg=='(' && just_asked_for_space_stat>0)
			{
				just_asked_for_space_stat++;
				return FALSE;	// ignore individual DLs space needed ouput
			}
			/* Roman Hodek: just_asked_for_space_stat is cleared too often IMHO...
			 * Recognize indiv. space messages by '(' _and_ ":\t"
			 */
			if (*msg=='(' && strstr(msg, N_(":\t")))
				return FALSE;
		}
		break;
		case CORE_STATUS_MSG_CANTCONNECT:
		{
			servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_CANTCONNECT);
			toolbar_set_server_data (NULL, 0, 0);
			last_connecting = 0;
		}
		break;
		case CORE_STATUS_MSG_CONNECTINGTO:
		{
			// Find server name or IP from whole line
			gchar *ipOrName = g_strdup(msg+14);
			gchar *lastbit = strstr(ipOrName, N_(" ..."));
			if (lastbit) *lastbit=0x00;
			servers_register_connect_status (ipOrName, SERVERS_CONNECT_STATUS_CONNECTING);
			toolbar_set_server_data (_(" [connecting] "), -1, -1);
			g_free(ipOrName);
			last_connecting = time(NULL);
		}
		break;
		case CORE_STATUS_MSG_PLACEDONCONNQ:
		{
			/* FIX ME: why no servers_register_connect_status() here? */
			toolbar_set_server_data (_(" [on connection queue] "), -1, -1);
		}
		break;
		case CORE_STATUS_MSG_DONELOADINGFILES:
			send_get_shared_files();
			send_get_shared_dirs();
			shared_page_set_stats_label_message (_("Done loading shared files."));
		break;
		case CORE_STATUS_MSG_CORRUPTEDDOWNLOAD:
			last_status_message_was_download_corrupted = TRUE;
		break;
		case CORE_STATUS_MSG_IMPORTFAILED:
			status_msg (N_("%s.\n"), _("GUI: This might be because you only have read, but no write permission on that file"));
		break;
		case CORE_STATUS_MSG_OUTOFDATE:
			status_msg (N_("%s.\n"), _("GUI: Server is rejecting us. *pah*"));
			servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_THINKS_WE_ARE_FIREWALLED);
		break;

		case CORE_STATUS_MSG_CLIENTID:
		{
			servers_page_set_client_id(msg);
			if (print_next_clientID==1)
				status_msg(N_("%s"), msg);
			if (option_disconnect_if_firewalled==1 && strlen(msg)>10)
			{
				gint id = atoi(msg+10);
				if (id>0 && id<0xffffff)
				{
					status_msg (N_("%s"), msg);
					status_msg (N_("%s.\n"), _("GUI: Server thinks you are firewalled. Disconnecting."));
					servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_THINKS_WE_ARE_FIREWALLED);
				}
			}
			print_next_clientID=0;
			just_asked_for_peers_stat=0;
		}
		return FALSE;

		case CORE_STATUS_MSG_OVERNETID:
		{
			gchar overnetid[33];
			static gboolean overnetmsg_printed = FALSE;
			if (sscanf(msg, N_("ID: %32[0-9a-fA-F]"), overnetid))
			{
				gchar *fullmsg = g_strdup_printf (N_("%s%s"), msg, (last_overnet_status!=NULL) ? last_overnet_status : N_(""));

				servers_page_set_client_id(fullmsg);

				if (print_next_clientID==1)
					status_msg(fullmsg);

				if (!overnetmsg_printed)
				{
					status_message_blue(_("GUI: This seems to be an OVERNET core (=not fully supported yet)\n"));
					mainwindow_set_title (N_("overnet"));
					toolbar_set_overnet_mode(TRUE);
					serversclist_set_overnet_mode(TRUE);
					parse_message_set_overnet_mode(TRUE);
					overnetmsg_printed = TRUE;
				}
				core_is_overnet = TRUE;
			}
			print_next_clientID=0;
			just_asked_for_peers_stat=0;
		}
		return FALSE;

		case CORE_STATUS_MSG_YOURIDHASCHANGED:	/* just supress this */
			just_asked_for_peers_stat=0;
			just_asked_for_space_stat=0;
		return FALSE;

		case CORE_STATUS_MSG_USERS:	
		{
			just_asked_for_peers_stat=0;
			just_asked_for_space_stat=0;

			if (IS_OVERNET)
			{
				guint users, files;
				if (sscanf(msg, N_("Users: %u   Files: %u"), &users, &files)==2)
				{
					toolbar_set_server_data (N_(" [overnet] "), files, users);
				}
			}
		}
		return FALSE;

		case CORE_STATUS_MSG_SPACE_TEMP:
			if (tab)
			{
				if (tempSpaceStr) g_free(tempSpaceStr);
				tempSpaceStr = g_strdup(tab+1);
				g_return_val_if_fail (tempSpaceStr!=NULL, FALSE);
				if (tempSpaceStr[strlen(tempSpaceStr)-1]=='\n')
					tempSpaceStr[strlen(tempSpaceStr)-1]=0x00;
				just_asked_for_peers_stat=0;
				if (just_asked_for_space_stat) return FALSE;
			}
		break;
		case CORE_STATUS_MSG_SPACE_IN:
			if (tab)
			{
				if (incomingSpaceStr) g_free(incomingSpaceStr);
				incomingSpaceStr = g_strdup(tab+1);
				g_return_val_if_fail (incomingSpaceStr!=NULL, FALSE);
				if (incomingSpaceStr[strlen(incomingSpaceStr)-1]=='\n')
					incomingSpaceStr[strlen(incomingSpaceStr)-1]=0x00;
				just_asked_for_peers_stat=0;
				if (just_asked_for_space_stat) return FALSE;
			}
		break;
		case CORE_STATUS_MSG_SPACE_NEEDED:
			if (tab)
			{
				int old_jafss = just_asked_for_space_stat; /* Roman Hodek: store it! */
				if (neededSpaceStr) g_free(neededSpaceStr);
				neededSpaceStr = g_strdup(tab+1);
				g_return_val_if_fail (neededSpaceStr!=NULL, FALSE);
				if (neededSpaceStr[strlen(neededSpaceStr)-1]=='\n')
					neededSpaceStr[strlen(neededSpaceStr)-1]=0x00;
				just_asked_for_peers_stat=0;
				just_asked_for_space_stat=0;
				// last entry we want - now print new label
				if ((neededSpaceStr) && (tempSpaceStr) && (incomingSpaceStr))
				{
					gboolean	conversion_error = FALSE;
					gfloat		needed= -1.0;
					gfloat		temp = -1.0;
					gfloat		incoming = -1.0;
					
#ifdef HAVE_STRTOF
					needed = strtof (neededSpaceStr, (gchar**)NULL);
					if (errno == ERANGE)
						conversion_error = TRUE;
#else
					needed = atof (neededSpaceStr);
#endif
#ifdef HAVE_STRTOF
					temp = strtof (tempSpaceStr, (gchar**)NULL);
					if (errno == ERANGE)
						conversion_error = TRUE;
#else
					temp = atof (tempSpaceStr);
#endif

#ifdef HAVE_STRTOF
					incoming = strtof (incomingSpaceStr, (gchar**)NULL);
					if (errno == ERANGE)
						conversion_error = TRUE;
#else
					incoming = atof (incomingSpaceStr);
#endif

					if (!conversion_error)
					{
						downloads_page_set_free_space (needed, temp, incoming);
						if (temp <= (option_minfreetempspace*1.0))
						{
							downloads_pause_all(temp);
						}
					} else downloads_page_set_free_space (-1.0, -1.0, -1.0);
				}
				if (old_jafss) return FALSE;
			}
		break;

		case CORE_STATUS_MSG_NOTALLWRITTEN:
			g_printerr (_("The above error _might_ be due to a full harddisk.\n"));
		break;

		case CORE_STATUS_MSG_NOMORETOEXTENDTO:
			statusbar_extra_message (N_(" %s"), msg);
		return FALSE;

		case CORE_STATUS_MSG_LOADINGSHAREDFILES:
		{
			if (lastMsgWasLoadingSharedFiles)
				return FALSE;
			lastMsgWasLoadingSharedFiles = TRUE;
			shared_page_set_stats_label_message (msg);
		}
		return TRUE;

		case CORE_STATUS_MSG_NOTCONNECTEDTOASERVER:
		{
			/* we must set the 'search' button sensitive again if we couldn't search
			 * because we are not connected to a server
			 */
			gtk_widget_set_sensitive(button_extend, FALSE);
			gtk_widget_set_sensitive(button_more, FALSE);
			gtk_widget_set_sensitive(button_addsearch, TRUE);
		}
		return TRUE;

		case CORE_STATUS_MSG_ALREADYCONNECTED:
		{
			return FALSE;
		}

		case CORE_STATUS_MSG_ERRORLOWID:
		{
			/* XXX - do this even if option_disconnect_if_firewalled is not set?
			 * (Problem is: whether server closes connection on lowid,
			 *  depends on server, not on core)
			 */
			if (option_disconnect_if_firewalled)
			{
				servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_CANTCONNECT);
				toolbar_set_server_data (NULL, 0, 0);
			}
		}
		return TRUE;

		case CORE_STATUS_MSG_OVERNETOPEN:
		{
			if (g_strcasecmp(msg,N_("(open)"))==0)
			{
				last_overnet_status = _(" (open) ");
			}
		}
		return FALSE;

		case CORE_STATUS_MSG_OVERNETFW:
		{
			if (g_strcasecmp(msg,N_("(firewalled)"))==0)
			{
				last_overnet_status = _(" (firewalled) ");
			}
		}
		return FALSE;

		case CORE_STATUS_MSG_OVERNETUNCLEAR:
		{
			if (g_strcasecmp(msg,N_("(unclear)"))==0)
			{
				last_overnet_status = _(" (unclear) ");
			}
		}
		return FALSE;

		default:
		break;
	}
	just_asked_for_peers_stat=0;
	just_asked_for_space_stat=0;
	lastMsgWasLoadingSharedFiles = FALSE;
	return TRUE;
}



static void
parse_message_extend_5x (void)
{
	gint j;
	for (j=1; j<6; j++)
	{
		if ((servers) && j<servers->len)
			send_extended_search();
	}
}


static void
parse_message_get_results (guint num)
{
	guint i;

	g_return_if_fail (current_search!=NULL);

	for (i=0; i<num; i++)
	{
		guint8		*servicehash = read_x( (IS_OVERNET) ? 16 : 0 );
		guint8		*hash = read_x(16);	// read unused fields too (they do no harm)
		guint8		*unused = read_x( (IS_OVERNET) ? 0 : (4+2) );
		GPtrArray	*taglist = taglist_read_from_ramfile (ramfile_now_active);
		NGS_tag		*nametag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILENAME_SHORT);
		gchar		*name = (nametag) ? (gchar*)nametag->data : N_("(???)");
		NGS_tag		*sizetag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILESIZE_SHORT);
		NGS_tag		*availtag = taglist_find_tag_from_tagname (taglist, TAGNAME_AVAILABILITY_SHORT);
		NGS_tag		*typetag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILETYPE_SHORT);
		NGS_tag		*formattag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILEFORMAT_SHORT);
		gchar		*format = (formattag) ? (gchar*)formattag->data : (gchar*) misc_get_extension_from_filename(name);
		NGS_tag		*artisttag = taglist_find_tag_from_tagname (taglist, N_("Artist"));
		NGS_tag		*titletag = taglist_find_tag_from_tagname (taglist, N_("Title"));
		NGS_tag		*albumtag = taglist_find_tag_from_tagname (taglist, N_("Album"));
		NGS_tag		*lengthtag = taglist_find_tag_from_tagname (taglist, N_("length"));
		NGS_tag		*bitratetag = taglist_find_tag_from_tagname (taglist, N_("bitrate"));
		NGS_tag		*codectag = taglist_find_tag_from_tagname (taglist, N_("codec"));
		gchar		*pos;

		if ((unused) && (!IS_OVERNET) && *((guint32*)unused)!=0 && *((guint16*)(unused+4))!=0)
		{
			g_printerr (N_("ATTENTION: Jed has changed FileMeta->toNetwork()! This needs to be fixed in the GUI.\n"));
			return;
		}

		/* remove bad characters (like a slash) from the filename */
		for (pos=name; *pos!=0x00; pos++)
			if (*pos == SLASH_CHAR) *pos = ' ';

		searchlist_add_or_update_record ( current_search, i, hash,
			(sizetag) ? GPOINTER_TO_UINT(sizetag->data) : 0,
			(availtag) ? GPOINTER_TO_UINT(availtag->data) : 0,
			name,
			(typetag) ? (gchar*)typetag->data : misc_get_filetype_from_extension (format),
			format,
			(artisttag) ? (gchar*)artisttag->data : NULL,
			(albumtag) ? (gchar*)albumtag->data : NULL,
			(titletag) ? (gchar*)titletag->data : NULL,
			(lengthtag) ? (gchar*)lengthtag->data : N_(""),
			(bitratetag) ? GPOINTER_TO_UINT(bitratetag->data) : 0,
			(codectag) ? (gchar*)codectag->data : N_(""));

/* FIX: what does the availability value in OVERNET mean??? (insanely huge values)
#warning **********************************~~~~~~~~~~~ REMOVE ME (line 964)
		if (availtag)
		{
			g_print ("availability: %12u %12d for %s\n", GPOINTER_TO_UINT(availtag->data), ((gint)GPOINTER_TO_UINT(availtag->data))*(-1), name);
		}
*/
		if (taglist) taglist_free(taglist);
		if (hash) g_free(hash);
		if (servicehash) g_free(servicehash);
		if (unused) g_free(unused);
	}

	// we only want to do this after all the links have been added, not after
	//	every single item
	if ((current_search) && (current_search->clist))
		searchclist_resort(current_search->clist);
}


// parse_message_from_client
//
// Reads a single message from the GUI and acts upon it

void
parse_message_from_client (ramfile *packet)
{
	static time_t	first_time_not_connecting = 0;
	static gboolean	we_are_connected = FALSE;

	g_return_if_fail (packet!=NULL);

	if (!window)
		return;


/* let's not do this: seems to provoke core segfaults through
 * double-connects at the same time (=> core bug):
 */
#ifdef HUMBUG

	/* Make sure we connect to a server if the option is set */
	/*  (the core is a bit sluggish in this respect)         */
	if (last_connecting==0)
	{
		if (first_time_not_connecting==0)
			first_time_not_connecting = time(NULL);
	} else first_time_not_connecting = 0;

	if ((option_always_connected) && (!we_are_connected))
	{
		if (first_time_not_connecting>0 && (time(NULL)-first_time_not_connecting)>45)
		{
			status_message_blue (N_("GUI DEBUG: connecting. NOW-first_time_not_connecting>45\n"));
			send_disconnect ();
			send_connect_to_server(0,0);
			first_time_not_connecting = 0;
		}
		else if (last_connecting>0 && (time(NULL)-last_connecting)>45)
		{
			status_message_blue (N_("GUI DEBUG: connecting. NOW-last_connecting>45\n"));
			/* XXX - should we send a disconnect here (possibly hung at 'placed on conn queue')? */
			send_connect_to_server(0,0);
			last_connecting = 0;
		}
	}

#endif
	// if we've just killed and restarted the core, we want a success message if it
	// worked - here it is
	if (show_restart_success_message_on_next_packet == 1)
	{
		status_message_blue (N_("%s.\n"), _("GUI: restart and reconnect successful"));
		show_restart_success_message_on_next_packet = 0;
	}
	
	// get and parse message
	ramfile_seek (packet, 5, RAMFILE_SEEK_FROM_START);	// skip header
	switch (read_8())
	{
		case -1:		return;	// error (packet short)

		case CORE_STATUS_MSG:
		{
			gchar *msg = read_i4str();
			last_status_message_was_download_corrupted = FALSE;
			if (msg)
			{
				if (parse_status_message_from_core(msg))
					status_msg (N_("%s\n"), msg);
				g_free(msg);
			}
		}
		break;

     		case CORE_ERROR_MSG:
		{
			gchar *msg = read_i4str();
			last_status_message_was_download_corrupted = FALSE;
			if (msg)
			{
				if (parse_status_message_from_core(msg))
					status_warning (N_("%s"), msg);
				g_free(msg);
			}
		}
		break;

		case CORE_HASHING:
		{
			gchar	*name = read_i4str();
			gchar	*txt = g_strdup_printf (_("Hashing %s ..."), name);
			g_return_if_fail (txt!=NULL);
			shared_page_set_stats_label_message(txt);
			g_free(txt);
			g_free(name);
		}
		break;

		case CORE_CONNECTED_TO:
		{
			gchar *servername = read_i4str();
			if (servername)
			{
				toolbar_set_server_data (servername, 0, 0);
				status_msg (_(" --- connected to server '%s' ---\n"), servername);
				g_free(servername);
				servername = NULL;
			}
			servers_reinstate_alwaysconnected_option();
			servers_clear_connect_to_list();
			// add timer to get a new serverlist 10s after the connect
			(void) gtk_timeout_add (10*1000, timer_get_new_serverlist_10s_after_connect, NULL);
			print_next_clientID=1;
			send_get_client_stats();
			gtk_widget_set_sensitive(button_addsearch, TRUE);
			gtk_widget_set_sensitive(button_extend, FALSE);
			gtk_widget_set_sensitive(button_more, FALSE);
			last_connecting = 0;
			we_are_connected = TRUE;
		}
		break;

		case CORE_SET_SERVER_STATS:
		{
			guint32 files = read_32();
			guint32 users = read_32();
			toolbar_set_server_data (NULL, (guint)files, (guint)users);
		}
		break;

		case CORE_DISCONNECTED:
		{
			status_msg (_(" --- disconnected from server ---\n"));
			toolbar_set_server_data (NULL,0,0);
			gtk_widget_set_sensitive(button_addsearch, FALSE);
			gtk_widget_set_sensitive(button_extend, FALSE);
			gtk_widget_set_sensitive(button_more, FALSE);
/*			can't do this, because 'placed on queue' is followed by disconnect as well
			servers_reinstate_alwaysconnected_option();
			servers_clear_connect_to_list();
*/
			last_connecting = 0;
			we_are_connected = FALSE;
			first_time_not_connecting = 0;
		}
		break;

		case CORE_NEW_SEARCH_RESULTS:
		{
			// counter, so we don't go on forever asking for more results
			static gint	requested_more_results = 0;

			guint		results, num=read_32();
			gboolean	more; // more results available?
			GString		*actstr;

			gtk_widget_set_sensitive (button_addsearch, TRUE);

			if (num==0)
			{
				statusbar_extra_message(_(" No results. Automatically extending search to five more servers..."));
				parse_message_extend_5x ();
				gtk_widget_set_sensitive (button_extend, TRUE);
				break;
			}

			g_return_if_fail (current_search!=NULL);
			g_return_if_fail (current_search->clist!=NULL);

			// Note: new searchlist and all that is done
			//	in search_page.c/add search button pressed

			gtk_clist_freeze(current_search->clist);
			parse_message_get_results (num);
			gtk_clist_thaw(current_search->clist);

			results = (current_search->results) ? (current_search->results)->len : 0;

			more = (read_8()!=0);
			gtk_widget_set_sensitive (button_more, more);
	
			actstr = g_string_sized_new (256);
			g_string_sprintf (actstr, _("Search results received. (%u results filtered) "), current_search->filtered);

			if (results<10 && (!more))
			{
				actstr = g_string_append (actstr, _("Not very many though - extending search..."));
// XXX
				if (current_search->isUserList)
					g_printerr (N_("FIX ME: different extend for user list! (bug)\n"));
				parse_message_extend_5x ();
			}

			// make sure we don't request forever
			if ((more) && requested_more_results<20)
			{
				actstr = g_string_append (actstr, _("More results available..."));
				// only ask automatically for more results if
				// we didn't get a single one that we don't
				// have yet - otherwise let user do it manually
				if (results==0)
				{
					gchar numstr[16];
					sprintf (numstr, N_("(%d)"), requested_more_results);
					actstr = g_string_append (actstr, numstr);
					gtk_widget_set_sensitive (button_addsearch, FALSE);
					// more search for users not available!
					if (!current_search->isUserList)
						send_more_search();
					else g_print (N_("FIX ME: different more search for user list! (bug)\n"));
					requested_more_results++;
				}
			} else {
				actstr = g_string_append (actstr, _("Search completed."));
				requested_more_results=0;
			}
			statusbar_extra_message (N_(" %s"), actstr->str);
			gtk_widget_set_sensitive (button_extend, TRUE);
			g_string_free(actstr,TRUE);
		}
		break;

		case CORE_SERVER_LIST:
		{
			guint i, num = read_32();
			servers_mark_all_for_removal ();
			for (i=0; i<num; i++)
			{
				guint8		*notused = read_x(16);
				guint32		 ip = read_IP();
				guint16		 port = read_16();
				GPtrArray	*taglist = taglist_read_from_ramfile (ramfile_now_active);
				NGS_tag		*nametag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILENAME_SHORT);
				NGS_tag		*desctag = taglist_find_tag_from_tagname (taglist, TAGNAME_SERVERDESCRIPTION_SHORT);
				NGS_tag		*userstag = taglist_find_tag_from_tagname (taglist, N_("users"));
				NGS_tag		*filestag = taglist_find_tag_from_tagname (taglist, N_("files"));
				NGS_tag		*pingtag = taglist_find_tag_from_tagname (taglist, TAGNAME_PING_SHORT);
				NGS_tag		*preftag = taglist_find_tag_from_tagname (taglist, TAGNAME_PREFERENCE_SHORT);
				servers_add_or_update_record ( ip, port,
					(nametag) ? (gchar*) nametag->data : NULL,
					(desctag) ? (gchar*) desctag->data : NULL,
					(userstag)? GPOINTER_TO_UINT(userstag->data) : 0,
					(filestag)? GPOINTER_TO_UINT(filestag->data) : 0,
					(pingtag) ? GPOINTER_TO_UINT(pingtag->data) : 0,
					(preftag) ? GPOINTER_TO_UINT(preftag->data) : 0);
				if (taglist) taglist_free(taglist);
				if (notused) g_free(notused);
			}
			servers_remove_all_old_records ();
			serversclist_update ();
			statusbar_extra_message (N_("     "));
		}
		break;

		case CORE_EXTENDING_SEARCH:
		{
			gchar *servername = read_i4str();
			if (servername)
			{
				statusbar_extra_message (_(" Extending search to server %s..."), servername);
				g_free(servername);
			}
		}
		break;

		case CORE_SEARCH_RESULT:
		{
			g_return_if_fail (current_search!=NULL);
			parse_message_get_results (1);
		}
		break;

		case CORE_NEW_DOWNLOAD:
		{
			guint8		*hash = read_x(16+4+2);	// read unused fields (4+2) too (they do no harm)
			GPtrArray	*taglist = taglist_read_from_ramfile (ramfile_now_active);
			NGS_tag		*nametag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILENAME_SHORT);
			NGS_tag		*sizetag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILESIZE_SHORT);
			guint8		 priority = read_8(); // get initial priority state
			gint		 i;	// needs to be gint, otherwise we can't catch error if partnum is not supplied by core
			guint		 partnum;
			guint8		 headerbuf[5]; /* to check if the next packet is already in the buffer */
			gchar		*pos;

			if ((hash) && (!IS_OVERNET) && (hash[16]!=0||hash[17]!=0||hash[18]!=0||hash[19]!=0||hash[20]!=0||hash[21]!=0))
			{
				g_printerr (N_("ATTENTION: Jed has changed FileMeta->toNetwork()! This needs to be fixed in the GUI.\n"));
				break;
			}

			// okay, the following part is tricky. Originally I planned to just send two bytes to represent
			// the x.part file number. So my test cores use(d) these. Now Jed is going to change this to
			// sending the full path+filename in the upcoming core, so we need to cater for both cases for now

			/* sometimes we get two ed2k-packets in one go. We'll have to check here whether
			 * there's already the next ed2k-packet in the buffer instead of
			 * the x.part number (alternatively we could have checked for the
			 * packet length of course, but this seems easier)
			 */
			(void) ramfile_peek (ramfile_now_active, headerbuf, ramfile_get_pos(ramfile_now_active), 5);
			if (headerbuf[0]==0xe3 && headerbuf[1]>0 && headerbuf[4]==0)
			{
				i=0;	/* ok, next packet is there => no partnum available */
			} else i = (gint) read_16();	// get temp filename number (we don't know if this isn't stringlen of full path now)

			if (i>0 && i<0xffff)
			{
				char *tmpfn = (char*) alloca((i+1)*sizeof(char));	// XXX - (i+1) ???
				if (tmpfn)
				{
					memset (tmpfn, 0x00, (i+1)*sizeof(char));
					tmpfn[0] = read_8();
					tmpfn[1] = read_8();
					// beginning of unix path or DOS path?
					if (tmpfn[0]=='/' || tmpfn[1]==':')
					{
						int pos;
						char *lastslash;
						for (pos=2; pos<i; pos++) { tmpfn[pos] = read_8(); }
						lastslash = strrchr (tmpfn, '/');	// find base filename
						if (!lastslash) lastslash = strrchr (tmpfn, '\\'); // might be DOS path
						if (lastslash)
						{
							i = atoi(lastslash+1); // convert initial digits to int
							if (i<0) i=0;
						} else i=0;
					} // else: just w16 as numeric value for x.part
				} else i=0; // mem alloc error
			} else i=0; // pre-59-gaps core
			partnum = i;

			if (partnum == 0 && (option_DLShowPartFileName))
			{
				static gboolean msgshown = FALSE;
				if (!msgshown)
				{
					g_printerr (_("This core does not support showing the x.part temporary filename of your downloads.\n"));
					msgshown = TRUE;
				}
			}

			/* warn if filename contains bad characters (like a slash) */
			if (nametag)
			{
				for (pos=(gchar*)nametag->data; *pos!=0x00; pos++)
				{
					if (*pos == SLASH_CHAR)
					{
						static gboolean msgshown = FALSE;
						if (!msgshown)
						{
							ok_dialog(_("warning"), _("Warning:\n"
										"At least one of the files in your download list\n"
										"contains characters in the filename that will\n"
										"cause problems later on (as for example a slash).\n\n"
										"You should cancel this file and re-download it\n"
										"using the 'download with custom filename' function\n"
										"from the search page.\n"));
							msgshown = TRUE;
						}
					}
				}
			}

			downloads_add_record ( hash,
				(nametag) ? (gchar*) nametag->data : NULL,
				(sizetag) ? GPOINTER_TO_UINT(sizetag->data) : 0,
				partnum, priority);

			if (CORE_IS_GAP_ENABLED)
				send_get_download_gaps (hash);

			if (taglist) taglist_free(taglist);
			if (hash) g_free(hash);
		}
		break;

		case CORE_REMOVE_DOWNLOAD:
		{
			guint8		 	 i, remove_hash[16];
			downloadRecord	*item;

			for (i=0; i<16; i++)
				remove_hash[i] = read_8();
			item = downloads_find_from_hash (remove_hash);
			if (item)
			{
				if (!item->cancelled)
				{
					gchar *badnosave = _("has been corrupted - file has been deleted from disk.\n");
					gchar *badsave   = _("has been corrupted - file has been saved and is in your temp folder.\n");
					gchar *goodsave  = _("should have been moved to your incoming folder.\n");
					status_message_blue (_("----- DOWNLOAD COMPLETE ---------------\n"));
					status_message_blue (N_("%s\n"), item->name);
					if ((item->corrupted) || (last_status_message_was_download_corrupted))
					{
						g_printerr (N_("%s"), (option_save_corrupted==0) ? badnosave : badsave);
					} else status_message_blue (N_("%s"), goodsave);
					exec_command_on_complete_download(item->name);
					status_message_blue (N_("---------------------------------------\n"));
					/* and add to 'completed' list (before the 'choplast()' call */
					downloads_completed_new_from_download (item);
					statusbar_extra_message (_(" %s has been completed."), choplast(item->name));
				} else status_msg (N_("%s: ed2k://|file|%s|%u|%s|\n"), _("GUI: download cancelled"), item->name, item->size, hash_to_hash_str(remove_hash));
				downloads_remove_record (item);
			} else g_printerr (N_("DEBUG: I was asked to remove a download that is not in the download list!\n"));
		}
		break;

		case CORE_DOWNLOAD_STATUS:
		{
			gtk_clist_freeze (DL_CLIST);
			parse_message_receive_dl_status_list();
			gtk_clist_thaw (DL_CLIST);
		}
		break;

		case CORE_NEW_UPLOAD:
		{
			gchar		*filename = read_i4str();
			guint8		*userhash = read_x(16+4+2);	// read unused fields too (they do no harm)
			GPtrArray	*taglist = taglist_read_from_ramfile (ramfile_now_active);
			NGS_tag		*usernametag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILENAME_SHORT);

			// Note: in this function, the file hash will be set to zero. With each
			//	upload status packet we get, we'll check whether the filehash is zero
			//	and try to find it via the filename (from downloads or shared files) if
			//	it is. Once we have the filehash, we can do upload statistics...
			uploads_add_record (userhash, filename,
				(usernametag) ? (gchar*)usernametag->data : N_("(???)"));

			if (taglist) taglist_free(taglist);
			if (userhash) g_free(userhash);
			if (filename) g_free(filename);
		}
		break;

		case CORE_REMOVE_UPLOAD:
		{
			guint8 *userhash = read_x(16);
			if (userhash)
			{
				PROCESS_GPTRARRAY( uploads , uploadRecord,
					if ((item) && memcmp(item->userhash, userhash, 16)==0)
					{
						uploads_remove_record(item);
						break;
					}
				);
				g_free(userhash);
			}
		}
		break;

		case CORE_UPLOAD_STATUS:
		{
			parse_message_receive_ul_status_list();
		}
		break;

		case CORE_NEW_UPLOAD_SLOT:
		{
			/* ignore the 'assigned slot' stuff - it's useless */
/*			guint	 slotnumber = read_32();
			gchar	*peername = read_i4str();
			gint	 row = gtk_clist_append (GTK_CLIST(qlist), &peername);
			gtk_clist_set_row_data (GTK_CLIST(qlist), row, GUINT_TO_POINTER(slotnumber));
			if (peername) g_free (peername);
*/
		}
		break;

		case CORE_REMOVE_UPLOAD_SLOT:
		{
			/* ignore the 'assigned slot' stuff - it's useless */
/*			guint	slotnumber = read_32();
			gint	row = gtk_clist_find_row_from_data (GTK_CLIST(qlist), GUINT_TO_POINTER(slotnumber));
			if (row>=0) gtk_clist_remove (GTK_CLIST(qlist), row);
*/
		}
		break;

		case CORE_SHARED_FILES:
		{
			guint i, num = read_32();

			shared_files_clist_before_update ();
			for (i=0; i<num; i++)
			{
				gchar		*cols[4];
				guint8		*hash = read_x(16);
				guint8		*unused = read_x(4+2);
				GPtrArray	*taglist = taglist_read_from_ramfile (ramfile_now_active);
				NGS_tag		*nametag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILENAME_SHORT);
				NGS_tag		*formattag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILEFORMAT_SHORT);
				NGS_tag		*sizetag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILESIZE_SHORT);
				NGS_tag		*typetag = taglist_find_tag_from_tagname (taglist, TAGNAME_FILETYPE_SHORT);
				guint32		 size = (sizetag) ? GPOINTER_TO_UINT(sizetag->data) : 0;

				cols[0] = ((nametag) && (nametag->data)) ? (gchar*) nametag->data : N_("(no name)");
				cols[1] = N_("");	// not used
				cols[2] = ((formattag) && (formattag->data)) ? (gchar*) formattag->data : (gchar*) misc_get_extension_from_filename (cols[0]);
				cols[3] = ((typetag) && (typetag->data)) ? (gchar*) typetag->data : (gchar*) misc_get_filetype_from_extension (cols[2]);
				shared_files_add_or_update_record (hash, cols[0], size, cols[2], cols[3]);
				if (taglist) taglist_free(taglist);
				if (hash) g_free(hash);
				if (unused) g_free(unused);
			}
			shared_files_clist_after_update ();
		}
		break;

		case CORE_SHARED_DIRS:
		{
			if (!shareddirs) shareddirs = g_ptr_array_new();
			if (shareddirs)
			{
				guint i, num;

				// flag all for removal
				PROCESS_GPTRARRAY (shareddirs, shareddirRecord,
					if (item) item->remove=TRUE;
				);

				num = read_16();
				for (i=0; i<num; i++)
				{
					gchar			*dirname = read_i4str();
					shareddirRecord	*sdr = shareddirs_find_from_dirname (dirname);
					if (!sdr)
					{
						sdr = g_new0 (shareddirRecord, 1);
						g_return_if_fail (sdr!=NULL);
						sdr->dirname = dirname;
						g_ptr_array_add (shareddirs, (gpointer)sdr);
					} else g_free(dirname);
					sdr->idxnum = i+1;		// dir nums start with 1
					sdr->remove = FALSE;
				}

				// remove all records that aren't in our new list any more
				PROCESS_GPTRARRAY (shareddirs, shareddirRecord,
					if ((item) && (item->remove)) shareddirs_remove_record(item);
				);

				shareddirsclist_update();
			}
		}
		break;

		case CORE_OPTIONS:
		{
			parse_message_receive_options();
			options_update_option_widgets();
		}
		break;

		case CORE_DOWNLOAD_GAP_DETAILS:
		{
			guint			 i;
			guint8			 hash[16];
			downloadRecord	*item;

			flag_core_knows_gaps = 1;

			for (i=0; i<16; i++)	{ hash[i]=read_8(); }

			item = downloads_find_from_hash (hash);

			if (item)
			{
//				int debug=0;
				guint gaps = read_16();
				init_gaplist (item, gaps);
/*				DEBUG(3,"received CORE_DOWNLOAD_GAP_DETAILS\n");
				if ((downloadlist+item)->sources>=3)
				{
					debug=1;
					printf ("no in DL list = %u (%s)\n", item, (downloadlist+item)->name);
					printf ("no of gaps = %u\n", gaps);
				}
*/				for (i=0; i<gaps; i++)
				{
					guint start = read_32();
					guint end   = read_32();
					guint val   = read_16();
					add_gap (item, start, end, val);
///					if (debug==1) printf ("gap #%03u: %08u - %08u (%08u) - value: %02u\n", i,start,end,end-start,val);
				}
//				force_dllist_redraw_all = 1;
				item->lasttime_got_gaps = time(NULL);
			} else g_printerr (N_("received gaplist for download not in downloadlist?!\n"));
//			g_print ("FIXME: gap details in parse_message\n");
		}
		break;

		case CORE_CLIENT_STATS:
		{
			gfloat	 tmp = float_read();
			gfloat	 in  = float_read();
			gfloat	 needed = float_read();
			guint	 clientid = read_32();
			gchar	*stattxt = g_strdup_printf (N_("ClientID: %u (%s)\n"), clientid, (clientid>0xFFFFFF) ? _("open") : _("firewalled"));
			guint16	 on_queue;
			guint16	 cons_now;

			if (stattxt)
			{
				(void) parse_status_message_from_core (stattxt);
				g_free(stattxt);
			}

			downloads_page_set_free_space (needed, tmp, in);
			if (tmp <= (option_minfreetempspace*1.0))
			{
				downloads_pause_all(tmp);
			}

			cons_now = read_16();
			on_queue = read_16();
			uploads_page_set_people_on_queue (on_queue);
			(void) toolbar_set_cons ((guint)cons_now);

			space_stats_last_updated = time(NULL);
		}
		break;

		default:
		{
			ramfile_dump (packet, _(" Received unknown message from core "));
		}
	}
}
