/***************************************************************************
                          search_data.c  -  description
                             -------------------
    begin                : Sun Jul 28 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "blacklist.h"
#include "colors.h"
#include "downloads_data.h"
#include "global.h"
#include "gui_flags.h"
#include "gui_options.h"
#include "mainwindow.h"
#include "misc.h"
#include "search_data.h"
#include "search_page.h"
#include "shared_files.h"
#include "filter.h"

searchList	*current_search = NULL;		// the last search - where new results are added

GtkWidget	*search_notebook = NULL;

GPtrArray	*searchlists = NULL;

static void				 searchclist_add_or_update_record (searchList *sl, searchlistRecord *slr);
static searchlistRecord	*searchlist_find_record_from_hash (searchList *sl, const guint8 *hash);

static void				 searchclist_remove_record (GtkCList *clist, searchlistRecord *slr);


// searchlist_clean_up_all
//
//

void
searchlist_clean_up_all (void)
{
	if (search_notebook)
	{
		if (searchlists)
		{
			PROCESS_GPTRARRAY_FROM_END ( searchlists, searchList,
				if (item) searchlist_destroy (item);
			);
		}

		if (window)
			(void) searchlist_make_new_searchlist (_("empty"));

		current_search = NULL;
	}
}


// searchlist_find_from_clist
//
// finds the searchList record associated with a given clist.
// returns record or NULL if not found

searchList *
searchlist_find_from_clist (GtkCList *clist)
{
	g_return_val_if_fail (clist!=NULL,NULL);

	if (searchlists)
	{
		PROCESS_GPTRARRAY ( searchlists, searchList,
			if ((item) && (item->clist == clist))
				return item;
		);
	}	
	return NULL;
}


// searchlist_find_from_frame
//
// finds the searchList record associated with a given frame (child in GtkNotebookPage).
// returns record or NULL if not found

searchList *
searchlist_find_from_frame (GtkWidget *frame)
{
	g_return_val_if_fail (frame!=NULL,NULL);

	if (searchlists)
	{
		PROCESS_GPTRARRAY ( searchlists, searchList,
			if ((item) && (item->frame == frame))
				return item;
		);
	}
	return NULL;
}


// searchlist_we_have_record
//
//

static gboolean
searchlist_we_have_record (const guint8 *hash)
{
	// XXX - true instead of false does the least damage?!
	g_return_val_if_fail (hash!=NULL, TRUE);

	if (downloads_find_from_hash (hash))
		return TRUE;

	if (shared_files_find_from_hash (hash))
		return TRUE;

	return FALSE;	
}


// searchlist_find_record_from_hash
//
//

static searchlistRecord *
searchlist_find_record_from_hash (searchList *sl, const guint8 *hash)
{
	g_return_val_if_fail (sl!=NULL, NULL);
	g_return_val_if_fail (hash!=NULL, NULL);
	if (sl->results)
	{
		PROCESS_GPTRARRAY ( sl->results, searchlistRecord,
			if ((item) && memcmp(item->hash, hash, 16)==0)
				return item;
		);
	}
	return NULL;
}


// searchlist_free_record
//
//

static void
searchlist_free_record (searchlistRecord *slr)
{
//	GSList	*node;

	g_return_if_fail (slr!=NULL);

	if (slr->name)
		g_free(slr->name);

	if (slr->mp3_artist)
		g_free(slr->mp3_artist);

	if (slr->mp3_title)
		g_free(slr->mp3_title);

	if (slr->mp3_album)
		g_free(slr->mp3_album);

	PROCESS_GSLIST (slr->altnames, gchar,
		if (item) g_free(item);
	);
	g_slist_free(slr->altnames);

/*	node = slr->altnames;
	while (node)
	{
		if (node->data) g_free(node->data);
		node=node->next;
	}
*/
	memset(slr,0x00,sizeof(slr));
	g_free(slr);
}



// searchlist_remove_record
//
//

void
searchlist_remove_record (searchList *sl, searchlistRecord *slr)
{
	g_return_if_fail (sl!=NULL);
	g_return_if_fail (slr!=NULL);

	if (sl->results)
	{
		g_ptr_array_remove (sl->results, (gpointer)slr);		// XXX - ..._remove_fast will segfault
		if ((sl->clist) && (window))
			searchclist_remove_record (sl->clist, slr);
		searchlist_free_record (slr);
	}
}


// searchlist_destroy
//
//	

void
searchlist_destroy (searchList *sl)
{
	gint idxpos = -1;

	g_return_if_fail (sl!=NULL);

	if (!window)
		return;

	if (sl->clist) 
		gtk_clist_freeze(sl->clist);

	if (sl->results)
	{
		PROCESS_GPTRARRAY_FROM_END ( sl->results, searchlistRecord,
			if (item) searchlist_remove_record (sl,item);
		);
	}

	if (sl->clist) 
		gtk_clist_thaw(sl->clist);

	if (sl->clist)
	{
		gint pagenum = gtk_notebook_page_num (GTK_NOTEBOOK(search_notebook), sl->frame);
		if (pagenum>=0) gtk_notebook_remove_page (GTK_NOTEBOOK(search_notebook), pagenum);
	}

	if (sl == current_search)
	{
		g_return_if_fail (searchlists!=NULL);
		PROCESS_GPTRARRAY_FROM_END ( searchlists, searchList,
			if (item == current_search)
			{
				idxpos = gpainc;	// gpainc is variable from macro
				break;
			}
		);
		g_return_if_fail(idxpos>=0);
	}

	// Note: there should always be at least one page, so we don't have to care about
	// active_searchlist, the notebook page switch function will take care of this

	g_ptr_array_remove (searchlists, (gpointer)sl);

	if ((sl == current_search) && (window) && (searchlists) && (searchlists->len>0))
	{
		gint pagenum;
		// if last search was removed, switch to search before, 
		//	otherwise stay at same search page number
		if (idxpos >= searchlists->len)
			idxpos--;
		current_search = g_ptr_array_index (searchlists, idxpos);
		g_return_if_fail (current_search!=NULL);
		pagenum = gtk_notebook_page_num (GTK_NOTEBOOK(search_notebook), current_search->frame);
		if (pagenum>=0) gtk_notebook_set_page (GTK_NOTEBOOK(search_notebook), pagenum);
	}

	if (sl->label)
		g_free(sl->label);

	memset (sl, 0x00, sizeof(sl));
	g_free(sl);

}


// searchlist_add_or_update_record
//
//

void
searchlist_add_or_update_record (	searchList *sl,
									guint idxpos,
									const guint8 *hash,
									guint32 size,
									guint32 availability,
									const gchar *name,
									const gchar *type,
									const gchar *format,
									const gchar *mp3_artist,
									const gchar *mp3_album,
									const gchar *mp3_title,
									const gchar *length,
									guint32 bitrate,
									const gchar *codec)
{
	searchlistRecord	*slr;
	gboolean			 we_have_this;

	g_return_if_fail (sl!=NULL);
	g_return_if_fail (sl->clist!=NULL);
	g_return_if_fail (hash!=NULL);
	g_return_if_fail (name!=NULL);

//	g_print ("%s:\n%03u - %s %10u %10u %s\n%s %s %s %s %s %s %u %s\n",
//				__FUNCTION__, idxpos, hash_to_hash_str(hash), size, availability, name,
//				type, format, mp3_artist, mp3_album, mp3_title, length, bitrate, codec);

	if ( filter_this_filename(name)==1
			|| ((!option_show_blacklist_items_grey) && (item_is_in_blacklist(hash,size))))
	{
		sl->filtered++;
		return;
	}

	we_have_this = searchlist_we_have_record (hash);
	if (option_show_own_files_in_search==0 && (we_have_this))
		return;

	slr = searchlist_find_record_from_hash (sl, hash);
	if (!slr)
	{
		const gchar *ext;

		slr = g_new0 (searchlistRecord, 1);		// sets all to 0 as well
		g_return_if_fail (slr!=NULL);

		memcpy (slr->hash, hash, 16);
		slr->we_have_this = we_have_this;

		if (!sl->results)
			sl->results = g_ptr_array_new();
		g_ptr_array_add (sl->results, (gpointer)slr);
		slr->name = g_strdup(name);
		slr->size = size;

		// XXX - isn't format and extension the same thing????
		ext = misc_get_extension_from_filename(name);
		if (ext) slr->extensionquark = g_quark_from_string(ext);

		if (!type)
		{
			if (ext) slr->typequark = g_quark_from_string(misc_get_filetype_from_extension (ext));
		}else slr->typequark = g_quark_from_string(type);
		
		if (format) slr->formatquark = g_quark_from_string(format);
		if (codec)
		{
			slr->codecquark = g_quark_from_string(codec);
			sl->hasCodec = (*codec!=0x00);
		}
		if (length)
		{
			slr->lengthquark = g_quark_from_string(length);
			sl->hasLength = (*length!=0x00);
		}
		if (mp3_artist)
		{
			slr->mp3_artist = g_strdup(mp3_artist);
			sl->hasArtist = TRUE;
		}
		if (mp3_album)
		{
			slr->mp3_album = g_strdup(mp3_album);
			sl->hasAlbum = TRUE;
		}
		if (mp3_title)
		{
			slr->mp3_title = g_strdup(mp3_title);
			sl->hasTitle = TRUE;
		}
		slr->bitrate = bitrate;
		if (bitrate>0) sl->hasBitrate = TRUE;
	} else {
		// If it is a different name than the one we got, check whether
		//	we already have this alternative name in our list. If not, add it.
		if (g_strcasecmp(slr->name,name)!=0)	
		{
			gboolean	 already_in_list = FALSE;
			GSList		*node = slr->altnames;

			while (node)	// is this alternative name already in our list?
			{
				if ((node->data) && g_strcasecmp((gchar*)node->data,name)==0)
				{
					already_in_list = TRUE;
					break;
				}
				node=node->next;
			}
			if (!already_in_list) slr->altnames=g_slist_append(slr->altnames, (gpointer)g_strdup(name));
		}
	}
	slr->availability += availability;
	searchclist_add_or_update_record (sl, slr);
}


// searchclist_add_or_update_record
//
//

static void
searchclist_add_or_update_record (searchList *sl, searchlistRecord *slr)
{
	gint		 row;
	gchar		 availstr[64], bitratestr[64], sizestr[64];

	g_return_if_fail (sl!=NULL);
	g_return_if_fail (sl->clist!=NULL);
	g_return_if_fail (slr!=NULL);

	row = gtk_clist_find_row_from_data (sl->clist, (gpointer)slr);
	if (row<0)
	{
		gchar	*empty[NO_OF_SEARCH_COLS];
		gchar	*frametxt;
		gint	 i;

		empty[0] = slr->name;
		g_snprintf (sizestr, sizeof(sizestr), N_("%u"), slr->size);
		empty[1] = sizestr;
		empty[2] = "";	// availability - will be set below
		empty[3] = g_quark_to_string (slr->typequark);
		empty[4] = g_quark_to_string (slr->lengthquark);
		g_snprintf (bitratestr, sizeof(bitratestr), N_("%u"), slr->bitrate);
		empty[5] = bitratestr;
		empty[6] = g_quark_to_string (slr->codecquark);;
		empty[7] = slr->mp3_artist;
		empty[8] = slr->mp3_title;
		empty[9] = slr->mp3_album;
		empty[10] = g_quark_to_string (slr->extensionquark);
		empty[11] = (gchar*) hash_to_hash_str (slr->hash);
		for (i=0; i<NO_OF_SEARCH_COLS; i++) if (!empty[i]) empty[i]=N_("");	// where g_quark_to_str() returned NULL ...
		row = gtk_clist_append(sl->clist, empty);
		gtk_clist_set_row_data(sl->clist, row, (gpointer)slr);

		gtk_clist_set_column_visibility (sl->clist, 1, (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 2, (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 3, (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 4, (sl->hasLength) && (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 5, (sl->hasBitrate)&& (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 6, (sl->hasCodec)  && (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 7, (sl->hasArtist) && (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 8, (sl->hasTitle)  && (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist, 9, (sl->hasAlbum)  && (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist,10, (!sl->isUserList));
		gtk_clist_set_column_visibility (sl->clist,11, (!sl->isUserList));

		gtk_clist_set_column_title (sl->clist, 0, (sl->isUserList) ? _("user name") : _("filename"));

		frametxt = g_strdup_printf (_(" %u search results "), (sl->clist)->rows);
		if (frametxt)
		{
			if (sl->frame) gtk_frame_set_label (GTK_FRAME(sl->frame), frametxt);
			g_free(frametxt);
		}
	} 

	if (slr->availability<99999)	// otherwise it's an ed2k-link
	{
		g_snprintf (availstr, sizeof(availstr), N_("%u"), slr->availability);
		gtk_clist_set_text (sl->clist, row, 2, availstr);
	} else gtk_clist_set_column_visibility (sl->clist, 2, FALSE);

	// XXX - this costs CPU cycles!
	// but it lets us update the clists automatically after we've added records etc.
	{
		static gboolean msgflag = FALSE;
		if (!msgflag)
			g_print (_("FIX ME: put searchlist_we_have_record() call in notebook switch function\n"));
		msgflag = TRUE;
	}
	slr->we_have_this = searchlist_we_have_record (slr->hash);

	if ((!sl->isUserList) && slr->availability > 8 && (!IS_OVERNET))
		gtk_clist_set_foreground (sl->clist, row, &blue);

//	if (option_show_own_files_in_search==1)
//		gtk_clist_set_selectable (clist, row, (!slr->we_have_this));

	if (item_is_in_blacklist(slr->hash, slr->size))
		gtk_clist_set_foreground (sl->clist, row, &grey);

}



// searchclist_remove_record
//
//

static void
searchclist_remove_record (GtkCList *clist, searchlistRecord *slr)
{
	searchList	*sl;
	gint		 row;

	g_return_if_fail (clist!=NULL);
	g_return_if_fail (slr!=NULL);

	row = gtk_clist_find_row_from_data (clist, (gpointer)slr);
	if (row>=0) gtk_clist_remove (clist, row);

	sl = searchlist_find_from_clist(clist);
	if (sl)
	{
		gchar *frametxt = g_strdup_printf (_(" %u search results "), (sl->clist)->rows);
		if (frametxt)
		{
			if (sl->frame) gtk_frame_set_label (GTK_FRAME(sl->frame), frametxt);
			g_free(frametxt);
		}
	}

}


// searchlists_make_new_searchlist
//
//

searchList *
searchlist_make_new_searchlist (const gchar *label)
{
	searchList	*newsl;

	g_return_val_if_fail (label!=NULL,NULL);
	g_return_val_if_fail (search_notebook!=NULL,NULL);

	if (!window) return NULL;		// don't do it if we're about to quit or starting in console mode

	// if current search list is empty, do not keep it
	if (current_search)
	{
		if ((!current_search->results) || ((current_search->results) && (current_search->results)->len==0))
			searchlist_destroy (current_search); // will remove page from notebook as well
	}

	newsl = g_new0 (searchList, 1);
	g_return_val_if_fail (newsl!=NULL,NULL);

	newsl->label = g_strdup(label);

	if (!searchlists) searchlists = g_ptr_array_new();
	g_ptr_array_add (searchlists, (gpointer) newsl);

	current_search = newsl;

	// this call needs to be made after the g_ptr_array_add()!
	searchnotebook_make_new_page(label, newsl);

	return newsl;
}












