/***************************************************************************
                          servers_data.c  -  maintains our server list
                             -------------------
    begin                : Wed Jul 3 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// TODO:
// * fix bug: servers with same IP/port twice in list (not confirmed yet)
//	(reported by Thomas Ritter, 15 July 2002)


#include "global.h"
#include "gui_options.h"
#include "misc.h"
#include "servers_data.h"
#include "servers_page.h"
#include "status_page.h"
#include "statusbar.h"
#include "send_command.h"

// An array of pointers to serverRecord items (note: not necessarily in the same order
//	that the command line client has them in its list
GPtrArray		*servers = NULL;



// What do we do if the user has selected multiple servers and clicks 'connect?'
//	we need to do the connection process on our own. This is a list of
//	GInetAddr structures we want to connect to
GList			*connectlist = NULL;

// We need this to save the original core option while we disable 'always stay connected'
//	while we do our own connect thing and reinstate it later on to what it was before
static gboolean		 option_alwaysStayConnectedSaved = FALSE;



// Pointer to the serverRecord we were last trying to connect to (if found)
static serverRecord	*servers_last_attempted_to_connect_to = NULL;



static void		 servers_record_free (serverRecord *srv);
static gint		 servers_placedOnQDeadlockAvoidTimerCallback (gpointer data);


// servers_connect_next
//
// Tries to connect to the next selected server.
// Returns FALSE if there are no more servers to try, otherwise TRUE.

gboolean
servers_connect_to_next_selected (void)
{
	GInetAddr	*ia;
	GList		*first;

	if (!connectlist) return FALSE;

	ia = (GInetAddr*) connectlist->data;

	first = connectlist;
	connectlist = connectlist->next;
	g_list_remove_link (connectlist, first);
	g_list_free_1 (first);

	if (ia)
	{
		guint32		 ip = misc_gnet_inetaddr_get_ip (ia);
		guint16		 port = gnet_inetaddr_get_port (ia);
		serverRecord	*srv = servers_find_from_address (ip,port);

		if ((srv) && (srv->nogood))
		{
			gchar *ipcanonicalstr = gnet_inetaddr_get_canonical_name(ia);
			status_msg (_("GUI: Skipping server '%s' - it has rejected us before.\n"),
					(srv->name) ? srv->name : ipcanonicalstr);
			g_free (ipcanonicalstr);
		} else send_connect_to_server (ip, port);
		gnet_inetaddr_delete(ia);
	}
	
	return TRUE;
}

// servers_clear_connect_to_list
//
// Clears the connect to queue when we actually got connected to a server.

void
servers_clear_connect_to_list (void)
{
	GInetAddr	*ia;
	GList		*node = connectlist;
	while (node)
	{
		ia = (GInetAddr*) node->data;
		if (ia) gnet_inetaddr_delete (ia);
		node = node->next;
	}
	g_list_free (connectlist);
	connectlist = NULL;
}


// servers_save_alwaysconnected_option
//
// Save the current 'always stay connected' option

void
servers_save_alwaysconnected_option ()
{
	option_alwaysStayConnectedSaved = option_always_connected;
}


// servers_reinstate_alwaysconnected_option
//
// Reinstate the core's 'always stay connected' option as it was before

void
servers_reinstate_alwaysconnected_option()
{
	if (option_alwaysStayConnectedSaved)
	{
		send_command (N_("auto"));	// switch 'always stay connected on again'
		send_get_options();
	}
}


// servers_init
//
// initialises the server data list

void
servers_init(void)
{
	if (!servers) servers = g_ptr_array_new();
	servers_last_attempted_to_connect_to = NULL;
}


static void
servers_record_free (serverRecord *srv)
{
	gint row;

	g_return_if_fail (srv!=NULL);

	if (srv->name)
		g_free(srv->name);

	if (srv->desc)
		g_free(srv->desc);

	if (srv == servers_last_attempted_to_connect_to)
		servers_last_attempted_to_connect_to = NULL;

	memset(srv,0x00,sizeof(srv));
	g_free(srv);

	// make sure record is removed from clist as well
	if (SRV_CLIST)
	{
		row = gtk_clist_find_row_from_data (SRV_CLIST, (gpointer)srv);
		if (row>=0) gtk_clist_remove (SRV_CLIST, row);
	}
}


// servers_cleanup
//
// frees up the server data list

void
servers_cleanup(void)
{
	if (clist_servers) gtk_clist_freeze(SRV_CLIST);
	if (servers)
	{
		// Free all elements, but not array structure itself
		PROCESS_GPTRARRAY_FROM_END( servers, serverRecord,
			if (item)
			{
				servers_record_free(item);
				g_ptr_array_remove_index_fast (servers, gpainc); // gpainc is from macro
			}	
		);
	}
	if (clist_servers) gtk_clist_thaw(SRV_CLIST);
	servers_last_attempted_to_connect_to = NULL;
}


// servers_find_from_address
//
// finds an item from the server data list by ip/port
// returns server record or NULL if not found

serverRecord *
servers_find_from_address (guint32 ip, guint16 port)
{
	g_return_val_if_fail (servers!=NULL, NULL);
	PROCESS_GPTRARRAY( servers, serverRecord,
		if ((item) && item->ip==ip && (port==0 || (port>0 && item->port==port)))
		{
			return item;
		}
	);
	return NULL;
}



// servers_add_or_update_record
//
// Adds a server record to the servers list if it's not already there,
//	or just updates the values if it's already there.

void
servers_add_or_update_record (guint32 ip, guint16 port, const gchar *name, const gchar *desc, guint32 users, guint32 files, guint32 ping, guint8 pref)
{
	serverRecord	*srv = servers_find_from_address (ip, port);
	gchar		*ourname = (name!=NULL)? g_strdup(name) : NULL;
	if (!srv)
	{
		// server is new => make new record
		srv = g_new0 (serverRecord, 1);
		g_return_if_fail (srv!=NULL);
		srv->ip = ip;
		srv->port = port;
		srv->nogood = FALSE;
		srv->last_seen_alive = time(NULL);
		srv->last_connected = (time_t) 0;
		srv->last_connecting = (time_t) 0;
		g_ptr_array_add (servers, (gpointer)srv);
	} else {
		if (srv->name) g_free(srv->name);
		if (srv->desc) g_free(srv->desc);
	}
	srv->name = NULL;
	srv->desc = NULL;
	// We don't want servernames that are too long
	if ((ourname) && strlen(ourname)>64)
	{
		ourname[63] = '*';	// indicate that it's been cut off
		ourname[64] = 0x00;
	}
	if (ourname) srv->name = g_strdup(ourname);	// more allocs+frees, but saves memory
	if (desc) srv->desc = g_strdup(desc);
	srv->users = users;
	srv->files = files;
	srv->ping = ping;
	srv->pref = pref;
	srv->remove = FALSE;
	if (users>0 && files>0 && ping>0) srv->last_seen_alive = time(NULL);
	if (ourname) g_free(ourname);
}


// servers_mark_all_for_removal
//
// Before we receive a new serverlist from the core, this routine flags every server,
//	and while the list is updated, the flags are removed, so that in the end only
//	servers that the core doesn't have any longer remain flagged (and are then
//	removed by servers_remove_all_old_records()

void
servers_mark_all_for_removal (void)
{
	PROCESS_GPTRARRAY( servers, serverRecord, if (item) item->remove = TRUE; );
}


// servers_remove_record
//
// removes the given record from the servers list AND the servers clist
// Returns TRUE if BOTH was successful, otherwise FALSE

gboolean
servers_remove_record (serverRecord *srv)
{
	gboolean	ret;
	gint		row;

	g_return_val_if_fail (srv!=NULL, FALSE);
	// remove record from clist (if found)
	row = serversclist_find_row_from_record (srv);
	if (row>=0) gtk_clist_remove (SRV_CLIST, row);
	// free server record data and remove from servers list
	servers_record_free(srv);
	ret = g_ptr_array_remove_fast(servers,(gpointer)srv);	// order of records doesn't matter
	return ((ret)&&(row>=0));
}


// servers_remove_all_old_records
//
// see servers_mark_all_for_removal()

void
servers_remove_all_old_records (void)
{
	if (clist_servers)
		gtk_clist_freeze(SRV_CLIST);

	PROCESS_GPTRARRAY_FROM_END( servers, serverRecord,
		if (item)
			if (item->remove) servers_remove_record (item);
	);

	if (clist_servers)
		gtk_clist_thaw(SRV_CLIST);
}



// servers_placedOnQDeadlockAvoidTimerCallback
//
// This is a timer callback set up after we've got a "Placed on Connection Queue"
//	when the 'ignore connection queue deadlock' options is not switched on.
// If we are still 'Placed on Connection queue' after so and so many seconds,
//	this function is called, so we can try to resolve the deadlock in the core.
//	(which might or might not work. In these cases, the core is somehow screwed
//	up internally anyway, so there's not always much we can do about it).

static gint
servers_placedOnQDeadlockAvoidTimerCallback (gpointer data)
{
	// if we are called this means that we haven't gotten a "can't connect" or "connected to" message for
	// 30 seconds => deadlock assumed, send 'x' as advanced command (as message doesn't work for some reason)
	servers_register_connect_status (NULL, SERVERS_CONNECT_STATUS_PLACEDONQ_DEADLOCK);
// XXX - maybe we should return TRUE here then?
	return 0;	// do not call us again (we'll be removed by servers_register...() anyway)
}



// servers_register_connect_status
//
// When the core is connecting to a server this function is called so we know what's
//	going on and can do certain stuff if we want to (like set servers to low
//	priority that think we are firewalled etc.)
//
// ipOrName is only set with SERVERS_CONNECT_STATUS_CONNECTING, otherwise NULL.
//
// This Function sets the local variable servers_last_attempted_to_connect_to if it can.

void
servers_register_connect_status (const gchar *ipOrName, serversConnectStatus status)
{
	static guint			placedOnQDeadlockAvoidTimer = 0;
	static serversConnectStatus	lastStatus = SERVERS_CONNECT_STATUS_INVALID;

	if (placedOnQDeadlockAvoidTimer>0)
	{
		gtk_timeout_remove (placedOnQDeadlockAvoidTimer);
		placedOnQDeadlockAvoidTimer = 0;
	}

	switch (status)
	{
		case SERVERS_CONNECT_STATUS_CONNECTING:
		{
			guint32		 n1,n2,n3,n4;

			servers_last_attempted_to_connect_to = NULL;
			g_return_if_fail (ipOrName);
			if (sscanf (ipOrName, N_("%u.%u.%u.%u"),&n1,&n2,&n3,&n4)!=4)
			{
				// find by name if it's not an IP address
				PROCESS_GPTRARRAY( servers, serverRecord,
					if ((item) && (item->name) && g_strncasecmp(item->name,ipOrName,60)==0)
					{
						servers_last_attempted_to_connect_to = item;
						break;
					}
				);
			} else servers_last_attempted_to_connect_to = servers_find_from_address (n1|n2<<8|n3<<16|n4<<24, 0);

			if (servers_last_attempted_to_connect_to)
			{
				servers_last_attempted_to_connect_to->last_connecting = time(NULL);
			}
		}
		break;
		case SERVERS_CONNECT_STATUS_PLACEDONQ:
		{
			/* Avoid "placed on connection queue..." deadlocks:
			 *	Start timer, and send 'x' as advanced command
			 *	if we haven't had a 'Can't connect to' or 'connected to'
			 *	for a certain amount of time.
			 */
			if (option_ignore_queue_deadlock==0)
				placedOnQDeadlockAvoidTimer = gtk_timeout_add (15*1000, servers_placedOnQDeadlockAvoidTimerCallback, NULL);
		}
		break;

		case SERVERS_CONNECT_STATUS_CANTCONNECT:	// note: ipOrName will be NULL now
		{
			if (servers_last_attempted_to_connect_to)
			{
				/* set server to low priority, but only if it was 'normal priority before' */
				if (servers_last_attempted_to_connect_to->pref == SERVER_PRIO_NORMAL)

/*					XXX - Tim: huh? what was that for? */
/*					|| (!servers) || ((servers) && servers->len<10)) */
				{
					send_set_server_priority (	servers_last_attempted_to_connect_to->ip,
												servers_last_attempted_to_connect_to->port,
												SERVER_PRIO_LOW);
					servers_last_attempted_to_connect_to->pref = SERVER_PRIO_LOW;
					serversclist_update ();
				}
/*					Let's not remove it just because we couldn't connect twice.
				else {
					send_remove_server (servers_last_attempted_to_connect_to->ip, servers_last_attempted_to_connect_to->port);
					servers_remove_record (servers_last_attempted_to_connect_to);	// remove from list + clist, too
				}
*/
			}
			servers_last_attempted_to_connect_to = NULL;
			if (!servers_connect_to_next_selected ())
			{
				servers_reinstate_alwaysconnected_option();
			}

		}
		break;

		case SERVERS_CONNECT_STATUS_DISCONNECTED:
		{
			/* set server priority to low on disconnect if we have
			 * been connected less than 1 minute. Set priority to
			 * high if we have been connected for more than 5 minutes
			 */
			if (servers_last_attempted_to_connect_to)
			{
				time_t time_connected = time(NULL) - servers_last_attempted_to_connect_to->last_connecting;
				if ( time_connected < 60)
				{
					send_set_server_priority (	servers_last_attempted_to_connect_to->ip,
												servers_last_attempted_to_connect_to->port,
												SERVER_PRIO_LOW);
				}
				if ( time_connected > (5*60) && time_connected!=time(NULL))
				{
					send_set_server_priority (	servers_last_attempted_to_connect_to->ip,
												servers_last_attempted_to_connect_to->port,
												SERVER_PRIO_HIGH);
				}
			}
			servers_last_attempted_to_connect_to = NULL;
		}
		break;

		case SERVERS_CONNECT_STATUS_PLACEDONQ_DEADLOCK:
		{
			if (lastStatus == SERVERS_CONNECT_STATUS_PLACEDONQ)
			{
				status_message_blue (_("GUI: detected possible core deadlock on connection attempt - trying to disconnect and reconnect.\n"));
				status_message_blue (_("GUI: explanation: this happens if server is not responding within 15-20s or core hangs on conenction queue.\n"));
				status_message_blue (_("GUI: Don't worry about this msg, it's just a 'play-it-safe' function by the GUI.\n"));
				status_message_blue (_("GUI: (you can switch this off on the options page)\n"));

				/* set server to low priority in any case */
				if (servers_last_attempted_to_connect_to)
					send_set_server_priority (servers_last_attempted_to_connect_to->ip, servers_last_attempted_to_connect_to->port, 2);

				send_disconnect();
				send_command (N_("x"));
			} else status_message_blue (_("GUI: PlacedOnQDeadlock w/o PlacedOnQ before!? (bug!)\n"));
		}
		break;

		case SERVERS_CONNECT_STATUS_THINKS_WE_ARE_FIREWALLED:
		{
			if (servers_last_attempted_to_connect_to)
			{
				servers_last_attempted_to_connect_to->nogood = TRUE;
				send_set_server_priority (	servers_last_attempted_to_connect_to->ip,
											servers_last_attempted_to_connect_to->port,
											SERVER_PRIO_LOW);
			}
		}
		/* Note: no break here! */

		case SERVERS_CONNECT_STATUS_DO_SERVER_ROTATE:
		{
			send_disconnect();	

			if (option_always_connected == 0)
				send_connect_to_server(0,0);	/* connect to any */
		}
		break;

		case SERVERS_CONNECT_STATUS_INVALID:
		default:
		{
			g_printerr (_("GUI: unknown status in %s (bug!)\n"), __FUNCTION__);
		}
		break;
	}
	lastStatus = status;
}


// servers_show_dead_alive_stats
//
// Show a line in the statusbar saying how many dead and alive servers we have.
// Called by the notebook switch page function

void
servers_show_dead_alive_stats (void)
{
	guint	dead=0, alive=0, users=0, files=0;

	if (!servers) return;
	PROCESS_GPTRARRAY ( servers, serverRecord,
		if (item->ping!=0 || (time(NULL)-item->last_seen_alive)<SERVER_WO_STATS_DEAD_SECONDS)
		{
			alive++;
			users += item->users;
			files += item->files;
		} else dead++;
	);
	statusbar_extra_message (_(" Serverlist: %u alive and %u dead servers with %u users sharing %u files."), alive, dead, users,files);
}

