/* servers_page.c
*
*  populates server page in notepad and deals with its functions
*
* (c) 2001-2002 Tim-Philipp Muller <t.i.m@orange.net>
*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// TODO (feature):
// * pop-up menu option "save current server list to server.met"

#include "global.h"
#include "misc.h"
#include "icons.h"
#include "statusbar.h"
#include "status_page.h"
#include "servers_data.h"
#include "servers_page.h"
#include "gui_options.h"
#include "gui_flags.h"
#include "send_command.h"
#include "http_get.h"

#include <sys/types.h>
#include <unistd.h>


// ---------- variables

GtkWidget			*clist_servers = NULL;		/* our server clist */
static GtkWidget	*servers_page_stats_label = NULL;
static GtkWidget	*servers_page_hbox = NULL;

/* lcoal functions */

static GtkWidget	*serversclist_create (void);
static GtkWidget	*server_page_create_buttons_hbox (void);
static gint			 serversclist_key_pressed (GtkWidget *widget, GdkEvent *event);
static void			 serversclist_column_resized (GtkCList *clist, int col, int new_width, gpointer data);
static gint			 serversclist_button_pressed (GtkWidget *widget, GdkEvent *event);
static void			 serversclist_column_clicked (GtkCList *clist, gint column, gpointer data);
static gint			 serversclist_compare_func (GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2);
static GtkWidget	*servers_page_create_stats_label (void);


/*
 *
 *
 */

void
servers_page_set_client_id (const gchar *s)
{
	const gchar	*open, *fw;
	gchar		*s2, *s3;
	guint		 pos;

	g_return_if_fail (s!=NULL);
	g_return_if_fail (servers_page_stats_label!=NULL);

	fw   = strstr(s,N_("(firewalled)"));
	open = strstr(s,N_("(open)"));

	if ((!fw)&&(!open))
	{
		gtk_label_set_text(GTK_LABEL(servers_page_stats_label), s);
		return;
	}

	pos = (fw!=NULL) ? (guint)(fw-s) : (guint)(open-s);

	s2 = g_strdup(s);
	g_strdelimit(s2, N_("("), 0x00);
	s3 = g_strdup_printf(N_("%s%s"), s2, (fw!=NULL) ? _("(firewalled)") : _("(open)"));
	gtk_label_set_text(GTK_LABEL(servers_page_stats_label), s3);
	g_free(s2);
	g_free(s3);
}


// serversclist_popupmenu_item_selected
//
// An item from the servers pop-up menu has been selected. Take some action...
//
// data contains the menuitem selected (lowest 8 bits), and in some cases an additional
//	number (second-lowest 8 bits), and some flags (second-highest 8 bits)
// Note: widget might be NULL

void
serversclist_popupmenu_item_selected (GtkWidget *widget, gpointer data)
{
	guint		menuitem =  GPOINTER_TO_UINT (data) & 0xff;
	guint		number   = (GPOINTER_TO_UINT (data) & 0xff00) >> 8;
	guint		flags    =  GPOINTER_TO_UINT (data) & 0xff0000;
//	gboolean	connect_any = ((flags&SERVERS_MENUITEM_FUNC_CONNECT_ANY_FLAG)!=0);
	gboolean	remove_dead = ((flags&SERVERS_MENUITEM_FUNC_REMOVE_DEAD_FLAG)!=0);
	gboolean	remove_all  = ((flags&SERVERS_MENUITEM_FUNC_REMOVE_ALL_FLAG)!=0);
	gboolean	remove_local = ((flags&SERVERS_MENUITEM_FUNC_REMOVE_LOCAL_FLAG)!=0);
	gboolean	unselect_all = TRUE;

	switch (menuitem)
	{
		case SERVERS_MENUITEM_CONNECT:
		{
			gboolean connectlistWasEmpty = (!connectlist);
			PROCESS_CLIST_SELECTION_LIST (SRV_CLIST, servers, serverRecord,
				// it doesn't make sense to send out multiple connect_to
				//	messages in a row if multi-selection is on
				//	=> we put the servers we want to connect to into
				//	the connect list and then do it one by one
				if (item)
				{
					// We only send a 'connect to' msg right away if
					//	there aren't any other servers in the
					//	connect queue that we haven't tried yet.
					if (g_list_length(SRV_CLIST->selection)>1 || (connectlist))
					{
						GInetAddr	*ia;
						gchar		*ipstr = g_strdup_printf (N_("%u.%u.%u.%u"), item->ip&0xff, (item->ip&0xff00)>>8, (item->ip&0xff0000)>>16, (item->ip&0xff000000)>>24);
						ia = gnet_inetaddr_new (ipstr, item->port);
						if (ia) connectlist = g_list_append (connectlist, (gpointer) ia);
						if (ipstr) g_free (ipstr);
					} else send_connect_to_server (item->ip, item->port);
				}
			)
			if (g_list_length(SRV_CLIST->selection)>1 && (connectlistWasEmpty))
			{
				servers_save_alwaysconnected_option ();
				send_command (N_("auto -"));
				send_disconnect();
				(void) servers_connect_to_next_selected ();
				status_message_blue (_("Trying to connect to one of the selected servers...\n"));
			} else statusbar_extra_message (_(" Added selected servers to list of servers to try to connect to.\n"));
		}
		break;

		case SERVERS_MENUITEM_CONNECT_ANY:
			send_connect_to_server(0,0);
		break;
		case SERVERS_MENUITEM_DISCONNECT:
			send_disconnect();
		break;
		case SERVERS_MENUITEM_REFRESH:
			send_get_serverlist();
		return;	// DO NOT UNSELECT ANY

		case SERVERS_MENUITEM_REMOVE:
		{
			/* list of serverRecord items to remove
			 * we need to make this list first, otherwise we remove records
			 *	while processing the selection list containing ROW NUMBERS,
			 *	and any removal of rows will mess up everything
			 */
			GList	*removelist = NULL;

			if (!servers)
				break;

			/* in case of remove dead or remove all make fake selection list */
			if (remove_dead)
			{
				PROCESS_GPTRARRAY( servers, serverRecord,
					if (item)
					{
						time_t secs_without_stats = time(NULL)-item->last_seen_alive;
						if ( item->ping==0 && secs_without_stats >= SERVER_WO_STATS_DEAD_SECONDS)
							removelist = g_list_append (removelist, (gpointer)item);
					}
				);
			}

			else if (remove_all)
			{
				/* remove all */

				PROCESS_GPTRARRAY( servers, serverRecord,
					if (item)
					{
						removelist = g_list_append (removelist, (gpointer)item);
					}
				);
			}

			else if (remove_local)
			{
				/* remove all servers with local IP */

				PROCESS_GPTRARRAY( servers, serverRecord,
					if ((item) && (misc_ip_is_local(item->ip)))
					{
						removelist = g_list_append (removelist, (gpointer)item);
					}
				);

			} else {

				/* only remove selected items */

				PROCESS_CLIST_SELECTION_LIST (SRV_CLIST, servers, serverRecord,
					if (item)
					{
						removelist = g_list_append (removelist, (gpointer)item);
					}
				);
			}

			if ((remove_dead) && g_list_length(removelist) == servers->len)
			{
				statusbar_extra_message (_(" Didn't remove any servers. Would have removed all - maybe we didn't get stats yet?"));
				break;
			}

			if (clist_servers) gtk_clist_freeze(SRV_CLIST);

			PROCESS_GLIST(removelist, serverRecord,
				if (item)
				{
					send_remove_server (item->ip, item->port);
					servers_remove_record (item);
				}
			);

			if (clist_servers) gtk_clist_thaw(SRV_CLIST);

			send_get_serverlist();
			statusbar_extra_message (_(" Removed %u%s servers (wait for refresh)."), g_list_length(removelist), (remove_dead) ? _(" dead") : N_("") );
			g_list_free(removelist);
		}
		break;

		case SERVERS_MENUITEM_SET_PRIORITY:
		{
			PROCESS_CLIST_SELECTION_LIST (SRV_CLIST, servers, serverRecord,
				if ((item) && (item->pref!=number))
				{
					send_set_server_priority (item->ip, item->port, number);
					item->pref=number;
				}
			);
			statusbar_extra_message (_(" Changed priority of selected servers."));
			// No need to fetch a new list from the core just for this
			serversclist_update ();
		}
		break;
		case SERVERS_MENUITEM_FETCH_SERVERLIST:
		{
			switch (number)
			{
				case 1:	http_get_retrieve_serverlist (option_serverlist_URL,0);
				break;
				case 2:	http_get_retrieve_serverlist (option_serverlist_URL2,0);
				break;
				case 3: http_get_retrieve_serverlist (STANDARD_SERVERLIST_1,0);
				break;
				case 4: http_get_retrieve_serverlist (STANDARD_SERVERLIST_2,0);
				break;
				default: g_printerr (_("GUI: fix me: retrieve unknown server list (SERVERS_MENUITEM_FETCH_SERVERLIST)???\n"));
			}
			statusbar_extra_message (_(" GUI: Trying to fetch serverlist via http..."));
		}
		break;
		case SERVERS_MENUITEM_SELECT_ALL:
		{
			gtk_clist_select_all (SRV_CLIST);
			unselect_all = FALSE;
		}
		break;
		case SERVERS_MENUITEM_UNSELECT_ALL:
		{
			gtk_clist_unselect_all (SRV_CLIST);
		}
		break;
		case SERVERS_MENUITEM_TOGGLE_ALWAYS_CONNECTED:
		{
			option_always_connected = (~option_always_connected) & 1;
			send_auto_connect (option_always_connected);
		}
		break;
		case SERVERS_MENUITEM_TOGGLE_AUTO_REMOVE_DEAD:
		{
			option_autoremove = (~option_autoremove) & 1;
			send_auto_server_remove (option_autoremove);
		}
		break;
		case SERVERS_MENUITEM_TOGGLE_IGNORE_QUEUE_DEADLOCKS:
			option_ignore_queue_deadlock = (~option_ignore_queue_deadlock) & 1;
		break;
		case SERVERS_MENUITEM_TOGGLE_AUTO_CLEAR_ON_FETCH:
			option_clear_slist_when_fetching_new_ones = (~option_clear_slist_when_fetching_new_ones) & 1;
		break;
		case SERVERS_MENUITEM_TOGGLE_FETCH_LIST_WHEN_FEW_SERVERS:
			option_fetch_when_too_few_servers = (~option_fetch_when_too_few_servers)&1;
		break;
		case SERVERS_MENUITEM_TOGGLE_DISCONNECT_ON_LOWID:
			option_disconnect_if_firewalled = (~option_disconnect_if_firewalled)&1;
		break;
		case SERVERS_MENUITEM_TOGGLE_UNSEL_AFTER_ACTION:
			option_SRVUnselectAfterAction = (~option_SRVUnselectAfterAction)&1;
		break;
		case SERVERS_MENUITEM_TOGGLE_SINGLE_MODE:
			option_SRVSingleSelectionMode = (~option_SRVSingleSelectionMode)&1;
			// set multi- or single- selection mode
			if (option_SRVSingleSelectionMode==1)
			{
				gtk_clist_set_selection_mode(SRV_CLIST, GTK_SELECTION_SINGLE);
			} else gtk_clist_set_selection_mode(SRV_CLIST, GTK_SELECTION_EXTENDED);
		break;
		case SERVERS_MENUITEM_TOGGLE_AUTO_REMOVE_INVALID:
			option_autoremoveinvalid = (~option_autoremoveinvalid) & 1;
		break;

		default:
		{
			g_printerr (_("GUI: Fix me - called with unknown value %u (%s)\n"), menuitem, __FUNCTION__);
		}
		break;
	}

	if (option_SRVUnselectAfterAction==1 && option_SRVSingleSelectionMode==0 && (unselect_all))
	{
		gtk_clist_unselect_all (SRV_CLIST);
	}
}


// serversclist_compare_func
//
// Compares two rows in the serverlist based on the sort column set.
// Returns -1, 0, or 1

static gint
serversclist_compare_func (GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2)
{
	serverRecord	*sr1, *sr2;

	g_return_val_if_fail (ptr1!=NULL,0);				// just making sure...
	g_return_val_if_fail (ptr2!=NULL,0);

	sr1 = (serverRecord*) ((GtkCListRow*)ptr1)->data;
	sr2 = (serverRecord*) ((GtkCListRow*)ptr2)->data;

	g_return_val_if_fail (sr1!=NULL,0);					// just making sure...
	g_return_val_if_fail (sr2!=NULL,0);

	switch (serverlist_sortcolumn)
	{
		case 0:	MISC_RETURN_STR_COMPARE(sr1->name,sr2->name);	// 0 - name
		case 1:	MISC_RETURN_NUM_COMPARE(GUINT32_TO_BE(sr1->ip),GUINT32_TO_BE(sr2->ip));		// 1 - ip
		case 2:	MISC_RETURN_NUM_COMPARE(sr1->port,sr2->port);	// 2 - port
		case 3:	MISC_RETURN_NUM_COMPARE(sr1->users,sr2->users);	// 3 - users
		case 4:	MISC_RETURN_NUM_COMPARE(sr1->files,sr2->files);	// 4 - files
		case 5:													// 5 - preference
		{
			const gint pref_values[3] = { 1, 2, 0 }; // normal, high, low
			MISC_RETURN_NUM_COMPARE(pref_values[CLAMP(sr1->pref,0,2)],pref_values[CLAMP(sr2->pref,0,2)]);
		}
		case 6: MISC_RETURN_NUM_COMPARE(sr1->ping,sr2->ping);	// 6 - ping
		case 7: MISC_RETURN_STR_COMPARE(sr1->desc,sr2->desc);	// 7 - description
		default:	
			g_printerr (_("FIXME: unknown server list sort column!?\n"));
		return 0;
	}
	g_assert_not_reached();
}


// serversclist_column_clicked
//
// A column has been clicked. Change the sort column or sort direction and resort the list

static void
serversclist_column_clicked (GtkCList *clist, gint column, gpointer data)
{

	// If Column clicked is the one already active for sorting, change sort order,
	//	otherwise set new sort column with default type=ascending
	if (column != serverlist_sortcolumn)
	{
		serverlist_sortcolumn = column;
		serverlist_sorttype   = GTK_SORT_ASCENDING;
		
	} else serverlist_sorttype = (serverlist_sorttype==GTK_SORT_ASCENDING) ? GTK_SORT_DESCENDING :  GTK_SORT_ASCENDING;

	gtk_clist_set_auto_sort (clist, FALSE);
	gtk_clist_set_sort_column (clist, serverlist_sortcolumn);
	gtk_clist_set_sort_type (clist, serverlist_sorttype);
	gtk_clist_set_compare_func (clist, serversclist_compare_func);	
	gtk_clist_sort (clist);
}



// serversclist_popup_options_menu
//
// Creates the options submenu for the pop-up menu for the serverlist
// row col:  clist row and column where click took place

static GtkWidget *
serversclist_popup_options_menu (guint row, guint col)
{
	GtkWidget	*menu;

	const gchar	*ALWAYS_CONNECTED_STR =			_(" always stay connected ");
	const gchar	*AUTO_REMOVE_DEAD_STR =			_(" automatically remove dead servers ");
	const gchar	*AUTO_REMOVE_INVALID_STR =		_(" automatically remove invalid servers ");
	const gchar	*IGNORE_QUEUE_DEADLOCKS_STR =	_(" ignore 'placed on connection queue...' deadlocks (GUI) ");
	const gchar	*AUTO_CLEAR_SERVERLIST_STR =	_(" auto-clear serverlist when fetching favourite list via http  ");
	const gchar	*FETCH_LIST_WHEN_FEW_STR =		_(" auto-fetch new servers when we have less than 10 servers ");
	const gchar	*DISCONNECT_ON_LOWID_STR =		_(" disconnect from server if it thinks that I'm firewalled ");
	const gchar	*UNSELECT_AFTR_ACTION_STR =		_(" unselect all items after each action ");
	const gchar	*SERVERS_SINGLE_MODE_STR = 		_(" use single selection mode in servers list ");

	gchar		**icon_yes = icon_optionon;
	gchar		**icon_no  = icon_optionoff;
	
	menu  = gtk_menu_new();

	/* This assignment is not 'by the (reference) book', but we don't
	 *	have any other choice to change the pos.func w/o popping
	 *	up the menu, do we?
	 */

	(void)	MENU_ITEM_ADD (ALWAYS_CONNECTED_STR,		serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_ALWAYS_CONNECTED, (option_always_connected==1)? icon_yes : icon_no );
			MENU_ITEM_ADD__________;
	(void)	MENU_ITEM_ADD (AUTO_CLEAR_SERVERLIST_STR,	serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_AUTO_CLEAR_ON_FETCH, (option_clear_slist_when_fetching_new_ones ==1) ? icon_yes : icon_no);
	(void)	MENU_ITEM_ADD (FETCH_LIST_WHEN_FEW_STR, 	serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_FETCH_LIST_WHEN_FEW_SERVERS,  (option_fetch_when_too_few_servers==1) ? icon_yes : icon_no);
			MENU_ITEM_ADD__________;
	(void)	MENU_ITEM_ADD (DISCONNECT_ON_LOWID_STR,		serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_DISCONNECT_ON_LOWID, (option_disconnect_if_firewalled==1) ? icon_yes : icon_no);
	(void)	MENU_ITEM_ADD (AUTO_REMOVE_DEAD_STR,		serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_AUTO_REMOVE_DEAD, (option_autoremove==1) ? icon_yes : icon_no);
	(void)	MENU_ITEM_ADD (AUTO_REMOVE_INVALID_STR,		serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_AUTO_REMOVE_INVALID, (option_autoremoveinvalid==1) ? icon_yes : icon_no);
	(void)	MENU_ITEM_ADD (IGNORE_QUEUE_DEADLOCKS_STR,	serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_IGNORE_QUEUE_DEADLOCKS, (option_ignore_queue_deadlock==1) ? icon_yes : icon_no);
			MENU_ITEM_ADD__________;
	(void)	MENU_ITEM_ADD (UNSELECT_AFTR_ACTION_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_UNSEL_AFTER_ACTION, (option_SRVUnselectAfterAction==1) ? icon_yes : icon_no);
	(void)	MENU_ITEM_ADD (SERVERS_SINGLE_MODE_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_TOGGLE_SINGLE_MODE, (option_SRVSingleSelectionMode==1) ? icon_yes : icon_no);

	return menu;

}


/* serversclist_popup_priority_menu
 *
 * creates a priority sub-menu for the main pop-up
 *
 */

static GtkWidget *
serversclist_popup_priority_menu (guint row, guint col)
{
	const gchar		*SRV_PRIO_HIGH_STR =	_(" high ");
	const gchar		*SRV_PRIO_NORM_STR =	_(" normal ");
	const gchar		*SRV_PRIO_LOW_STR =		_(" low ");
	GtkWidget		*menu = NULL;

	menu  = gtk_menu_new();
	MENU_ITEM_ADD (SRV_PRIO_HIGH_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_SET_PRIORITY|SERVERS_MENUITEM_FUNC_NUMBER_PRIO_HIGH, icon_threestars);
	MENU_ITEM_ADD (SRV_PRIO_NORM_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_SET_PRIORITY|SERVERS_MENUITEM_FUNC_NUMBER_PRIO_NORMAL, icon_twostars);
	MENU_ITEM_ADD (SRV_PRIO_LOW_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_SET_PRIORITY|SERVERS_MENUITEM_FUNC_NUMBER_PRIO_LOW, icon_onestar);

	return menu;
}

/* serversclist_popup_connect_menu
 *
 */

static GtkWidget *
serversclist_popup_connect_menu (guint row, guint col)
{
	const gchar		*SRV_CONNECT_SEL_STR = _(" connect to any selected ");
	const gchar		*SRV_CONNECT_ANY_STR = _(" connect to any ");
	GtkWidget		*menu = NULL;

	menu  = gtk_menu_new();
	(void)MENU_ITEM_ADD (SRV_CONNECT_SEL_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_CONNECT, icon_connect);
	(void)MENU_ITEM_ADD (SRV_CONNECT_ANY_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_CONNECT_ANY, icon_connect);
	return menu;
}

/* serversclist_popup_serverlist_menu
 *
 */

static GtkWidget *
serversclist_popup_serverlist_menu (guint row, guint col)
{
	const gchar     *SRV_FETCH_FIRST_STR =  _(" fetch new serverlist (favourite)");
	const gchar     *SRV_FETCH_SECOND_STR = _(" fetch new serverlist (secondary)");
	const gchar     *SRV_REFRESH_STR =      _(" refresh list (update values) ");
	const gchar		*SRV_REMOVE_LOCAL = 	_(" remove invalid servers ");
	const gchar     *SRV_REMOVE_ALL_STR =   _(" clear serverlist ");
	GtkWidget 		*menu = NULL;

	menu = gtk_menu_new();
	(void)  MENU_ITEM_ADD (SRV_FETCH_FIRST_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_FETCH_SERVERLIST|(1<<8), icon_download);
	(void)  MENU_ITEM_ADD (SRV_FETCH_SECOND_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_FETCH_SERVERLIST|(2<<8), icon_download);
	(void)  MENU_ITEM_ADD (SRV_REFRESH_STR,         serversclist_popupmenu_item_selected, SERVERS_MENUITEM_REFRESH, icon_refresh);
			MENU_ITEM_ADD__________;
	(void)  MENU_ITEM_ADD (SRV_REMOVE_LOCAL, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_LOCAL_FLAG, icon_cancel);
			MENU_ITEM_ADD__________;
	(void)  MENU_ITEM_ADD (SRV_REMOVE_ALL_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_ALL_FLAG, icon_cancel);

	return menu;
}


/* serversclist_popup_menu
 *
 * Creates a pop-up menu for the servers list and pops it up
 * row, col: clist row and column where click took place
 *
 */

static void
serversclist_popup_menu (guint row, guint col, guint button, guint32 activate_time)
{
	// menu widget isn't destroyed automatically => we have to destroy it after use
	GtkWidget		*menu = NULL;
	GtkWidget		*mi[16]; /* if we have more rows than that in the pop-up menu, we should be beaten anyway */
	GtkWidget		*optitem;
	GtkWidget		*listitem;
	GtkWidget		*connectitem;
	guint			 n_prio, n;

	const gchar	*SRV_CONNECT_STR =	_(" connect... ");
	const gchar	*SRV_PRIO_STR =		_(" priority... ");
	const gchar	*SRV_LIST_STR =		_(" server list... ");
	const gchar	*SRV_DISCONNECT_STR = 	_(" disconnect from current server ");
	const gchar	*SRV_REMOVE_SEL_STR =	_(" remove selected servers ");
	const gchar	*SRV_REMOVE_DEAD_STR =	_(" remove dead servers ");
	const gchar	*SRV_OPTIONS_STR = 		_(" server pool options ");

	menu  = gtk_menu_new();

	n=0;

	/* connect... */
	connectitem =	MENU_ITEM_ADD (SRV_CONNECT_STR, NULL, 0, icon_connect);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(connectitem), serversclist_popup_connect_menu(row,col));
	(void)			MENU_ITEM_ADD__________;

	/* disconnect */
	(void)		MENU_ITEM_ADD (SRV_DISCONNECT_STR,	serversclist_popupmenu_item_selected, SERVERS_MENUITEM_DISCONNECT, icon_disconnect);
	(void)		MENU_ITEM_ADD__________;

	/* priority... */
	n_prio = n;
	mi[n++] =	MENU_ITEM_ADD (SRV_PRIO_STR, NULL, 0, icon_twostars);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi[n_prio]), serversclist_popup_priority_menu(row,col));
	(void)  	MENU_ITEM_ADD__________;

	/* server list... */
	listitem = 	MENU_ITEM_ADD (SRV_LIST_STR, NULL, 0, icon_servers);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(listitem), serversclist_popup_serverlist_menu(row,col));
	(void)  	MENU_ITEM_ADD__________;

	/* remove selected */
	mi[n++] =	MENU_ITEM_ADD (SRV_REMOVE_SEL_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_REMOVE, icon_cancel);

	/* remove dead servers */
	(void)		MENU_ITEM_ADD (SRV_REMOVE_DEAD_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_REMOVE|SERVERS_MENUITEM_FUNC_REMOVE_DEAD_FLAG, icon_skull);
	(void)		MENU_ITEM_ADD__________;

	optitem =	MENU_ITEM_ADD (SRV_OPTIONS_STR, NULL, 0, icon_options);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM(optitem), serversclist_popup_options_menu(row,col));

	/* No items selected? => make those functions that need items selected insensitive */
	if (!SRV_CLIST->selection)
	{
		guint i;
		for (i=0; i<n; i++)
			gtk_widget_set_sensitive (GTK_WIDGET(mi[i]), FALSE);
	}

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL,   NULL   , NULL, button, activate_time);
	
}


/* servers_toolbar_icon_popup_menu
 *
 * Creates a pop-up menu for the servers icon in the toolbar
 *
 */

void
servers_toolbar_icon_popup_menu (guint button, guint32 activate_time)
{
	GtkWidget		*menu = NULL;

	const gchar		*SRV_CONNECT_ANY_STR = _(" connect to any ");
	const gchar		*SRV_DISCONNECT_STR = 	_(" disconnect from current server ");
	const gchar     *SRV_FETCH_FIRST_STR =  _(" fetch new serverlist (favourite)");
	const gchar     *SRV_FETCH_SECOND_STR = _(" fetch new serverlist (secondary)");
	const gchar     *SRV_REFRESH_STR =      _(" refresh list (update values) ");

	menu  = gtk_menu_new();

	(void) MENU_ITEM_ADD (SRV_CONNECT_ANY_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_CONNECT_ANY, icon_connect);
	(void) MENU_ITEM_ADD (SRV_DISCONNECT_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_DISCONNECT, icon_disconnect);
	(void) MENU_ITEM_ADD__________;
	(void) MENU_ITEM_ADD (SRV_FETCH_FIRST_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_FETCH_SERVERLIST|(1<<8), icon_download);
	(void) MENU_ITEM_ADD (SRV_FETCH_SECOND_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_FETCH_SERVERLIST|(2<<8), icon_download);
	(void) MENU_ITEM_ADD__________;
	(void) MENU_ITEM_ADD (SRV_REFRESH_STR, serversclist_popupmenu_item_selected, SERVERS_MENUITEM_REFRESH, icon_refresh);

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL,   NULL   , NULL, button, activate_time);

}


// serversclist_button_pressed
//
// Shows the Servers pop-up menu on right-click on the clist
// Connects to server on double-click on row (if double-click not disabled in options)
//
// This right-click pop-up menu code was inspired by code posted by Mike Hicks to
//	the gtk-list@redhat.com mailing list on 25/11/98. Thank you, Mike :)

static gint
serversclist_button_pressed (GtkWidget *widget, GdkEvent *event)
{
	gint row, col;

	last_time_of_user_click = time(NULL);

	if (event->type == GDK_BUTTON_PRESS)
	{
		// convert GdkEvent to GdkEventButton
		GdkEventButton *bevent = (GdkEventButton *) event;
		if (bevent->button == 3)
		{
			gtk_clist_get_selection_info (SRV_CLIST, (int) bevent->x ,(int) bevent->y, &row, &col);
			if ((SRV_CLIST) && g_list_length(SRV_CLIST->selection)<=1)
			{
				gtk_clist_unselect_all (SRV_CLIST);
				gtk_clist_select_row (SRV_CLIST,row,1);
			}
			serversclist_popup_menu(row, col, bevent->button, bevent->time);	
		}
		return TRUE;	 // event has been dealt with
	}

	// double-click? => connect to this one server
	if ((event->type) == GDK_2BUTTON_PRESS && option_disable_doubleclick==0)
	{
		// convert GdkEvent to GdkEventButton
		GdkEventButton *bevent = (GdkEventButton *) event;
		g_return_val_if_fail(SRV_CLIST!=NULL,FALSE);
		gtk_clist_get_selection_info (SRV_CLIST, (int) bevent->x ,(int) bevent->y, &row, &col);
		if ((servers) && row < servers->len && row>=0)	// otherwise click on empty space
		{
			GList			*original_selection, *fake_selection;

			original_selection = SRV_CLIST->selection;
			// Make fake selection with only one item for downloads_popupmenu_item_selected()
			fake_selection = g_list_append (NULL, GUINT_TO_POINTER(row));
			SRV_CLIST->selection = fake_selection;
			serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(SERVERS_MENUITEM_CONNECT));
			if (fake_selection) g_list_free (fake_selection);
			SRV_CLIST->selection = original_selection;
		}
		return TRUE;
	}

	return FALSE;
}


// servers_column_resized
//
// A column has been resized. Save new width so that we can save it with options.

static void
serversclist_column_resized (GtkCList *clist, int col, int new_width, gpointer data)
{
	column_widths[SERVER_COLS][col] = new_width;
}


// serversclist_key_pressed
//
// a key has been pressed in the server list

static gint
serversclist_key_pressed (GtkWidget *widget, GdkEvent *event)
{
	guint	action = SERVERS_MENUITEM_INVALID;
	guint	key = event->key.keyval;
	if (key>='A' && key<='Z') key = key + ('a'-'A');	// to lower case
	last_time_of_user_click = time(NULL);
	switch (key)
	{
		case DELETE_KEY:	action = SERVERS_MENUITEM_REMOVE;
					break;
		case 'h':		action = SERVERS_MENUITEM_SET_PRIORITY|SERVERS_MENUITEM_FUNC_NUMBER_PRIO_HIGH;
					break;
		case 'l':		action = SERVERS_MENUITEM_SET_PRIORITY|SERVERS_MENUITEM_FUNC_NUMBER_PRIO_LOW;
					break;
		case 'n':		action = SERVERS_MENUITEM_SET_PRIORITY|SERVERS_MENUITEM_FUNC_NUMBER_PRIO_NORMAL;
					break;
		case 'c':		action = SERVERS_MENUITEM_CONNECT;
					break;
		case 'u':		action = SERVERS_MENUITEM_UNSELECT_ALL;
					break;

		default:		return FALSE;
	}
	serversclist_popupmenu_item_selected (NULL, GUINT_TO_POINTER(action));
	return TRUE; // has been dealt with - if key is not in list we return false right away
}


/* servers_page_create_stats_label
 *
 *
 */

static GtkWidget *
servers_page_create_stats_label (void)
{
	GtkWidget	*hbox, *emptylabel, *slabel;

	hbox = gtk_hbox_new(FALSE,0);
	gtk_widget_show(hbox);

	slabel = gtk_label_new(_(" Servers - use the toolbar or drag'n'drop to add a server "));
	gtk_label_set_justify (GTK_LABEL(slabel), GTK_JUSTIFY_LEFT);
	gtk_widget_show(slabel);

	emptylabel = gtk_label_new(N_(" "));
	gtk_widget_show(emptylabel);

	servers_page_stats_label = gtk_label_new (NULL);
	gtk_label_set_justify (GTK_LABEL(servers_page_stats_label), GTK_JUSTIFY_RIGHT);
	gtk_widget_show(servers_page_stats_label);

	gtk_box_pack_start (GTK_BOX(hbox), slabel, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(hbox), emptylabel, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(hbox), servers_page_stats_label, FALSE, FALSE, 0);

	if (option_nogreen == 0)
		set_style_recursively (hbox, style_notebook_label);

	return hbox;
}


/* serversclist_create
 *
 *
 *
 */

static GtkWidget *
serversclist_create (void)
{
	GtkWidget	 *scrollwin;
	gchar		**tabtitles = g_strsplit ( _("name|ip|port|users|files|preference|ping|description"), N_("|"), 0);
	gint		  i;

	g_return_val_if_fail (tabtitles!=NULL,NULL);

	// creating server clist window and set its properties

	clist_servers = gtk_clist_new_with_titles (NO_OF_SERVER_COLS, tabtitles);	/* 7 columns */
	gtk_signal_connect (GTK_OBJECT (clist_servers), N_("click-column"),
			    GTK_SIGNAL_FUNC(serversclist_column_clicked), NULL);
	gtk_signal_connect (GTK_OBJECT (clist_servers), N_("button_press_event"),
				GTK_SIGNAL_FUNC(serversclist_button_pressed), GTK_OBJECT(clist_servers));
	gtk_signal_connect (GTK_OBJECT (clist_servers), N_("key_press_event"),
				GTK_SIGNAL_FUNC(serversclist_key_pressed), GTK_OBJECT(clist_servers));
	gtk_signal_connect (GTK_OBJECT (clist_servers), N_("resize-column"),
				GTK_SIGNAL_FUNC(serversclist_column_resized), NULL);

	gtk_clist_set_auto_sort (GTK_CLIST(clist_servers),FALSE);
	gtk_clist_set_compare_func (GTK_CLIST(clist_servers), serversclist_compare_func);
	gtk_clist_sort (GTK_CLIST(clist_servers));

	// set multi- or single- selection mode
	if (option_SRVSingleSelectionMode==1)
	{
		gtk_clist_set_selection_mode(SRV_CLIST, GTK_SELECTION_SINGLE);
	} else gtk_clist_set_selection_mode(SRV_CLIST, GTK_SELECTION_EXTENDED);

	/* set columns resizable by user (which is not default in gtk2) */
	for (i=0; i<NO_OF_SERVER_COLS; i++)
	{
		gtk_clist_set_column_resizeable(GTK_CLIST(clist_servers), i, TRUE);
	}

	// set default column widths
	set_clist_columns (GTK_CLIST(clist_servers), SERVER_COLS, NO_OF_SERVER_COLS);

	// create a scroll window to contain the clist later on
	scrollwin = gtk_scrolled_window_new (NULL,NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrollwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(scrollwin), clist_servers);
	gtk_widget_show (clist_servers);
	gtk_widget_show (scrollwin);

	g_strfreev(tabtitles);

	return (scrollwin);
}

// servers_entry_widget_changed
//
//

static void
servers_entry_widget_changed (GtkWidget *widget, gpointer data)
{
	return;
}

static GtkWidget *
server_page_create_buttons_hbox (void)
{
	GtkWidget	*button_newlist, *button_cleardead, *label;
	GtkWidget	*hbox_servers, *button_connect, *button_disconnect;

	hbox_servers = gtk_hbox_new (FALSE, 5);

	/* remove dead servers button */
	new_icon_button_pack_and_signal_connect(
						&button_cleardead,
						hbox_servers,
						_(" remove dead "),
						serversclist_popupmenu_item_selected,
						GUINT_TO_POINTER( SERVERS_MENUITEM_REMOVE | SERVERS_MENUITEM_FUNC_REMOVE_DEAD_FLAG ),
						0,
						icon_skull);

	/* fetch serverlist button */
	new_icon_button_pack_and_signal_connect (
						&button_newlist,
						hbox_servers,
						_(" fetch serverlist "),
						serversclist_popupmenu_item_selected,
						GUINT_TO_POINTER( SERVERS_MENUITEM_FETCH_SERVERLIST | (1<<8) ),
						0,
						icon_download);

	/* connect any button */
	new_icon_button_pack_and_signal_connect (
						&button_connect,
						hbox_servers,
						_(" connect "),
						serversclist_popupmenu_item_selected,
						GUINT_TO_POINTER( SERVERS_MENUITEM_CONNECT_ANY ),
						0,
						icon_connect);

	/* disconnect */
	new_icon_button_pack_and_signal_connect (
						&button_disconnect,
						hbox_servers,
						_(" disconnect "),
						serversclist_popupmenu_item_selected,
						GUINT_TO_POINTER( SERVERS_MENUITEM_DISCONNECT ),
						0,
						icon_disconnect);

	label = gtk_label_new (_(" you can also click on the toolbar server icon "));
	gtk_widget_show(label);
	gtk_box_pack_end (GTK_BOX(hbox_servers), label, TRUE, TRUE, 0);
	if (option_nogreen == 0)
		set_style_recursively (label, style_notebook_label);

	gtk_widget_show (hbox_servers);
	servers_page_hbox = hbox_servers;
	return (hbox_servers);
}



/* servers_page_create
 *
 *
 *
 */

GtkWidget *
servers_page_create (void)
{
	GtkWidget *vbox; 

	vbox = gtk_vbox_new (FALSE, 5);	/* FALSE = not homogenous */
	gtk_widget_show (vbox);

	gtk_box_pack_start (GTK_BOX(vbox), servers_page_create_stats_label(), FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), serversclist_create(), TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), server_page_create_buttons_hbox (), FALSE, FALSE, 5);

	return vbox;
}



// serversclist_find_row_from_record
//
// finds the row of a server in the clist from the server record
// returns -1 if not found, otherwise row

gint
serversclist_find_row_from_record (serverRecord *item)
{
	return (item==NULL) ? -1 : gtk_clist_find_row_from_data (SRV_CLIST, (gpointer)item);
}



// serversclist_update
//
// Sync the GUI's server clist with the server records in 'servers'

void
serversclist_update (void)
{
	gchar			 *content[NO_OF_SERVER_COLS] = { N_(""), N_(""), N_(""), N_(""), N_(""), N_(""), N_(""), N_(""),  };
	static gchar	**pref_str = NULL;

	if (!pref_str)
	{
		pref_str = g_strsplit ( _("normal|high|low"), N_("|"), 0);
		g_return_if_fail (pref_str!=NULL);
	}

	if (!clist_servers)
		return;

	gtk_clist_freeze(SRV_CLIST);
	PROCESS_GPTRARRAY( servers, serverRecord,
		gchar	coltxt[48];
		gint	row = serversclist_find_row_from_record (item);

		if (!item) continue;
		if (row<0)
		{
			row = gtk_clist_append (SRV_CLIST, content);
			gtk_clist_set_row_data (SRV_CLIST, row, (gpointer)item);
		}
		gtk_clist_set_text (SRV_CLIST, row, 0, (item->name) ? item->name : N_(""));
		g_snprintf (coltxt, 48, N_("%u.%u.%u.%u"), item->ip&0xff, (item->ip&0xff00)>>8, (item->ip&0xff0000)>>16, (item->ip&0xff000000)>>24);
		gtk_clist_set_text (SRV_CLIST, row, 1, coltxt);
		g_snprintf (coltxt, 48, N_("%u"), item->port);
		gtk_clist_set_text (SRV_CLIST, row, 2, coltxt);
		g_snprintf (coltxt, 48, N_("%u"), item->users);
		gtk_clist_set_text (SRV_CLIST, row, 3, coltxt);
		g_snprintf (coltxt, 48, N_("%u"), item->files);
		gtk_clist_set_text (SRV_CLIST, row, 4, coltxt);
		gtk_clist_set_text (SRV_CLIST, row, 5, pref_str[CLAMP(item->pref,0,2)]);
		g_snprintf (coltxt, 48, N_("%u"), item->ping);
		gtk_clist_set_text (SRV_CLIST, row, 6, coltxt);

		// mark servers with no stats dead'n'red if we haven't seen stats for 60s
		if (item->ping>0 && item->users>0 && item->files>0)
		{
			strncpy (coltxt, (item->desc) ? item->desc : N_(""), 48);			
			item->last_seen_alive = time(NULL);
		} else g_snprintf (coltxt, 48, _("no stats for %u secs"), (guint) (time(NULL) - item->last_seen_alive));
		gtk_clist_set_text (SRV_CLIST, row, 7, coltxt);
	
		if (item->ping==0 && item->users==0 && item->files==0
		    && (time(NULL)-item->last_seen_alive)>=SERVER_WO_STATS_DEAD_SECONDS)
		{
			gtk_clist_set_foreground (SRV_CLIST, row, &red);
		} else gtk_clist_set_foreground (SRV_CLIST, row, &black);
	);
	gtk_clist_set_auto_sort (SRV_CLIST, FALSE);
	gtk_clist_set_sort_column (SRV_CLIST, serverlist_sortcolumn);
	gtk_clist_set_sort_type (SRV_CLIST, serverlist_sorttype);
	gtk_clist_set_compare_func (SRV_CLIST, serversclist_compare_func);
	gtk_clist_sort (SRV_CLIST);
	gtk_clist_thaw(SRV_CLIST);
}


/******************************************************************************
 *
 *   serversclist_set_overnode_mode
 *
 *
 */

void
serversclist_set_overnet_mode (gboolean overnetOn)
{
	if (!clist_servers)
		return;

	gtk_widget_set_sensitive(clist_servers, (!overnetOn));

	if (overnetOn)
	{
		gtk_widget_hide(servers_page_hbox);
	}
	else
	{
		gtk_widget_show(servers_page_hbox);
	}
}












