/***************************************************************************
                          shared_dirs.c  -  description
                             -------------------
    begin                : Sat Jul 27 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "shared_dirs.h"
#include "shared_page.h"

GPtrArray	*shareddirs = NULL;	// contains pointers to gchar* directorynames


static void		shareddirs_free_record (shareddirRecord *sdr);
static void		shareddirsclist_remove_record (shareddirRecord *sdr);

// shareddirs_find_from_dirname
//
// finds shareddirRecord record from directory string
// returns record or NULL if not found

shareddirRecord *
shareddirs_find_from_dirname (const gchar *dirname)
{
	if ((shareddirs) && (dirname))
	{
		PROCESS_GPTRARRAY (shareddirs, shareddirRecord,
			if ((item) && (item->dirname) && strcmp(dirname,item->dirname)==0)
				return item;
		);
	}
	return NULL;
}


// shareddirs_free_record
//
// free a shareddirRecord

static void
shareddirs_free_record (shareddirRecord *sdr)
{
	g_return_if_fail (sdr!=NULL);

	G_FREE(sdr->dirname);
	memset (sdr,0x00,sizeof(sdr));
	g_free(sdr);
}


// shareddirs_remove_record
//
// remove record from shared dirs list and from clist

void
shareddirs_remove_record (shareddirRecord *sdr)
{
	g_return_if_fail (sdr!=NULL);
	g_return_if_fail (shareddirs!=NULL);
	shareddirsclist_remove_record (sdr);
	g_ptr_array_remove (shareddirs, (gpointer)sdr);
	shareddirs_free_record (sdr);
}

/*******************************************************************************
 *
 *   shareddirs_send_remove_records_sort_func
 *
 *   callback for g_slist_sort() to sort a linked list of shareddirRecords
 *    in reverse order of their index numbers (highest numbers first)
 */

static gint
shareddirs_send_remove_records_sort_func (gconstpointer p1, gconstpointer p2)
{
	shareddirRecord *sdr1 = (shareddirRecord*)p1;
	shareddirRecord *sdr2 = (shareddirRecord*)p2;
	gint             n1, n2;
	
	g_return_val_if_fail (p1!=NULL,0);
	g_return_val_if_fail (p2!=NULL,0);

	n1 = sdr1->idxnum;	/* below comparison makes ascending order - just changing the */
	n2 = sdr2->idxnum;	/* items to compare around will make it descending...         */
	
	if (n1>n2)
	{
		return (-1);
	}
	return ((gint) (n2>n1));
}

/*******************************************************************************
 *
 *   shareddirs_send_remove_records
 *
 */

void
shareddirs_send_remove_records (GSList *recordlist)
{
	GSList *node;
	
	g_return_if_fail (recordlist != NULL);

	node = g_slist_sort (recordlist, shareddirs_send_remove_records_sort_func);
	while (node)
	{
		shareddirRecord *sdr = (shareddirRecord*) node->data;
		if (sdr)
		{
			status_message_blue (_("GUI: unsharing directory '%s' (#%u)\n"), sdr->dirname, sdr->idxnum);
			send_unshare_dir_no(sdr->idxnum);

		}
		node = node->next;
	}

	shareddirs_send_get (2000);

}


/*******************************************************************************
 *
 *   shareddirs_send_get
 *
 *   updates the list of shared dirs in X millisecs
 *
 */

static gint
shareddirs_send_get_callback (gpointer data)
{
	send_get_shared_dirs ();
	return 0; /* don't call us again */
}

void
shareddirs_send_get (guint millisecs)
{
	if (millisecs == 0)
	{
		send_get_shared_dirs ();
		return;
	}
	
	(void) gtk_timeout_add(millisecs, shareddirs_send_get_callback, NULL);
}


/*******************************************************************************
 *
 *   shareddirs_add_subdirs_to_list
 *
 *   Takes a list of sharedDirRecord items and adds all 
 *    subdirectories of 'dirname' that we share to that list
 *
 */
 
GSList *
shareddirs_add_subdirs_to_list (GSList *list, const gchar *dirname)
{
	GSList *newlist = list;
	gint    topdir_len;
	
	g_return_val_if_fail (dirname != NULL, newlist);
	g_return_val_if_fail (shareddirs != NULL, newlist);	
	g_return_val_if_fail (shareddirs->len >= 1, newlist);	

	topdir_len = strlen(dirname);
	
	PROCESS_GPTRARRAY ( shareddirs, shareddirRecord,
		if ((item) && (item->dirname))
		{
			gint checkdir_len = strlen(item->dirname);
			gint comparelen = MIN(topdir_len, checkdir_len);
			
			/* a candidate for a subdirectory must have at least two chars more than the super-dir (slash+one char) */
			if ((topdir_len+2 <= checkdir_len) && (strncmp(item->dirname, dirname, comparelen) == 0))
			{
				if (g_slist_index(newlist,(gpointer)item) == -1)
				{
					newlist = g_slist_append (newlist, (gpointer)item);
				}
			}
		}
	);
	
	return newlist;
}



// shareddirsclist_remove_record
//
// removes a record from the clist

static void
shareddirsclist_remove_record (shareddirRecord *sdr)
{
	gint row;
	g_return_if_fail (sdr!=NULL);
	row = gtk_clist_find_row_from_data (GTK_CLIST(clist_shared), (gpointer)sdr);
	if (row>=0) gtk_clist_remove (GTK_CLIST(clist_shared), row);
}


// shareddirsclist_update
//
// updates the shared dirs clist

void
shareddirsclist_update (void)
{
	if (!clist_shared) return;

	gtk_clist_freeze (GTK_CLIST(clist_shared));

	// add records that are in shared dirs list, but not in clist
	if (shareddirs)
	{
		PROCESS_GPTRARRAY ( shareddirs, shareddirRecord,
			if ((item) && (item->dirname) 
					&& gtk_clist_find_row_from_data(GTK_CLIST(clist_shared),(gpointer)item)<0)
			{
				gint newrow = gtk_clist_append (GTK_CLIST(clist_shared), &item->dirname );
				gtk_clist_set_row_data (GTK_CLIST(clist_shared), newrow, (gpointer)item);
			}
		);
	} else gtk_clist_clear (GTK_CLIST(clist_shared));
	gtk_clist_thaw (GTK_CLIST(clist_shared));
}


// shareddirs_cleanup
//
// clear shared dirs list and clist

void
shareddirs_cleanup (void)
{
	if (shareddirs)
	{
		PROCESS_GPTRARRAY_FROM_END (shareddirs, shareddirRecord,
			if (item)
				shareddirs_remove_record(item);
		);
	}
	if (clist_shared)
		gtk_clist_clear (GTK_CLIST(clist_shared));
}


