/* shared_page.c
*
*  populates shared page in notepad and deals with its functions
*
*  (c) 2001 Tim-Philipp Muller <t.i.m@orange.net>
*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "misc.h"
#include "misc_gtk.h"
#include "gui_options.h"
#include "shared_dirs.h"
#include "shared_files.h"
#include "shared_page.h"
#include "status_page.h"
#include "statusbar.h"
#include "gui_flags.h"
#include "guicoreconn.h"
#include "send_command.h"

#include <ctype.h>


// global variables

int  sharedlist_sortcolumn = 0; /* we don't need that at the moment */
int  sharedlist_sorttype   = GTK_SORT_ASCENDING;

GtkWidget *clist_shared;	/* clist for shared dires */
GtkWidget *spanedwin;		/* paned window for both  */
GtkWidget *ew_addshared;        /* we need this global, so the pop-up menu functions can */
                                /* access it (in case the core runs not on localhost)    */
GtkWidget *shared_hbox;         /* global so we can hide it if on local host             */

static gchar		*lastdir = NULL;
static GtkWidget	*shared_page_stats_label = NULL;

/* functions */

static void	 shared_add_this_dir (char *path, int add_whole_structure);


static void
shared_clist_column_clicked (GtkCList *clist, gint column, gpointer data)
{
	last_time_of_user_click = time(NULL);

	/* column that is already active for sorting? -> change sort type */

	if (column == sharedlist_sortcolumn) {
		if (sharedlist_sorttype == GTK_SORT_ASCENDING)
			sharedlist_sorttype = GTK_SORT_DESCENDING;
		else
			sharedlist_sorttype = GTK_SORT_ASCENDING;
	} else {
		/* set new sort column; default type = ascending */
		sharedlist_sortcolumn = column;
		sharedlist_sorttype   = GTK_SORT_ASCENDING;
	}

	gtk_clist_set_auto_sort (clist, TRUE);
	gtk_clist_set_sort_column (clist, sharedlist_sortcolumn);
	gtk_clist_set_sort_type (clist, sharedlist_sorttype);
	gtk_clist_sort (clist);
}


static int
check_if_path_is_allowed (char *addpath)
{
	/* check if path is one we don't want to allow */

	if ( ( (addpath[0]			== 0x00 ) | 	/* path entered was '/' */
	       (strncmp(addpath,N_("/dev"),4)  	== 0 ) |
	       (strncmp(addpath,N_("/etc"),4)  	== 0 ) |
	       (strncmp(addpath,N_("/sbin"),5) 	== 0 ) |
	       (strncmp(addpath,N_("/bin"),4)  	== 0 ) |
	       (strncmp(addpath,N_("/boot"),5) 	== 0 ) |
	       (strncmp(addpath,N_("/usr/bin"),8) 	== 0 ) |
	       (strncmp(addpath,N_("/usr/sbin"),9)	== 0 ) |
	       (strcmp (addpath,N_("/home"))		== 0 )) != 0) {
		ok_dialog (_("add path to shared"), _("I don't think this is a sensible "
			   "path to share. Sorry."));
		return (0);
	}

	/* check if path starts with a '/' */

	if ((strcmp(clienthost,N_("localhost"))==0) && (addpath[0] != '/')) {
		ok_dialog (_("add path to shared"), _("Paths in linux/unix usually start with a '/'\n"
			   "Example: /home/joe/documents/public\n"
			   "Type the path you want to share in the box at the bottom left before you hit the 'share' button"));
		return (0);
	}

	/* home path shared -> caution! */

	if (strncmp(addpath,N_("/home/"),6) == 0) {
		if (strstr (addpath+6, N_("/")) == NULL ) {	/* meaning: string is /home/xyz   */
			ok_dialog (_("add path to shared"), _("You are trying to share your home "
			           "directory or your _whole_ home directory structure.\n"
				   "This is a security risk, as the donkey will automatically "
				   "include all hidden files and hidden subdirectories,\nwhich "
				   "might contain secret information like usernames and "
				   "passwords (think about the configuration file of\n your "
				   "mail client, or your .ssh subdirectory, or your .gnupg "
				   "subdirectory if you are using pgp public keys...).\n"
				   "However, you can share specific folders from your home "
				   "directory if you enter something like "
				   "\"/home/joe/sharethis\""));
			return (0);
		}
	}

	/* check if path exists */

	if ((strcmp(clienthost, N_("localhost"))==0) && (access(addpath,F_OK | R_OK))) {
		ok_dialog ( _("add path to shared"),
		_("The path you entered does either not exist or you don't have read permissions for it"));
		return (0);
	}

	return (1);
}


static void
shared_add_shared_button_clicked (GtkWidget *widget, gpointer data)
{
	gchar	*add_dir;

	g_return_if_fail (data!=NULL);

	add_dir = gtk_editable_get_chars (GTK_EDITABLE(data), 0, -1);
	g_return_if_fail (add_dir!=NULL);

	last_time_of_user_click = time(NULL);
        shared_add_this_dir (add_dir, 0);
        gtk_entry_set_text (GTK_ENTRY(data), SHARED_PATH_ENTRY_WIDGET_DEFAULT_TEXT);
	g_free(add_dir);
}


static void 
shared_add_structure_button_clicked (GtkWidget *widget, gpointer data)
{
	gchar	*add_dir;

	g_return_if_fail (data!=NULL);

	add_dir = gtk_editable_get_chars (GTK_EDITABLE(data), 0, -1);
	g_return_if_fail (add_dir!=NULL);

	last_time_of_user_click = time(NULL);
        shared_add_this_dir (add_dir, 1);
        gtk_entry_set_text (GTK_ENTRY(data), SHARED_PATH_ENTRY_WIDGET_DEFAULT_TEXT);
	g_free(add_dir);
}


/* POP-UP MENU FUNCTIONS -----------------------------------------------------------------------*/

static void
shared_add_this_dir (char *path, int add_whole_structure) // auxiliary to file select dialg
{
	gint 	 plen;
	gchar	*addpath;

	if (!path) return;
	plen = strlen(path);
	addpath = alloca (plen+10);
        strncpy (addpath,path,plen+10);

        // chop of tailing '/' if there
	if (addpath[plen-1] == '/') addpath[plen-1] = 0x00;

	if (check_if_path_is_allowed(addpath) == 1)
	{
		if (add_whole_structure == 1)
		{
 			/* with overnet core v43 (from 27 Jan 2003) 'a+' command is broken (patch submitted to Jed) */
			status_message_blue (_("Adding directory tree to shared files. If this does not work, it's the core's fault (really)."));
			send_add_shared_dir_structure (addpath);
		} 
		else
		{
			send_add_shared_dir (addpath);
		}

		shareddirs_send_get(3000);
		statusbar_extra_message (_(" The core is hashing the newly shared files..."));
		// filelist will update automatically, because we'll get a
		// "Done loading shared file message" when the new files have been hashed
	}
}

static void
shared_dir_select_dialogue_ok_button_clicked (GtkWidget *widget, GtkFileSelection *fs)
{
	if (lastdir) g_free(lastdir);
	lastdir = g_strdup(gtk_file_selection_get_filename (GTK_FILE_SELECTION(fs)));
	shared_add_this_dir (lastdir, 0);
	gtk_widget_destroy (GTK_WIDGET(fs));
}

static void
shared_select_dialog_cancel_button_clicked (GtkWidget *widget, gpointer gdata)
{
	if (lastdir) g_free(lastdir);
	lastdir = g_strdup(gtk_file_selection_get_filename (GTK_FILE_SELECTION(widget)));
	gtk_widget_destroy (widget);
}

static void
shared_tree_select_dialogue_ok_button_clicked (GtkWidget *widget, GtkFileSelection *fs)
{
	if (lastdir) g_free(lastdir);
	lastdir = g_strdup(gtk_file_selection_get_filename (GTK_FILE_SELECTION(fs)));
	shared_add_this_dir (lastdir, 1);
	gtk_widget_destroy (GTK_WIDGET(fs));
}

static void
shared_dir_select_dialogue (int add_structure_flag)
{
	GtkWidget	*filew;
	gchar		*title = g_strdup_printf (_(" Select a directory%s to share "),
							(add_structure_flag==0) ? N_("") : _(" tree"));

	g_return_if_fail (title!=NULL);
	filew = gtk_file_selection_new (title);

	if (!lastdir) lastdir = g_strdup_printf(N_("%s%c"), g_get_home_dir(), SLASH_CHAR);
	gtk_file_selection_set_filename (GTK_FILE_SELECTION(filew), lastdir);

	// cancel = destroy dialogue widget
	gtk_signal_connect_object (GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), N_("clicked"),
                             (GtkSignalFunc) shared_select_dialog_cancel_button_clicked, GTK_OBJECT(filew));
	// ok button
	gtk_signal_connect (GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), N_("clicked"),
		(add_structure_flag==0) ? (GtkSignalFunc) shared_dir_select_dialogue_ok_button_clicked
		: (GtkSignalFunc) shared_tree_select_dialogue_ok_button_clicked, filew);

	// from the gtk-app-devel-list, 25-03-99, by Kemal 'Disq' Hadimili - thank you. */
	gtk_widget_hide(GTK_FILE_SELECTION(filew)->file_list);
	gtk_widget_hide(GTK_FILE_SELECTION(filew)->file_list->parent);

	gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION(filew));
	gtk_widget_show(filew);
	g_free(title);
}

static void
shared_menuitem_add_dir (GtkWidget *widget, gpointer data)
{
        shared_dir_select_dialogue (0);
}

static void
shared_menuitem_add_tree (GtkWidget *widget, gpointer data)
{
        shared_dir_select_dialogue (1);
}


/*******************************************************************************
 *
 *   shared_menuitem_unshare_selected_dirs
 *
 *   unshare selected directories
 */

static void
shared_menuitem_unshare_selected_dirs (GtkWidget *widget, gpointer data)
{
	GList  *node;
	GSList *newlist = NULL;
	
	g_return_if_fail (clist_shared != NULL);
	
	node = GTK_CLIST(clist_shared)->selection;
	g_return_if_fail (node != NULL);
	
	while (node)
	{
		shareddirRecord	*sdr = (shareddirRecord*) gtk_clist_get_row_data(GTK_CLIST(clist_shared),GPOINTER_TO_INT(node->data));
		if (sdr)
		{
			if (g_slist_index(newlist,(gpointer)sdr) == -1)
			{
				newlist = g_slist_append(newlist, (gpointer)sdr);
			}
		}
		node = node->next;
	}		
	
	shareddirs_send_remove_records (newlist);
	statusbar_extra_message (_(" Unshared selected directories."));
}



/*******************************************************************************
 *
 *   shared_menuitem_unshare_selected_dir_trees
 *
 *   unshare selected directories and all of their subdirectories
 *
 */

static void
shared_menuitem_unshare_selected_dir_trees (GtkWidget *widget, gpointer data)
{
	GSList *sdlist=NULL;
	GList  *node;
	
	if (!clist_shared) 
		return;

	/* first, make a list of shareddirRecords of the directories selected */
	node = GTK_CLIST(clist_shared)->selection;
	while (node)
	{
		gint		row = GPOINTER_TO_INT(node->data);
		gpointer	sdr = gtk_clist_get_row_data(GTK_CLIST(clist_shared),row);
		if (sdr) 
		{
			if (g_slist_index(sdlist,(gpointer)sdr) == -1)
			{
				sdlist = g_slist_append (sdlist, sdr);
			}
			sdlist = shareddirs_add_subdirs_to_list (sdlist, ((shareddirRecord*)sdr)->dirname);
			
		}
		node=node->next;
	}
	
	shareddirs_send_remove_records (sdlist);
	statusbar_extra_message (_(" Unshared selected directories and sub-directories. (list update might take a while)"));
}


static void
shared_menuitem_refresh (GtkWidget *widget, gpointer data)
{
	shareddirs_send_get(0);
	send_get_shared_files();
}

static void
shared_menuitem_rehash (GtkWidget *widget, gpointer data)
{
	time_t timefrom = time(NULL);
	
	status_message_blue (N_("%s\n"), _(" Forcing core to rehash shared files. If you see a nonsense error message, ignore it."));
	
	while ((gtk_events_pending()) && time(NULL)-timefrom<2) 
		gtk_main_iteration_do(FALSE);
	
	/* old core versions: recheck all dirs by adding bogus directory                              */
	/* new core versions: add directory that is already shared to make it update the shared files */
	if (core_build_date >= CORE_JANUAR_2003)
	{	
		shareddirRecord *sdir;
		
		g_return_if_fail (shareddirs != NULL);
		g_return_if_fail (shareddirs->len > 0);
		
		sdir = g_ptr_array_index (shareddirs, 0);
		g_return_if_fail (sdir != NULL);
		g_return_if_fail (sdir->dirname != NULL);
		
		send_add_shared_dir (sdir->dirname);
		
		return;
	} 
	
	send_add_shared_dir (_("/ignore_this_error_message_it_is_to_force_the_core_to_rehash"));
}

static void 
shared_column_resized (GtkCList *clist, int col, int new_width, gpointer data)
{
	column_widths[SHARED_COLS][col] = new_width;
}


// right-click pop-up menu code by Mike Hicks, found on the gtk-list@redhat.com
//  mailing list on 25/11/98
//

static gint
shared_button_pressed (GtkWidget *widget, GdkEvent *event)
{
	GtkWidget   *menu, *item1, *item2, *item3, *item4, *item5, *item6, *item8, *item9, *item10;
	int	    row,col;

	last_time_of_user_click = time(NULL);
	if (event->type == GDK_BUTTON_PRESS)
	{
		// convert GdkEvent to GdkEventButton
		GdkEventButton *bevent = (GdkEventButton *) event;
		if (bevent->button == 3)
		{
			gtk_clist_get_selection_info (GTK_CLIST(clist_shared),
								(int) bevent->x ,(int) bevent->y, &row, &col);
			gtk_clist_select_row (GTK_CLIST(clist_shared),row,col);

			menu  = gtk_menu_new();
			item1 = gtk_menu_item_new_with_label (_(" add directory to share... "));
			item2 = gtk_menu_item_new_with_label (_(" add directory tree to share... "));
			item3 = gtk_menu_item_new ();
			item4 = gtk_menu_item_new_with_label (_(" unshare selected directories "));
			item5 = gtk_menu_item_new_with_label (_(" unshare selected directory trees "));
			item6 = gtk_menu_item_new ();
			item9 = gtk_menu_item_new_with_label (_(" refresh list "));
			item8 = gtk_menu_item_new ();
			item10= gtk_menu_item_new_with_label (_(" rehash shared files "));

			gtk_signal_connect (GTK_OBJECT(item1), N_("activate"), GTK_SIGNAL_FUNC(shared_menuitem_add_dir), NULL );
			gtk_signal_connect (GTK_OBJECT(item2), N_("activate"), GTK_SIGNAL_FUNC(shared_menuitem_add_tree),  NULL );
			gtk_signal_connect (GTK_OBJECT(item4), N_("activate"), GTK_SIGNAL_FUNC(shared_menuitem_unshare_selected_dirs), NULL);
			gtk_signal_connect (GTK_OBJECT(item5), N_("activate"), GTK_SIGNAL_FUNC(shared_menuitem_unshare_selected_dir_trees), NULL);
			gtk_signal_connect (GTK_OBJECT(item9), N_("activate"), GTK_SIGNAL_FUNC(shared_menuitem_refresh), NULL);
			gtk_signal_connect (GTK_OBJECT(item10),N_("activate"), GTK_SIGNAL_FUNC(shared_menuitem_rehash), NULL);

			// dir selection dialogue only makes sense if core runs on localhost
			if (CORE_IS_LOCAL)
			{
				gtk_menu_append (GTK_MENU(menu), item1); gtk_widget_show (GTK_WIDGET (item1));
				gtk_menu_append (GTK_MENU(menu), item2); gtk_widget_show (GTK_WIDGET (item2));
				gtk_menu_append (GTK_MENU(menu), item3); gtk_widget_show (GTK_WIDGET (item3));
			}
			gtk_menu_append (GTK_MENU(menu), item4); gtk_widget_show (GTK_WIDGET (item4));
			gtk_menu_append (GTK_MENU(menu), item5); gtk_widget_show (GTK_WIDGET (item5));
			gtk_menu_append (GTK_MENU(menu), item6); gtk_widget_show (GTK_WIDGET (item6));
			gtk_menu_append (GTK_MENU(menu), item9); gtk_widget_show (GTK_WIDGET (item9));
			gtk_menu_append (GTK_MENU(menu), item8); gtk_widget_show (GTK_WIDGET (item8));
			gtk_menu_append (GTK_MENU(menu),item10); gtk_widget_show (GTK_WIDGET (item10));

			// set most pop-up entries insensitive if there's nothing selected
			if ((GTK_CLIST(clist_shared)->selection == NULL) || (!CORE_IS_ALIVE))
			{
				gtk_widget_set_sensitive (GTK_WIDGET(item4), FALSE);
				gtk_widget_set_sensitive (GTK_WIDGET(item5), FALSE);
			}
			if (!CORE_IS_ALIVE)
			{
				gtk_widget_set_sensitive (GTK_WIDGET(item1), FALSE);
				gtk_widget_set_sensitive (GTK_WIDGET(item2), FALSE);
				gtk_widget_set_sensitive (GTK_WIDGET(item9), FALSE);
				gtk_widget_set_sensitive (GTK_WIDGET(item10), FALSE);
			}

//			DEBUG_2VAR (3, "BUTTON 3 (right button) is pressed (r=%d,c=%d)\n", row,col);
			gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
			bevent->button, bevent->time);
		}
		return TRUE;	 // event has been dealt with
	}

	return FALSE;
}



/* Populate shared Page  ---------------------------------------------------------------
*
*/

GtkWidget *
create_scrolled_dirs_shared_clist (void)
{
	GtkWidget	*scrollwin;
	gchar		*tabtitles = _(" path ");

	// creating shared clist window and set its properties

	clist_shared = gtk_clist_new_with_titles (1, &tabtitles);	// 1 columns
	gtk_widget_show(clist_shared);
	gtk_signal_connect (GTK_OBJECT (clist_shared), N_("click-column"),
			    GTK_SIGNAL_FUNC(shared_clist_column_clicked), NULL);
	gtk_signal_connect (GTK_OBJECT (clist_shared), N_("button_press_event"),
				GTK_SIGNAL_FUNC(shared_button_pressed), GTK_OBJECT(clist_shared));
	gtk_clist_set_auto_sort (GTK_CLIST(clist_shared),TRUE);
	gtk_clist_sort (GTK_CLIST(clist_shared));

	// always exactly one item selected (unless less is empty)
	gtk_clist_set_selection_mode(GTK_CLIST (clist_shared), GTK_SELECTION_MULTIPLE);
	// define look
	gtk_clist_set_shadow_type(GTK_CLIST (clist_shared), GTK_SHADOW_NONE);

        // set default column widths
	set_clist_columns (GTK_CLIST(clist_shared),SHARED_COLS,1);

	// create a scrollbar window to contain the clist later on

	scrollwin = gtk_scrolled_window_new (NULL,NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrollwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW(scrollwin), clist_shared);
	gtk_widget_show (clist_shared);
	gtk_widget_show (scrollwin);

	return (scrollwin);
}



GtkWidget *
create_sharedlist_buttons_hbox (void)
{
	GtkWidget *hbox_shared, *button_addshared, *button_add_structure;

	hbox_shared = gtk_hbox_new (FALSE,5);

	ew_addshared = gtk_entry_new_with_max_length (513);
	gtk_signal_connect (GTK_OBJECT (ew_addshared), N_("activate"),
			    GTK_SIGNAL_FUNC (shared_add_shared_button_clicked),
			    ew_addshared);
	/* make num block enter key behave like normal enter key */
	gtk_signal_connect (GTK_OBJECT (ew_addshared), N_("insert_text"),
				GTK_SIGNAL_FUNC (misc_gtk_entry_insert_text_handler), NULL);
	gtk_entry_set_text (GTK_ENTRY (ew_addshared), SHARED_PATH_ENTRY_WIDGET_DEFAULT_TEXT);
	gtk_box_pack_start (GTK_BOX (hbox_shared), ew_addshared, TRUE, TRUE,2);
	gtk_widget_show (ew_addshared);

	new_button_pack_and_signal_connect (&button_addshared, hbox_shared, _(" share directory "),
                                                shared_add_shared_button_clicked, ew_addshared, 1);

	new_button_pack_and_signal_connect (&button_add_structure, hbox_shared, _(" share directory structure "),
                                                shared_add_structure_button_clicked, ew_addshared, 0);

	gtk_widget_show (hbox_shared);
	return (hbox_shared);
}



/* shared_page_set_stats_label_message
 *
 *
 *
 *
 */

void
shared_page_set_stats_label_message (const gchar *txt)
{
	gchar *txt2, *flabel, *newline;

	g_return_if_fail (txt!=NULL);
	g_return_if_fail (shared_page_stats_label!=NULL);

	txt2 = g_strdup (txt);
	g_return_if_fail (txt2!=NULL);

	/* replace all newline characters with whitespaces */
	while ((newline = strchr(txt2,'\n')))
		*newline = ' ';

	/* shows if shared files are being hashed or not */
	flabel = g_strdup_printf (N_("%s- %s"), _(" Shared "), txt2);	
	gtk_label_set_text (GTK_LABEL(shared_page_stats_label), flabel);

	g_free (txt2);
	g_free (flabel);
}

/* shared_page_create_stats_label
 *
 *
 *
 *
 */

static GtkWidget *
shared_page_create_stats_label (void)
{
	GtkWidget	*hbox, *emptylabel;

	hbox = gtk_hbox_new(FALSE,0);
	gtk_widget_show(hbox);

	emptylabel = gtk_label_new(N_(" "));
	gtk_widget_show(emptylabel);

	shared_page_stats_label = gtk_label_new (_(" Shared "));
	gtk_label_set_justify (GTK_LABEL(shared_page_stats_label), GTK_JUSTIFY_LEFT);
	gtk_widget_show(shared_page_stats_label);

	gtk_box_pack_start (GTK_BOX(hbox), shared_page_stats_label, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(hbox), emptylabel, TRUE, TRUE, 0);

	if (option_nogreen == 0)
		set_style_recursively (hbox, style_notebook_label);

	return hbox;
}


/* shared_page_create
 *
 *
 *
 *
 */

GtkWidget *
shared_page_create (void)
{

#ifdef NOCOMPILE
	GtkWidget *table_shared;

	table_shared = gtk_table_new (1,2,FALSE);	/* 1x2, not homogeneous */

	/* create a paned window */

	spanedwin = gtk_hpaned_new ();
	gtk_widget_show (spanedwin);
	gtk_paned_add2 (GTK_PANED(spanedwin), shared_files_create_scrolled_clist());
	gtk_paned_add1 (GTK_PANED(spanedwin), create_scrolled_dirs_shared_clist());
	gtk_paned_set_position (GTK_PANED(spanedwin), paned_win_position[SHARED_PANE_POS]);

	/* create scrolled shared clist and add it to shared table */

	gtk_table_attach (GTK_TABLE (table_shared),
			  spanedwin,
			  0,1,0,1, GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			  	   GTK_FILL | GTK_SHRINK | GTK_EXPAND, 5,5);

	/* create horizontal packing box for shared control buttons and add it to table */
        /* only necessary if the core runs on remote host, otherwise we can use the     */
        /*   right-click pop-up menu functions with dir selection dialogue              */

        gtk_table_attach (GTK_TABLE(table_shared),
	        	  (shared_hbox = create_sharedlist_buttons_hbox()),
		          0,1,1,2, GTK_SHRINK | GTK_FILL | GTK_EXPAND,
		                   GTK_SHRINK | GTK_FILL, 50,5);

	gtk_widget_show (table_shared);
	return table_shared;
#endif

	GtkWidget *vbox, *statslabel;

	vbox = gtk_vbox_new(FALSE,0);
	gtk_widget_show(vbox);

	statslabel = shared_page_create_stats_label();
	gtk_box_pack_start (GTK_BOX(vbox), statslabel, FALSE, FALSE, 0);

	/* create a paned window */
	spanedwin = gtk_hpaned_new ();
	gtk_widget_show (spanedwin);
	gtk_paned_add2 (GTK_PANED(spanedwin), shared_files_create_scrolled_clist());
	gtk_paned_add1 (GTK_PANED(spanedwin), create_scrolled_dirs_shared_clist());
	gtk_paned_set_position (GTK_PANED(spanedwin), paned_win_position[SHARED_PANE_POS]);
	gtk_box_pack_start (GTK_BOX(vbox), spanedwin, TRUE, TRUE, 0);

	shared_hbox = create_sharedlist_buttons_hbox();
	gtk_box_pack_start (GTK_BOX(vbox), shared_hbox, FALSE, FALSE, 0);

	return vbox;
}


