/***************************************************************************
                          stats.c  -  description
                             -------------------
    begin                : Wed Jul 31 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "colors.h"
#include "global.h"
#include "gui_options.h"
#include "icons.h"
#include "misc.h"
#include "notebook.h"
#include "stats.h"
#include "status_page.h"
#include "uploads_page.h"
#include "downloads_page.h"

gint	option_timeline_active = STATS_TIMELINE_LAST_3MINS;
gint	option_timeline_mode = STATS_TIMELINE_DISPLAYMODE_BARS;

gfloat	option_stats_up_wrong = 1024.0;
gfloat	option_stats_down_wrong = 2048.0;
gfloat	option_stats_up_crop = 150.0;
gfloat	option_stats_down_crop = 150.0;

static GtkWidget *ew_up_wrong, *ew_down_wrong, *ew_crop_up, *ew_crop_down;


static time_t		 time_first_of_packets = 0;
static time_t		 time_first_of_minutes = 0;
static time_t		 time_first_of_hours = 0;
static time_t		 time_first_of_days = 0;

static gfloat		 up_max = 0.0;
static gfloat		 down_max = 0.0;

static GPtrArray	*stats_packets = NULL;
static GPtrArray	*stats_minutes = NULL;
static GPtrArray	*stats_hours = NULL;
static GPtrArray	*stats_days = NULL;

static GtkWidget	*chart = NULL;	// will be our GtkDrawingArea
static chartPixmap	*cpm = NULL;	// the pixmap we draw on (we use double-buffering to avoid flicker)

static GdkGC		*penblack = NULL;
static GdkGC		*penred = NULL;
static GdkGC		*penblue = NULL;
static GdkGC		*pengreen = NULL;
static GdkGC		*penwhite = NULL;

static statsRecord	*stats_make_average_stat_record (GPtrArray *statsarray, time_t from, time_t to);
static void			 stats_remove_old_records (GPtrArray *statsarray, gint maxsecondsago);
static void			 stats_repaint ();


/* stats_register_upload_download_stats_packet
 *
 * regtime: time at which these up/down values are registered. Usually time(NULL),
 *          but when we load the stats and haven't run the GUI for some time we
 *          need to fill the stats buffers with 0/0 values for the times that the
 *          GUI has not run (due to the current plotting routines; when these are
 *          changed to calculate the X position from the time value, then we won't
 *          need this anymore).
 *
 */

void
stats_register_upload_download_stats_packet ( time_t regtime,
                                              gfloat upspeed,
                                              gfloat downspeed,
                                              guint  no_up,
                                              guint  no_down )
{
	static gfloat	 lastup=0, lastdown=0;
	statsRecord		*statrec;

	if (!stats_packets)
		stats_packets = g_ptr_array_new();
	g_return_if_fail (stats_packets!=NULL);

	if (time_first_of_packets == 0)
		time_first_of_packets = regtime;

	statrec = g_new0(statsRecord, 1);
	g_return_if_fail (statrec!=NULL);

	up_max = MAX (up_max, upspeed);
	down_max = MAX (down_max, downspeed);

	if (upspeed >= option_stats_up_wrong)
		upspeed = lastup;
	if (downspeed >= option_stats_down_wrong)
		downspeed = lastdown;

	lastup=upspeed;
	lastdown=downspeed;

	statrec->up = upspeed;
	statrec->down = downspeed;
	statrec->no_uploads = no_up;
	statrec->no_downloads = no_down;
	statrec->when = regtime;

//	g_print ("registered packet #%u - %lu\n", stats_packets->len+1, time(NULL));

	g_ptr_array_add(stats_packets, (gpointer)statrec);

	if ((stats_packets) && regtime-time_first_of_packets >= 60 && time_first_of_packets>0 )
	{
		statsRecord	*statrec_min;

		if (!stats_minutes)
			stats_minutes = g_ptr_array_new();
		g_return_if_fail (stats_minutes!=NULL);

		statrec_min = stats_make_average_stat_record (stats_packets, time_first_of_packets, regtime);
		g_return_if_fail (statrec_min!=NULL);

		if (time_first_of_minutes == 0)
			time_first_of_minutes = regtime;

		g_ptr_array_add(stats_minutes, (gpointer)statrec_min);
		time_first_of_packets = 0;
		/* remove all packets stats */
		stats_remove_old_records (stats_packets, 3*(60+1));		// keep one packet more than necessary
	}

	if ((stats_minutes) && regtime-time_first_of_minutes >= 60*60 && time_first_of_minutes>0 )
	{
		statsRecord	*statrec_hour;

		if (!stats_hours)
			stats_hours = g_ptr_array_new();
		g_return_if_fail (stats_hours!=NULL);

		statrec_hour = stats_make_average_stat_record (stats_minutes, time_first_of_minutes, regtime);
		g_return_if_fail (statrec_hour!=NULL);

		if (time_first_of_hours == 0)
			time_first_of_hours = regtime;
	
		g_ptr_array_add(stats_hours, (gpointer)statrec_hour);
//		g_print ("stats_hours: added AVG at %lu (first_of_minutes == %lu)\n", statrec_hour->when, time_first_of_minutes);
		time_first_of_minutes = 0;
		// remove all older than 12 hours
		stats_remove_old_records (stats_minutes, 60*60*(12+1));	
	}

	if ((stats_hours) && regtime-time_first_of_hours >= 60*60*24 && time_first_of_hours>0 )
	{
		statsRecord	*statrec_days;

		if (!stats_days)
			stats_days = g_ptr_array_new();
		g_return_if_fail (stats_days!=NULL);

		statrec_days = stats_make_average_stat_record (stats_hours, time_first_of_hours, regtime);
		g_return_if_fail (statrec_days!=NULL);

		if (time_first_of_days == 0)
			time_first_of_days = regtime;

		g_ptr_array_add(stats_days, (gpointer)statrec_days);
		time_first_of_hours = 0;
		// remove all older than 7 days
		stats_remove_old_records (stats_hours, 60*60*24*(7+1));	
	}

	if (current_page == NOTEBOOK_PAGE_STATS)
		stats_repaint();
}


static statsRecord *
stats_make_average_stat_record (GPtrArray *statsarray, time_t from, time_t to)
{
	statsRecord *newsr;
	guint		 num = 0;

	g_return_val_if_fail (statsarray!=NULL, NULL);
	g_return_val_if_fail (from < to, NULL);

	newsr = g_new0 (statsRecord,1);
	g_return_val_if_fail (newsr!=NULL,NULL);

	newsr->up = 0.0;
	newsr->down = 0.0;
	newsr->no_uploads = 0;
	newsr->no_downloads = 0;
	
	PROCESS_GPTRARRAY_FROM_END ( statsarray, statsRecord,
		if (item) 
		{
			if (item->when < from)	// it's sorted, and we're processing from end, so we can break here
				break;
			newsr->up += item->up;
			newsr->down += item->down;
			newsr->no_uploads += item->no_uploads;
			newsr->no_downloads += item->no_downloads;
			num++;
		}
	);

	if(num>0)
	{
		newsr->up /= (num*1.0);
		newsr->down /= (num*1.0);
		newsr->no_uploads /= num;
		newsr->no_downloads /= num;
	}

	newsr->when = from + ((to-from)/2);

	return newsr;
}


// stats_remove_old_records
//
// removes all records from a stats array older than 'maxsecondsago'

static void
stats_remove_old_records (GPtrArray *statsarray, gint maxsecondsago)
{
	time_t	too_old = time(NULL)-maxsecondsago;

	g_return_if_fail (statsarray!=NULL);

	PROCESS_GPTRARRAY_FROM_END ( statsarray, statsRecord,
		if ((item) && item->when <= too_old)
		{
			// remove item from array
			g_ptr_array_remove (statsarray, (gpointer)item);
			// free item
			memset(item,0x00,sizeof(item));
			g_free(item);
		}
	);
}


// stats_find_highest_samples
//
// Finds the highest upload/download sample
// Returns FALSE on error, otherwise TRUE

static gboolean
stats_find_highest_samples (GPtrArray *statsarray, guint maxsamples, guint *maxupsample, guint *maxdownsample)
{
//	time_t	too_old = time(NULL)-maxsecondsago;
	guint	c = 0;

	*maxupsample = 0;
	*maxdownsample = 0;

	g_return_val_if_fail (statsarray!=NULL, FALSE);

	PROCESS_GPTRARRAY_FROM_END ( statsarray, statsRecord,
//		if ((item) && item->when < too_old)		// array is sorted, and we're checking from end
//			break;
		if (c>maxsamples)
			break;
		if (item)
		{
			*maxupsample = MAX(*maxupsample, item->up);
			*maxdownsample = MAX(*maxdownsample, item->down);
		} //else g_print ("no item!\n");
		c++;
//		g_print ("c = %u\n", c);
	);

	*maxupsample = CLAMP(*maxupsample, *maxupsample, option_stats_up_crop);
	*maxdownsample = CLAMP(*maxdownsample, *maxdownsample, option_stats_down_crop);

	return TRUE;
}


// stats_get_sample
//
// gets the x_last sample (ie. last sample - x samples)
// Returns statsRecord or NULL if sample not found

static statsRecord *
stats_get_sample (GPtrArray *statsarr, guint x_last)
{
	guint count=0;
	g_return_val_if_fail (statsarr!=NULL, NULL);

	PROCESS_GPTRARRAY_FROM_END ( statsarr, statsRecord,
		// greater than because the code could skip the wanted item if item==NULL at count==x_last
		if ((item) && count>=x_last)
			return item;
		count++;
	);
	return NULL;
}


// stats_repaint_paint_points
//
// displays the points and the graphs

static void
stats_repaint_paint_points (GdkPoint *points_up, GdkPoint *points_down, guint num, gfloat hstep, guint midy)
{
	guint i;

	g_return_if_fail (points_up!=NULL);	
	g_return_if_fail (points_down!=NULL);

//	g_print ("paint_points - num=%u, hstep=%f\n", num, hstep);

	// First, draw the individual points (if the view mode requires that)
	for (i=0; i<num; i++)
	{	
		switch (option_timeline_mode)
		{
			case STATS_TIMELINE_DISPLAYMODE_LINES:
			{
				// draw little filled circles to mark a dot in this mode
				gdk_draw_arc (cpm->pixmap, penblue, TRUE, points_up[i].x-2, points_up[i].y-2, 4, 4, 0, 64*360);
				gdk_draw_arc (cpm->pixmap, pengreen, TRUE,points_down[i].x-2-1, points_down[i].y-2-1, 4, 4, 0, 64*360);
			}
			break;
			case STATS_TIMELINE_DISPLAYMODE_BARS:	
			{
//				gint spacing = (hstep>=5.0) ? 2 : -1;	// space between the bars
				gint spacing = -1;
				gint xminus = (guint) (hstep/2.0)-spacing;
				gint up_h = midy-points_up[i].y;				// height of the upload bar
				gint down_h = points_down[i].y-midy;	// height of the download bar

				if (up_h>0)
				{
//					g_print ("\t RECT UP - %d %d %d %d\n", points_up[i].x-xminus, points_up[i].y,
//										((guint)hstep)-spacing, up_h);
					gdk_draw_rectangle (cpm->pixmap, penblue, TRUE,
										points_up[i].x-xminus, points_up[i].y,
										((guint)hstep)-spacing, up_h);
				}
				if (down_h>0)
				{
//					g_print ("\t RECT DOWN - %d %d %d %d\n", points_down[i].x-xminus, midy,
//										((guint)hstep)-spacing, down_h);
					gdk_draw_rectangle (cpm->pixmap, pengreen, TRUE,
										points_down[i].x-xminus, midy,
										((guint)hstep)-spacing, down_h);
				}
				// draw bars
			}
			default:
			break;
		}
	}

	// Secondly, do anything else that's required:
	switch (option_timeline_mode)
	{
		case STATS_TIMELINE_DISPLAYMODE_LINES:
		{
			// connect points with lines
			gdk_draw_lines (cpm->pixmap, penblue,  points_up, num);
			gdk_draw_lines (cpm->pixmap, pengreen, points_down, num);
		}
		break;
		case STATS_TIMELINE_DISPLAYMODE_BARS:
		default:
		break;
	}
}


/* stats_repaint_get_timeline_parameters
 *
 * sets the parameters that result from different choices of the timeline
 *
 */

static void
stats_repaint_get_timeline_parameters (guint *hunits, guint *tunits, guint *vunits, GPtrArray **statsarr, gchar **unitstr)
{
	/* hunit_table:		number of samples we need to graph
	 * tunit_table:		amount of seconds each single sample is 'worth'
	 * vunit_table:     number of samples between to vertical lines
	 * statsarr_table:	which array to take the samples from
	 */
	const guint	 hunit_table[STATS_TIMELINE_LAST] = {	  60, 	    60,  6*60,	  12,	 24,    48,    72,  7*24,       31,       365 };
	const guint	 tunit_table[STATS_TIMELINE_LAST] = {  	   3,       60,	   60, 60*60, 60*60, 60*60, 60*60, 60*60, 60*60*24,  60*60*24 };
	/* total duration of the graph                       3mn        1h     6h    12h    24h    48h    72h  1week    31days    365days */
	const guint	 vunit_table[STATS_TIMELINE_LAST] = { 	  20, 	    10,    60,	   6,	  6,    24,    24,    24,        1,        30 };
	const gchar* vustr_table[STATS_TIMELINE_LAST]=  {"1 min", "10 min", "1 h", "6 h", "6 h", "1 d", "1 d", "1 d",    "1 d",  "1 month"};
	GPtrArray	*statsarr_table[STATS_TIMELINE_LAST] = { stats_packets, stats_minutes, stats_minutes,
														 stats_hours, stats_hours, stats_hours,
														 stats_hours, stats_hours,	stats_days, stats_days};

	static gchar	unitstring[128];

	g_return_if_fail (hunits!=NULL);
	g_return_if_fail (tunits!=NULL);
	g_return_if_fail (vunits!=NULL);
	g_return_if_fail (statsarr!=NULL);

	*statsarr = NULL;
	if (option_timeline_active>=0 && option_timeline_active<STATS_TIMELINE_LAST)
	{
		*hunits = hunit_table[option_timeline_active];
		*tunits = tunit_table[option_timeline_active];
		*vunits = vunit_table[option_timeline_active];
		*statsarr = statsarr_table[option_timeline_active];
	} else g_printerr (_("BUG: unknown time line in %s?!\n"), __FUNCTION__);

	switch (*tunits)
	{
		case 3:			g_snprintf (unitstring, sizeof(unitstring), _("%u minutes/%s"), (*hunits * *tunits)/60,vustr_table[option_timeline_active]);
			break;
		case 60:		g_snprintf (unitstring, sizeof(unitstring), _("%u hours/%s"), *hunits/60,vustr_table[option_timeline_active]);
			break;
		case 60*60:		if (*hunits<=24)
							g_snprintf (unitstring, sizeof(unitstring), _("%u hours/%s"), *hunits,vustr_table[option_timeline_active]);
						else g_snprintf (unitstring, sizeof(unitstring), _("%u days/%s"), *hunits/24,vustr_table[option_timeline_active]);
			break;
		case 60*60*24:	g_snprintf (unitstring, sizeof(unitstring), _("%u days/%s"), *hunits,vustr_table[option_timeline_active]);
			break;
		default:		g_snprintf (unitstring, sizeof(unitstring), N_("???"));
			break;
	}
	*unitstr = unitstring;
}



// stats_repaint_draw_kBs_lines
//
// draws horizontal lines every X kB/s and puts a text label on these lines
// maxsample:	highest kB/s value we got in our sample (both up and down)
// ystep:		the number of vertical pixels for every 1.0kB/s

static void
stats_repaint_draw_kBs_lines (guint maxsample, guint maxx, guint maxy, guint midy, gfloat ystep, gboolean textonly)
{
#if ((GTK_MAJOR_VERSION == 2) || (GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 3))
	g_print ("***** GTK2 FIXME: %s\n", __FUNCTION__);
#else
	gint	ypos, ypos2;
	guint	kBs_per_line = 5;	// how many kB/s the distance between two lines represents
	guint	count = 0;			// counter - we're drawing the n-th line

	g_return_if_fail (chart!=NULL);
	g_return_if_fail (cpm!=NULL);
	g_return_if_fail (ystep>0.0);

	if (midy<=0)	// don't do anything on pixmap creation (would loop forever)
		return;

	if (maxsample<5) kBs_per_line = 1;
	if (maxsample<10) kBs_per_line = 2;
	if (maxsample>50)  kBs_per_line = 10;
	if (maxsample>100) kBs_per_line = 20;
	if (maxsample>250) kBs_per_line = 50;
	if (maxsample>500) kBs_per_line = 100;

	// label lines above middle
	ypos = ypos2 = midy;
	count = 0;
	while (ypos>0 || ypos2<maxy)
	{
		GtkStyle	*defstyle = gtk_widget_get_default_style ();
		gchar 		 axistxt[64];

		g_snprintf (axistxt, sizeof(axistxt), N_("%u kB/s"), count*kBs_per_line);

		if (!textonly)
		{
			// draw horizontal line every 5kB/s
			if (ypos>0)
				gdk_draw_line (cpm->pixmap, penblack,  0,  ypos, maxx, ypos);
			if (ypos2<maxy)
				gdk_draw_line (cpm->pixmap, penblack, 0,  ypos2, maxx, ypos2);
//				gdk_draw_line (cpm->pixmap, penblack, 0,  midy+(midy-ypos), maxx, midy+(midy-ypos));
		}

		if (ypos>0 && ypos<midy && (defstyle))	// don't label 0kB/s line
			gdk_draw_string (cpm->pixmap, defstyle->font, penblack, 5, ypos-2, axistxt);
		if (ypos2>midy && ypos2<maxy && (defstyle))	// don't label 0kB/s line
			gdk_draw_string (cpm->pixmap, defstyle->font, penblack, 5, ypos2-2, axistxt);
//			gdk_draw_string (cpm->pixmap, defstyle->font, penblack, 5, midy+(midy-ypos)-2, axistxt);

		ypos  -= ystep*kBs_per_line;
		ypos2 += ystep*kBs_per_line;
		count++;
	}

	// label lines below middle
	ypos = midy;
	count = 0;
	while (ypos < maxy)
	{
		GtkStyle	*defstyle = gtk_widget_get_default_style ();
		gchar 		 axistxt[64];

		g_snprintf (axistxt, sizeof(axistxt), _("%u kB/s"), count*kBs_per_line);

		if (!textonly)
		{
			// draw horizontal line every 5kB/s
			if (ypos>0)
			gdk_draw_line (cpm->pixmap, penblack,  0,  ypos, maxx, ypos);
//			gdk_draw_line (cpm->pixmap, penblack, 0,  midy+(midy-ypos), maxx, midy+(midy-ypos));
		}

		if ((ypos<midy) && (defstyle))	// don't label 0kB/s line
		{
			gdk_draw_string (cpm->pixmap, defstyle->font, penblack, 5, ypos-2, axistxt);
			gdk_draw_string (cpm->pixmap, defstyle->font, penblack, 5, midy+(midy-ypos)-2, axistxt);
		}

		ypos += ystep*kBs_per_line;
		count++;
	}
#endif
}


/* stats_get_all_samples_in_interval
 *
 * gets all samples in the intervall [t_from;t_to] from the given samples array
 * returns a GSList with the samples that needs to be freed by the caller
 *
 */

static GSList *
stats_get_all_samples_in_interval (GPtrArray *statsarr, time_t t_from, time_t t_to)
{
	GSList	*samples = NULL;

	g_return_val_if_fail (t_to > t_from, NULL);

	if (statsarr)
	{
		PROCESS_GPTRARRAY ( statsarr, statsRecord,
			if ((item) && (item->when >= t_from) && (item->when <= t_to))
				samples = g_slist_append (samples, (gpointer)item);
		);
	}
	return samples;
}


/* stats_repaint
 *
 * repaints our graphs on the pixmap
 *
 */

static void
stats_repaint ()
{
	GPtrArray		*statsarr = NULL;
	guint			 hunits=0;	/* number of samples (=horizontal 'units')			*/
	guint			 tunits=0;	/* number of seconds each sample represents			*/
	guint			 vunits=0;	/* number of samples between vertical lines			*/
	guint			 vline =0;	/* counter for vertical lines						*/
	guint			 maxx, maxy, xstep, maxupsample, maxdownsample;
	gfloat			 step;
	gchar			*unitstr = NULL;	/* the interval that is graphed as a string	*/
	GtkStyle		*defstyle = gtk_widget_get_default_style ();

	g_return_if_fail (cpm!=NULL);
	g_return_if_fail (chart!=NULL);

	maxx = chart->allocation.width;
	maxy = chart->allocation.height;

	/* clear drawing area and draw a horizontal line in the middle */
	gdk_draw_rectangle (cpm->pixmap, penwhite, TRUE, 0, 0, maxx, maxy);
	gdk_draw_line (cpm->pixmap, penblack, 0, maxy/2, maxx, maxy/2);

	stats_repaint_get_timeline_parameters (&hunits, &tunits, &vunits, &statsarr, &unitstr );

	if (statsarr)		// maybe we don't have any samples yet...
	{
//		g_print ("Samples to graph: %u (in arr:%u) tunits=%u\n", hunits, statsarr->len, tunits);
		if (stats_find_highest_samples (statsarr, hunits/*tunits*/, &maxupsample, &maxdownsample))
		{
			guint		 maxsample = MAX(maxupsample,maxdownsample)+1;	// 1 for good measure
			gfloat		 ystep = (maxy*1.0)/(maxsample*2.0);
			GdkPoint	*points_up = g_new0 (GdkPoint, hunits+1);
			GdkPoint	*points_down = g_new0 (GdkPoint, hunits+1);
/*			GdkPoint	*points_up = NULL;
			GdkPoint	*points_down = NULL;
*/			guint		 midy, maxupsample2, maxdownsample2, maxvertical;

//			g_print ("max up sample   = %u\n", maxupsample);
//			g_print ("max down sample = %u\n", maxdownsample);
			maxupsample2 = MAX(maxdownsample/10, maxupsample);
			maxdownsample2 = MAX(maxupsample/10, maxdownsample);
			maxvertical = maxupsample2+maxdownsample2+2;	// 2 for good measure

			if (maxupsample == maxdownsample)
				midy = maxy/2;
			else midy = (guint)( (maxy*1.0) * ((maxupsample2*1.0)/(maxvertical*1.0)) );

//			if (maxy-midy<10) midy=maxy-10;
//				if (midy<10) midy=10;
			g_return_if_fail (points_up!=NULL);
			g_return_if_fail (points_down!=NULL);

			// remove old line in exact middle
			gdk_draw_line (cpm->pixmap, penwhite, 0, maxy/2, maxx, maxy/2);

			stats_repaint_draw_kBs_lines (maxsample, maxx, maxy, midy, ystep, FALSE);
//			gdk_draw_line (cpm->pixmap, penblack, 0, midy, maxx, midy);

/* OLD ROUTINE:
   - gets last X samples and takes their values
   - *assumes* that samples are in right order, right intervals + complete
*/
			step = (maxx*1.0)/(hunits*1.0);
			for (xstep=0; xstep<=hunits; xstep++)
			{
				guint		 xpos = (guint) xstep*step;
				guint		 ypos_up = midy;
				guint		 ypos_down = midy;
				statsRecord	*statsrec = stats_get_sample (statsarr, xstep);

				if (statsrec)
				{
					gfloat upval = CLAMP(statsrec->up, 0.0, option_stats_up_crop);
					gfloat downval = CLAMP(statsrec->down, 0.0, option_stats_down_crop);

					ypos_up = midy-(gint)(upval*ystep);
					ypos_down = midy+(gint)(downval*ystep);
				} 

				points_up[xstep].x = maxx-xpos;		// make the graph shift leftwards
				points_up[xstep].y = ypos_up;
				points_down[xstep].x = maxx-xpos;
				points_down[xstep].y = ypos_down;

				if (++vline==vunits)
				{
					gdk_draw_line(cpm->pixmap,penblack,maxx-xpos,0,maxx-xpos,maxy);
					vline=0;
				}
				/* g_print ("vline : %u maxx=%u xpos=%u\n", vline, maxx, xpos); */
			}

#ifdef SKIPTHIS
/* NEW ROUTINE (not finished + not working yet)

   - gets all samples between time t1 and t2
   - plots them on the x-axis depending on t
*/
			if (1==1)
			{
				guint	 interval = hunits * tunits;	/* interval displayed in secs */
				guint	 interval_float = 1.0*interval;
				time_t	 t_from = time(NULL)-interval;
				time_t	 t_to = time(NULL);
				GSList	*samples = stats_get_all_samples_in_interval (statsarr, t_from, t_to);
				if (g_slist_length(samples)>0)
				{
					points_up = g_new0(GdkPoint,g_slist_length(samples)+1);
					points_down = g_new0(GdkPoint,g_slist_length(samples)+1);
					xstep = 0;
					PROCESS_GSLIST (samples, statsRecord,
						if (xstep>=g_slist_length(samples))
						{
							g_printerr (N_("debug: in %s - xstep>= g_slist_length(samples)\n"), __FUNCTION__);
							continue;
						}
						if (item)
						{
							gfloat	upval = CLAMP(item->up, 0.0, option_stats_up_crop);
							gfloat	downval = CLAMP(item->down, 0.0, option_stats_down_crop);
							gfloat	xpos_float = (((item->when-t_from)*1.0)/interval_float)*(maxx*1.0);
							guint	xpos_int = (guint) xpos_float;
							guint	ypos_up = midy-(gint)(upval*ystep);
							guint	ypos_down = midy+(gint)(downval*ystep);

							/* make the graph shift leftwards */
							points_up[xstep].x = maxx-xpos_int;
							points_up[xstep].y = ypos_up;
							points_down[xstep].x = maxx-xpos_int;
							points_down[xstep].y = ypos_down;
							xstep++;
						}
					);
					g_slist_free(samples);
				}
			}
#endif /* ifdef SKIPTHIS */

			stats_repaint_paint_points (points_up, points_down, hunits+1, step, midy);

			// Now draw the labels on top of the graph again, but not the lines
			stats_repaint_draw_kBs_lines (maxsample, maxx, maxy, midy, ystep, TRUE);

			// draw a new thick line in 'dynamic' vertical middle with an 'arrow'
			gdk_draw_rectangle (cpm->pixmap, penblack, TRUE, 0, midy-1, maxx, 3);
			gdk_draw_arc (cpm->pixmap, penblack, TRUE, maxx-20, midy-10, 40, 20, (180-30)*64, 60*64);

			g_free(points_up);
			g_free(points_down);
		}
	} // else: no samples yet in array

#if ((GTK_MAJOR_VERSION == 2) || (GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 3))
	g_print ("***** GTK2 FIXME: %s\n", __FUNCTION__);
#else
	// display interval
	if ((unitstr) && (defstyle))
		gdk_draw_string (cpm->pixmap, defstyle->font, penblack, maxx-150, 25, unitstr);
#endif

	// now update the whole drawing area
	if (1==1)
	{
		GdkRectangle update_area;

		update_area.x = 0;
		update_area.y = 0;
		update_area.width = maxx;
		update_area.height = maxy;

		gtk_widget_draw (chart, &update_area);
	}
}


// stats_create_pen
//
// creates a graphics context for the given colour

static GdkGC *
stats_create_pen (GdkColor *c)
{
	GdkGC *gc;
	g_return_val_if_fail (cpm!=NULL, NULL);
	gc = gdk_gc_new(cpm->pixmap);
	gdk_gc_set_foreground(gc, c);
	return gc;	
}


// stats_configure_event
//
// called when the drawing are is created and/or it's resized

static gint
stats_configure_event (GtkWidget *widget, GdkEventConfigure *event, gpointer data)
{
	if (!cpm)
	{
		cpm = g_new(chartPixmap, 1);
		cpm->gc = NULL;
		cpm->pixmap = NULL;
	}

	// free old pixmap
	if (cpm->pixmap)
		gdk_pixmap_unref(cpm->pixmap);

	// create a new pixmap
	cpm->pixmap = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);

	// create pens if we don't have them yet
	if (!penblack)
	{
		penblack = stats_create_pen (&black);
		penred = stats_create_pen (&red);
		penblue = stats_create_pen (&blue);
		pengreen = stats_create_pen (&dark_green);
		penwhite = stats_create_pen (&white);
	}

	// clear new pixmap
	gdk_draw_rectangle (cpm->pixmap, widget->style->white_gc, TRUE, 0,0, widget->allocation.width, widget->allocation.height);
	stats_repaint ();

	return TRUE;
}


// stats_expose_event
//
// called when the drawing are needs to be redrawn in whole or in parts

static gint
stats_expose_event (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	if ((cpm) && (cpm->pixmap))
	{
		// copy relevant part of pixmap over into the drawing area
		gdk_draw_pixmap (widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)], cpm->pixmap,
							event->area.x, event->area.y, 
							event->area.x, event->area.y, event->area.width, event->area.height);
	}
	return TRUE;
}


// stats_ignore_dialog_ok
//
//

static void
stats_ignore_dialog_ok (GtkWidget *widget, gpointer ignore_dialog)
{
	gchar	*upwstr = gtk_editable_get_chars (GTK_EDITABLE(ew_up_wrong), 0, -1);
	gchar	*upcstr = gtk_editable_get_chars (GTK_EDITABLE(ew_crop_up), 0, -1);
	gchar	*downwstr = gtk_editable_get_chars (GTK_EDITABLE(ew_down_wrong), 0, -1);
	gchar	*downcstr = gtk_editable_get_chars (GTK_EDITABLE(ew_crop_down), 0, -1);

	option_stats_up_wrong = option_stats_down_wrong = 1024.0;	// if sth goes wrong,
	option_stats_up_crop = option_stats_down_crop = 150.0;		// use default values

	if ((upwstr)&&(upcstr)&&(downwstr)&&(downcstr))
	{
		option_stats_up_wrong = (gfloat) atof(upwstr);
		option_stats_up_crop  = (gfloat) atof(upcstr);
		option_stats_down_wrong = (gfloat) atof(downwstr);
		option_stats_down_crop  = (gfloat) atof(downcstr);
	} else g_printerr (_("Couldn't retrieve text from stats ignore dialog!\n"));

	if (ignore_dialog)
		gtk_widget_destroy(GTK_WIDGET(ignore_dialog));

	if (upwstr) g_free(upwstr);
	if (upcstr) g_free(upcstr);
	if (downwstr) g_free(downwstr);
	if (downcstr) g_free(downcstr);

	stats_repaint ();
}


// stats_ignore_dialog_destroy 
//
//

static void
stats_ignore_dialog_destroy  (GtkWidget *widget, gpointer ignore_dialog)
{
	if (ignore_dialog)
		gtk_widget_destroy(GTK_WIDGET(ignore_dialog));
}



// stats_popupmenu_item_selected
//
// a button has been pressed or an item from the pop-up menu been selected

static void
stats_popupmenu_item_selected (GtkWidget *widget, gpointer data)
{
	const guint tlarr[STATS_MENUITEM_YEAR+1] =
						{
						 STATS_TIMELINE_LAST_3MINS,	STATS_TIMELINE_LAST_HOUR,
						 STATS_TIMELINE_LAST_6HOURS, STATS_TIMELINE_LAST_12HOURS,
						 STATS_TIMELINE_LAST_24HOURS, STATS_TIMELINE_LAST_48HOURS,
						 STATS_TIMELINE_LAST_72HOURS, STATS_TIMELINE_LAST_WEEK,
						 STATS_TIMELINE_LAST_MONTH, STATS_TIMELINE_LAST_YEAR
						};
	guint menuitem = GPOINTER_TO_UINT(data);

	switch (menuitem)
	{
		case STATS_MENUITEM_3MIN:		case STATS_MENUITEM_HOUR:
		case STATS_MENUITEM_6HOURS:		case STATS_MENUITEM_12HOURS:
		case STATS_MENUITEM_24HOURS:	case STATS_MENUITEM_2DAYS:
		case STATS_MENUITEM_3DAYS:		case STATS_MENUITEM_WEEK:
		case STATS_MENUITEM_MONTH:		case STATS_MENUITEM_YEAR:
			option_timeline_active = tlarr[menuitem];
		break;

		case STATS_MENUITEM_NEXT:
			option_timeline_active = (option_timeline_active+1) % STATS_TIMELINE_LAST;
		break;

		case STATS_MENUITEM_IGNORE_DIALOG:
		{
			GtkWidget	*ignore_dialog, *button_ok, *button_cancel;
			gchar		 defstr[32];
			gchar 		*UP_WRONG	=   _("  Upload speed: values higher than this are definitively wrong (kB/s) ");
			gchar 		*UP_CROP	=	_("  Upload speed: do not DISPLAY values higher than this in graph (kB/s) ");
			gchar 		*DOWN_WRONG	=	_("  Download speed: values higher than this are definitively wrong (kB/s) ");
			gchar 		*DOWN_CROP	=	_("  Download speed: do not DISPLAY values higher than this in graph (kB/s) ");

			ignore_dialog = gtk_dialog_new ();
			gtk_signal_connect (GTK_OBJECT (ignore_dialog), N_("destroy"),
					GTK_SIGNAL_FUNC (stats_ignore_dialog_destroy), (gpointer) ignore_dialog);
			gtk_window_set_title (GTK_WINDOW (ignore_dialog), _(" configure upload/download speed values to ignore "));
			gtk_widget_set_uposition (GTK_WIDGET (ignore_dialog), 320, 240);
			new_icon_button_pack_and_signal_connect (&button_ok, GTK_DIALOG(ignore_dialog)->action_area, _("   OK    "),
					    stats_ignore_dialog_ok, (gpointer) ignore_dialog, 1, icon_ok);
			new_icon_button_pack_and_signal_connect (&button_cancel, GTK_DIALOG(ignore_dialog)->action_area, _(" Cancel "),
					    stats_ignore_dialog_destroy, (gpointer) ignore_dialog, 1, icon_cancel);
			add_new_entry_widget_to_box (GTK_DIALOG(ignore_dialog)->vbox, &ew_up_wrong, UP_WRONG, 10);
			add_new_entry_widget_to_box (GTK_DIALOG(ignore_dialog)->vbox, &ew_crop_up,  UP_CROP, 10);
			add_new_entry_widget_to_box (GTK_DIALOG(ignore_dialog)->vbox, &ew_down_wrong, DOWN_WRONG, 10);
			add_new_entry_widget_to_box (GTK_DIALOG(ignore_dialog)->vbox, &ew_crop_down, DOWN_CROP, 10);
			g_snprintf(defstr, sizeof(defstr), N_("%.0f"), option_stats_up_wrong);
			gtk_entry_set_text (GTK_ENTRY(ew_up_wrong), defstr);
			g_snprintf(defstr, sizeof(defstr), N_("%.0f"), option_stats_down_wrong);
			gtk_entry_set_text (GTK_ENTRY(ew_down_wrong), defstr);
			g_snprintf(defstr, sizeof(defstr), N_("%.0f"), option_stats_up_crop);
			gtk_entry_set_text (GTK_ENTRY(ew_crop_up), defstr);
			g_snprintf(defstr, sizeof(defstr), N_("%.0f"), option_stats_down_crop);
			gtk_entry_set_text (GTK_ENTRY(ew_crop_down), defstr);

			gtk_widget_show(ignore_dialog);
		}
		break;

		default:	g_printerr (_("BUG: unknown menuitem number in %s\n"), __FUNCTION__);
		break;
	}
	stats_repaint();
}


// stats_button_press_event
//
// called when a button has been pressed on the screen
static gint
stats_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	// single right-click?
	if ((event) && event->type == GDK_BUTTON_PRESS && (event->button == 1 || event->button == 3) )
	{
		GtkWidget *menu, *mi;

		const gchar	*LAST_3MINS =		_(" last 3 minutes");
		const gchar *LAST_HOUR = 		_(" last hour");
		const gchar *LAST_6HOURS = 		_(" last 6 hours");
		const gchar *LAST_12HOURS = 	_(" last 12 hours");
		const gchar *LAST_24HOURS = 	_(" last 24 hours");
		const gchar *LAST_2DAYS = 		_(" last 2 days");
		const gchar *LAST_3DAYS =		_(" last 3 days");
		const gchar *LAST_WEEK =		_(" last week");
		const gchar *LAST_MONTH =		_(" last month");
		const gchar *LAST_YEAR =		_(" last year");

		gboolean stats_minutes_has_data = ((stats_minutes) && stats_minutes->len>0);
		gboolean stats_hours_has_data = ((stats_hours) && stats_hours->len>0);
		gboolean stats_days_has_data  = ((stats_days) && stats_days->len>0);

		menu = gtk_menu_new();
		mi = MENU_ITEM_ADD (LAST_3MINS, stats_popupmenu_item_selected, STATS_MENUITEM_3MIN, icon_stats);
		mi = MENU_ITEM_ADD (LAST_HOUR, stats_popupmenu_item_selected, STATS_MENUITEM_HOUR, icon_stats);
		gtk_widget_set_sensitive(mi, stats_minutes_has_data);
		mi = MENU_ITEM_ADD (LAST_6HOURS, stats_popupmenu_item_selected, STATS_MENUITEM_6HOURS, icon_stats);
		gtk_widget_set_sensitive(mi, stats_minutes_has_data);
		mi = MENU_ITEM_ADD (LAST_12HOURS, stats_popupmenu_item_selected, STATS_MENUITEM_12HOURS, icon_stats);
		gtk_widget_set_sensitive(mi, stats_hours_has_data);
		mi = MENU_ITEM_ADD (LAST_24HOURS, stats_popupmenu_item_selected, STATS_MENUITEM_24HOURS, icon_stats);
		gtk_widget_set_sensitive(mi, stats_hours_has_data);
		(void)	MENU_ITEM_ADD__________;
		mi = MENU_ITEM_ADD (LAST_2DAYS, stats_popupmenu_item_selected, STATS_MENUITEM_2DAYS, icon_stats);
		gtk_widget_set_sensitive(mi, stats_hours_has_data);
		mi = MENU_ITEM_ADD (LAST_3DAYS, stats_popupmenu_item_selected, STATS_MENUITEM_3DAYS, icon_stats);
		gtk_widget_set_sensitive(mi, stats_hours_has_data);
		(void)	MENU_ITEM_ADD__________;
		mi = MENU_ITEM_ADD (LAST_WEEK, stats_popupmenu_item_selected, STATS_MENUITEM_WEEK, icon_stats);
		gtk_widget_set_sensitive(mi, stats_hours_has_data);
		mi = MENU_ITEM_ADD (LAST_MONTH, stats_popupmenu_item_selected, STATS_MENUITEM_MONTH, icon_stats);
		gtk_widget_set_sensitive(mi, stats_days_has_data);
		mi = MENU_ITEM_ADD (LAST_YEAR, stats_popupmenu_item_selected, STATS_MENUITEM_YEAR, icon_stats);
		gtk_widget_set_sensitive(mi, stats_days_has_data);
		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, event->button, event->time);
	}
	return TRUE;
}



// stats_mode_button_clicked
//
// switches to the next display mode

static void
stats_mode_button_clicked (GtkWidget *widget, gpointer data)
{
	option_timeline_mode = (option_timeline_mode+1) % STATS_TIMELINE_DISPLAYMODE_LAST;
	stats_repaint();
}



// stats_create_page
//
// creates a drawing area for the statistics

GtkWidget *
stats_create_page(void)
{
	GtkWidget	*hbox, *vbox, *button_mode, *button_timeline, *button_ignore;

	vbox = gtk_vbox_new(FALSE,5);
	hbox = gtk_hbox_new(FALSE,5);

	chart = gtk_drawing_area_new();
	gtk_signal_connect (GTK_OBJECT(chart), N_("expose_event"), GTK_SIGNAL_FUNC(stats_expose_event), NULL);
	gtk_signal_connect (GTK_OBJECT(chart), N_("configure_event"), GTK_SIGNAL_FUNC(stats_configure_event), NULL);
	gtk_signal_connect (GTK_OBJECT(chart), N_("button_press_event"), GTK_SIGNAL_FUNC(stats_button_press_event), NULL);

	gtk_widget_set_events (GTK_WIDGET(chart), GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);

	new_button_pack_and_signal_connect (&button_mode, hbox, _(" switch graph mode "),
                                                stats_mode_button_clicked, NULL, 1);

	button_timeline = gtk_button_new_with_label (_(" switch interval displayed "));
	gtk_widget_show(button_timeline);
	gtk_signal_connect (GTK_OBJECT(button_timeline), N_("button_press_event"), GTK_SIGNAL_FUNC(stats_button_press_event), NULL);
	gtk_box_pack_start (GTK_BOX(hbox), button_timeline, FALSE, FALSE, 0);

	new_button_pack_and_signal_connect (&button_ignore, hbox, _(" ignore values... "),
                                                stats_popupmenu_item_selected, GUINT_TO_POINTER(STATS_MENUITEM_IGNORE_DIALOG), 1);

	gtk_box_pack_start (GTK_BOX(vbox), chart, TRUE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

	gtk_widget_show(hbox);
	gtk_widget_show(vbox);
	gtk_widget_show(chart);

	return vbox;
}


// stats_save_array_to_disk
//
// writes the data from the given array of samples to the given file,
//	using the marker provided

static void
stats_save_array_to_disk (GPtrArray *arr, gchar marker, FILE *f)
{
	g_return_if_fail (f!=NULL);
	if (arr)
	{
		PROCESS_GPTRARRAY (arr, statsRecord,
			if (item)
			{
				fprintf (f, N_("%c %f %f %u %u %lu\n"), marker, item->up, item->down,
							item->no_uploads, item->no_downloads, item->when);
			}
		);
	}
}


// stats_save_to_disk
//
// Saves some statistic data to disk
// * the general up/download stats (up/down total, secs up/downloaded)
// * the up/down rates over time

void
stats_save_to_disk (void)
{
	FILE		*f;
	const gchar	*filename = complete_pref_path (N_("gui_stats.txt"), GUI_CONFIG_USE_INSTANCE);

	g_return_if_fail ((filename));

	if ((f = fopen(filename, N_("w")) )!=NULL)
	{
		fprintf (f, N_("# ed2k_gui --- gui_stats --- config format version 0.3.0\n"));
		fprintf (f, _("#\n# This file is created by the ed2k gui. It contains your\n"));
		fprintf (f, _("# overall statistics. DO NOT CHANGE THE ABOVE HEADER!\n#\n"));
		fprintf (f, _("# total uploaded - total downloaded - stats last cleared - secs DL was nonzero - secs UL was nonzero\n"));
		fprintf (f, N_("GENERAL STATS: %f %f %lu %lu %lu\n"),
					total_uploaded,
					total_downloaded,
					stats_last_cleared,
					seconds_download_was_nonzero,
					seconds_upload_was_nonzero);
		fprintf (f, N_("FIRST OF TIMES: %lu %lu %lu %lu\n"),
					time_first_of_packets,
					time_first_of_minutes,
					time_first_of_hours,
					time_first_of_days);
		if (stats_packets) stats_save_array_to_disk (stats_packets, 'p', f);
		if (stats_minutes) stats_save_array_to_disk (stats_minutes, 'm', f);
		if (stats_hours) stats_save_array_to_disk (stats_hours, 'h', f);
		if (stats_days) stats_save_array_to_disk (stats_days, 'd', f);

		fclose (f);
	} else status_system_error_msg(_("couldn't save stats.txt"));
}

void
stats_read_from_disk (void)
{
	time_t		 newest_record = 0;
	gchar		**line, **linearray;
	guint		 vermaj, vermin, verminmin;
	const gchar	*cfn = complete_pref_path (N_("gui_stats.txt"), GUI_CONFIG_USE_INSTANCE);

	g_return_if_fail (cfn!=NULL);
	// Now read in the file
	linearray = misc_get_array_of_lines_from_textfile (cfn);
	if (!linearray)	// there should be at least our header!
	{
		g_printerr (_("gui_stats.txt: couldn't read file or file empty.\n"));
		return;
	}

	line = linearray;
	if ((*line==NULL) || sscanf(*line, N_("# ed2k_gui --- gui_stats --- config format version %u.%u.%u\n"), &vermaj, &vermin, &verminmin)!=3)
	{
		g_printerr (_("gui_stats.txt: invalid header?\n"));
		g_strfreev (linearray);
		return;
	}

	while (*(++line))
	{
		gfloat	up, down;
		guint	no_up, no_down;
		time_t	when;
		gchar	marker;

		if (**line=='#' || **line==0x00)
			continue;

		if (g_strncasecmp(*line, N_("GENERAL STATS:"), 13)==0)
		{
			sscanf(*line, N_("GENERAL STATS: %f %f %lu %lu %lu\n"),
				&total_uploaded,
				&total_downloaded,
				&stats_last_cleared,
				&seconds_download_was_nonzero,
				&seconds_upload_was_nonzero
			);
			continue;
		}

		if (g_strncasecmp(*line, N_("FIRST OF TIMES:"), 12)==0)
		{
			sscanf (*line, N_("FIRST OF TIMES: %lu %lu %lu %lu\n"),
				&time_first_of_packets,
				&time_first_of_minutes,
				&time_first_of_hours,
				&time_first_of_days
			);
			continue;
		}

		if (6 == sscanf(*line, N_("%c %f %f %u %u %lu\n"), &marker, &up, &down, &no_up, &no_down, &when))
		{
			static gfloat	  lastup=0.0, lastdown=0.0;
			statsRecord		 *statrec = g_new0(statsRecord,1);
			GPtrArray		**addarr = NULL;
			g_return_if_fail (statrec!=NULL);

			if (up >= option_stats_up_wrong)
				up = lastup;
			if (down >= option_stats_down_wrong)
				down = lastdown;

			lastup=up;
			lastdown=down;
			statrec->up = up;
			statrec->down = down;
			statrec->no_uploads = no_up;
			statrec->no_downloads = no_down;
			statrec->when = when;

			if (when>newest_record)
				newest_record=when;

			switch (marker)	// bad style, but shorter
			{
				case 'p': addarr = &stats_packets;
				case 'm': if (!addarr) addarr = &stats_minutes;
				case 'h': if (!addarr) addarr = &stats_hours;
				case 'd': if (!addarr) addarr = &stats_days;
				break;
				default: g_printerr (_("Unknown marker in gui_stats.txt (ignored)\n"));
				break;
			}

			if (addarr)
			{
				if (!*addarr) *addarr = g_ptr_array_new();
				g_ptr_array_add (*addarr, (gpointer)statrec);
			} else g_printerr (N_("addarr==NULL!?\n"));
		}
	}

	/* newest_record>0 can only happen if we've done the above loop else newest_record==0 */
	if (newest_record == 0)
	{
		g_printerr (_("gui_stats.txt: No valid stats records found after reading the file [harmless error].\n"));
		g_strfreev (linearray);
		return;
	} else {
		gint 	i;
		gint 	diff = time(NULL)-newest_record;
		time_t	newest_record_then = newest_record;	// while we register, the content in newest_record changes

		// Fill the time between last GUI shutdown and now with empty values
		if (diff>=3)
		{
			for (i=0; i<diff/3; i++)
				stats_register_upload_download_stats_packet (newest_record_then+(i*3), 0.0, 0.0, 0, 0);
		}
	}
	g_strfreev (linearray);
}


