/***************************************************************************
                          uploads_data.c  -  description
                             -------------------
    begin                : Fri Jul 26 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "downloads_data.h"
#include "global.h"
#include "shared_files.h"
#include "uploads_data.h"
#include "uploads_page.h"
#include "send_command.h"

GPtrArray	*uploads = NULL;


// functions
static void		uploads_free_record (uploadRecord *ul);



// uploads_add_record
//
// adds an upload to the uploads list and to the uploads CList

void
uploads_add_record (const guint8 *userhash, const gchar *filename, const gchar *username)
{
	uploadRecord	*ul;

	g_return_if_fail (userhash!=NULL);
	g_return_if_fail (filename!=NULL);
	if (!uploads) uploads = g_ptr_array_new();
	g_return_if_fail (uploads!=NULL);

	ul = g_new0 (uploadRecord, 1);	// sets all fields incl. filehash to zero as well
	g_return_if_fail (ul!=NULL);

	memcpy (ul->userhash, userhash, 16);
	ul->namequark = g_quark_from_string(filename);
	ul->username = (username) ? g_strdup (username) : g_strdup("(???)");
	ul->size = 0xFFFFFFFF;

	g_ptr_array_add (uploads, (gpointer)ul);	
	uploadsclist_add_record (ul);
}


// uploads_free_record
//
// frees an uploadRecord an all its elements

static void
uploads_free_record (uploadRecord *ul)
{
	g_return_if_fail (ul!=NULL);

	if (ul->username)
		g_free(ul->username);

	memset(ul,0x00,sizeof(ul));
	g_free(ul);
}


// uploads_remove_record
//
// removes a record from the uploads list and from the uploads CList

void
uploads_remove_record (uploadRecord *ul)
{
	g_return_if_fail (ul!=NULL);
	g_return_if_fail (uploads!=NULL);
	(void) g_ptr_array_remove (uploads, (gpointer)ul);	// order matters!
	uploadsclist_remove_record (ul);
	uploads_free_record(ul);
}


// uploads_clean_up
//
// clear upload list and CList

void
uploads_cleanup (void)
{
	if (uploads)
	{
		PROCESS_GPTRARRAY( uploads, uploadRecord,
			if (item) uploads_remove_record(item);
		);
		g_ptr_array_free (uploads,TRUE);
		uploads = NULL;
	}
	uploadsclist_clear();
}


// uploads_check_record_completeness
//
// checks whether we've found the filehash going with this upload, and if not,
//	tries to find it.
//
// returns TRUE if we have it/found it, or FALSE if not.

gboolean
uploads_check_record_completeness (uploadRecord *ul)
{
	static const guint8	 zerohash[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	const gchar			*filename;
	downloadRecord		*dl; 
	sharedfileRecord	*sf;
	static time_t		 last_sent_for_sf_update = (time_t) 0;

	g_return_val_if_fail (ul!=NULL, FALSE);

	if (ul->size!=0xFFFFFFFF || memcmp(ul->hash, zerohash, 16)!=0)
		return TRUE;

	// try to find hash from shared files by matching file name quark
	sf = shared_files_find_from_filenamequark (ul->namequark);
	if (sf)
	{
		memcpy (ul->hash, sf->hash, 16);
		ul->size = sf->filesize;
		return TRUE;
	}

	// try to find hash from downloads by matching filename
	filename = g_quark_to_string (ul->namequark);
	dl = downloads_find_from_filename (filename);
	if (dl)
	{
		memcpy (ul->hash, dl->hash, 16);
		ul->size = dl->size;
		return TRUE;
	}

	// still haven't found it? Probably our 'shared files' list is not up to date
	// => refresh it (only if it wasn't refreshed in the last three minutes)
	if (time(NULL) - last_sent_for_sf_update > 3*60);
	{
		send_get_shared_files();
		last_sent_for_sf_update = time(NULL);
	}

	return FALSE;
}

